/**
  @file /ai/Diplomacy/diplomacy.h
  @brief Header modulu diplomacie.
  
  Header modulu diplomacie.
  
  @author PZ
  @version 0.1
*/

#ifndef __PETR_ZITA_DIPLOMACY__
#define __PETR_ZITA_DIPLOMACY__

#pragma warning( disable : 4290 )

#include <deque>
#include <iterator>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "ai/ai_makra.h"
#include "ai/aiscriptids.h"
#include "ai/aiscriptcaller.h"
#include "common/exc.h"
#include "common/mm.h"
#include "common/utils.h"
#include "common/TCL/tcl_script.h"
#include "ai/Diplomacy/diptypes.h"
#include "ai/Diplomacy/dipcoefs.h"
#include "ai/Diplomacy/diplomacytransciever.h"

namespace ai_ns {

/** Prostor jmen obalujici modul Diplomacie. */
namespace diplomacy_ns {

/**
  Trida diplomatickeho enginu pocitajici a vyhodnocujici veskerou diplomacii.
*/
class CDiplomacyEngine : public CDiplomacyTransciever
{
public:
  /** Konstruktor */
  CDiplomacyEngine();

  /** Destruktor */
  ~CDiplomacyEngine();

  /** Metoda inicializujici ci resetujici diplomaticky engine. */
  void initDiplomacy();

  /** Metoda synchonizujici vsechny zaznamy diplomatu s aktualnim stavem. */
  void syncDiplomats();

  /** 
    Metoda vraci aktualni diplomaticky vztah mezi dvema diplomaty dle jejich World ID.
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
  */
  TRelationship getRelationshipByWorldIDs(int dip1,int dip2) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /**
    Metoda nastavuje aktualni diplomaticky vztah mezi dvema diplomaty dle jejich World ID.
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
  */
  void setRelationshipByWorldIDs(TRelationship relshp,int dip1,int dip2) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /** 
    Metoda vraci aktualni navrhovany diplomaticky vztah jednim diplomatem druhemu dle jejich World ID.
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
  */
  TRelationship getOfferRelationshipByWorldIDs(int dip1,int dip2) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /**
    Metoda nastavuje aktualni navrhovany diplomaticky vztah jednim diplomatem druhemu dle jejich World ID.
    Zaroven vraci synchronizacni informace.
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
    @return TPacket_DIPLOMACY_CHANGE*  paket se synchronizacnimi informacemi
  */
  TPacket_DIPLOMACY_CHANGE* setOfferRelationshipByWorldIDs(TRelationship relshp,int dip1,int dip2) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /**
    Metoda vraci aktualni vlastnosti diplomatickeho vztahu jednoho diplomata k druhemu dle jejich World ID. 
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
  */
  TRelationshipProperties getOtherRelationshipPropertiesByWorldIDs(int dip1,int dip2) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /**
    Metoda nastavuje aktualni vlastnosti diplomatickeho vztahu jednoho diplomata k druhemu dle jejich World ID. 
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
  */
  void setOtherRelationshipPropertiesByWorldIDs(TRelationshipProperties & relshp_prop,int dip1,int dip2) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /**
    Prida diplomata dle jeho typu a World ID do vnitrnich struktur modulu Diplomacie. 
    @throw E_8K_AI_Diplomacy_MaximumNumberOfDiplomats
    @throw E_8K_AI_Diplomacy_DiplomatWorldIDsNotUnique
  */
  void addDiplomat(int world_player_ID,TDiplomatType dt) throw(E_8K_AI_Diplomacy_MaximumNumberOfDiplomats,E_8K_AI_Diplomacy_DiplomatWorldIDsNotUnique);

  /**
    Odebere diplomata dle jeho World ID ze vnitrnich struktur modulu Diplomacie. 
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
  */
  void removeDiplomat(int world_player_ID) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /** Metoda zpusobi pro AI hrace vyhlaseni valky nejhorsimu nepriteli. */
  void AIWarDeclaration(int ai_diplomat_world_id,int enemy_diplomat_of_ai_world_id) throw(E_8K_AI_Diplomacy_NotAIDiplomat,E_8K_AI_Diplomacy_DiplomatNotFound);

  /** Metoda zmeni typ diplomata (AI/clovek). */
  void changeDiplomatType(int diplomat_world_id,TDiplomatType new_diplomat_type) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /**
    Metoda nalezne nejhorsiho nepritele pro daneho diplomata. 
    @param world_player_ID        World ID diplomata
    @param excluded_diplomats     seznam World ID diplomatu, kteri maji byt z hledani vynechani
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
    @return int                   World ID nejvetsiho nepritele
  */
  int findWorstEnemyByWorldIDs(int world_player_ID,std::list<int>* excluded_diplomats) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /**
    Metoda provadi synchronizaci na zacatku kola pro daneho diplomata.
    Vraci synchronizacni informace.
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
    @return TPacket_DIPLOMACY_CHANGE*  paket se synchronizacnimi informacemi
  */
  TPacket_DIPLOMACY_CHANGE* playerTurnBegin(int world_player_ID) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /**
    Metoda provadi diplomaticke akce na konci kola pro daneho diplomata.
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
  */
  void playerTurnEnd(int world_player_ID) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /**
    Metoda nalezne ID diplomata v modulu Diplomacie dle jeho World ID.
    @throw E_8K_AI_Diplomacy_DiplomatNotFound
  */
  int findDiplomatIDByWorldID(int world_player_ID) throw(E_8K_AI_Diplomacy_DiplomatNotFound);

  /**
    Metoda zapocita danou udalost do diplomatickych vztahu.
    @throw E_8K_AI_Diplomacy_RelationError
    @throw E_8K_AI_Diplomacy_UnknownEvent
    @throw E_8K_AI_Diplomacy_InvalidEventWeight
  */
  void calculateEvent(TEvent & new_event) throw(E_8K_AI_Diplomacy_RelationError,E_8K_AI_Diplomacy_UnknownEvent,E_8K_AI_Diplomacy_InvalidEventWeight);

  /** Metoda vrati pocet diplomatu, jez Diplomacie prave zpracovava. */
  int getDiplomatsNumber();


private:
  /** Pocet diplomatu, jez Diplomacie prave zpracovava. */
  int diplomats_number;

  /** Matice diplomatickych relaci mezi diplomaty. */
  TDipRelation DiplomatRelations[MAX_PLAYERS][MAX_PLAYERS];

  /** Pole se zaznamy o diplomatech. */
  TDipInfo DiplomatRecords[MAX_PLAYERS];

  /** Strutura s nactenymi koeficienty pro Diplomaticky Engine. */
  TCoefsRecord Coeficients;


  /** Metoda provede resetovani a reinicializaci cele diplomacie. */
  void clear();

  /**
    Metoda nacte koeficienty specificke pro Diplomacii z TCL skriptu.
  */
  void loadCoeficients();

  /** Metoda vrati priznak, zdali je na dane pozici aktivni diplomat. */
  bool diplomatDefined(int id);

  /** Metoda vrati prvni volnou pozici pro diplomata. */
  int getFirstFreeDiplomatSlot();

  /** Metoda vrati World ID diplomata na dane pozici. */
  int getDiplomatWorldID(int id);

  /** Metoda vraci aktualni diplomaticky vztah mezi diplomaty na danych pozicich. */
  TRelationship getRelationshipByIDs(int dip1,int dip2);

  /** Metoda nastavuje aktualni diplomaticky vztah mezi diplomaty na danych pozicich. */
  void setRelationshipByIDs(TRelationship relshp,int dip1,int dip2);

  /** Metoda vraci aktualni navrhovany diplomaticky vztah jednim diplomatem druhemu na danych pozicich. */
  TRelationship getOfferRelationshipByIDs(int dip1,int dip2);

  /** Metoda nastavuje aktualni navrhovany diplomaticky vztah jednim diplomatem druhemu na danych pozicich. */
  void setOfferRelationshipByIDs(TRelationship relshp,int dip1id,int dip2id);

  /** Metoda vraci aktualni vlastnosti diplomatickeho vztahu jednoho diplomata k druhemu na danych pozicich. */
  TRelationshipProperties getOtherRelationshipPropertiesByIDs(int dip1,int dip2);
  
  /** Metoda nastavuje aktualni vlastnosti diplomatickeho vztahu jednoho diplomata k druhemu na danych pozicich. */
  void setOtherRelationshipPropertiesByIDs(TRelationshipProperties & relshp_prop,int dip1,int dip2);
  
  /** Metoda nastavi hodnoty relace na neutralni. */
  void setNeutralRelation(TDipRelation & nr);

  /**
    Metoda vrati diplomaticky vztah odpovidajici dane duvere.
    @throw E_8K_AI_Diplomacy_InvalidBelief
  */
  TRelationship getBeliefRelationship(int bel) throw(E_8K_AI_Diplomacy_InvalidBelief);

  /**
    Metoda provede zmeny diplomatickych vztahu mezi dvema diplomaty vzhledem k jejich navrhovanym vztahum.
    @throw E_8K_AI_Diplomacy_InvalidBelief
  */
  void changeOfficialRelations(int dipid1,int dipid2);

  /**
    Metoda spocte a vrati pokutu, jenz musi diplomat zaplatim za poruseni diplomatickeho kontraktu.
    @throw E_8K_AI_Diplomacy_InvalidBelief
  */
  int countPenalty(TRelationship old_relationship,TRelationship new_relationship);

};


} // end of namespace "diplomacy_ns"
} // end of namespace "ai_ns"

#endif

