/**
  @file /ai/MapAnalyzer/mapanalyzer.h
  @brief Header modulu analyzy mapy.
  
  Header modulu analyzy mapy.
  
  @author PZ
  @version 0.1
*/

#ifndef __PETR_ZITA_MAPANALYZER__
#define __PETR_ZITA_MAPANALYZER__

#pragma warning( disable : 4290 )

#include <stdlib.h>
#include <time.h>

#include "ai/ai_makra.h"
#include "ai/MapAnalyzer/mapanalyzertransciever.h"
#include "common/exc.h"
#include "common/mm.h"
#include "common/types.h"
#include "common/utils.h"
#include "common/Interface.h"
#include "world/useful.h"
#include "world/plan/plan.h"
#include "world/buildings/building.h"
#include "world/units/unit.h"

namespace ai_ns {

/** Prostor jmen obalujici modul Analyzatoru mapy. */
namespace mapanalyzer_ns {

/** Datovy typ urcujici popisujici staveny most. */
typedef struct type_bridge_build_info
{
  int bridge_position;  ///< souradnice mostu
  int bridge_input;     ///< vstupni hex mostu
  int bridge_output;    ///< vystupni hex mostu
  int orientation;      ///< orientace mostu
} TBridgeBuildInfo;

/**
  Trida enginu analyzy mapy jenz pocita rozlicne ulohy a hleda optimalni pozice na datovych strukturach mapy.
*/
class CMapAnalyzerEngine : public CMapAnalyzerTransciever// trida engine pro analyzu mapy
{
public:
  /** Konstruktor. */
  CMapAnalyzerEngine();

  /** Destruktor. */
  ~CMapAnalyzerEngine();

private:
  /** Sada ukazatelu zpristupnujici veskere struktury WORLDu. */
  TPacket_AI_LetsGo* SIP;

  /** 
    Metoda inicializujici analyzator mapy predanim ukazatelu na struktury WORLDu.
    @param initPackage - sada ukazatelu na struktury WORLDu (TPacket_AI_LetsGo*)
  */
  virtual void initMapAnalyzer(TPacket_AI_LetsGo* initPackage);

  /** 
    Metoda vrati souradnice nejblizsi volne budovy daneho hrace, jenz je schopna vyprodukovat danou jednotku.
    @param playerID       World ID hrace
    @param hexid          souradnice hexu, od ktereho se vzdalenost pocita
    @param unit_type      ID typu jednotky jenz je pozadovana
    @return               souradnice nejblizsi prazdne budovy, jenz je schopna jednptku vyprodukovat
  */
  virtual int getNearestFreeProductionBuildingForUnitTrain(int playerID,int hexid,int unit_type);

  /** 
    Metoda vrati nejvhodnejsi misto pro postaveni produkcni budovy.
    @param playerID                         World ID hrace
    @param building_type                    ID typu produkcni budovy
    @param planned_buildings_coordinates    kontajner s hexy, na nichz hrac planuje stavet budovy
    @param assigned_builder_position        World ID stavitele, jenz byl stavbe prirazen (relevantni pouze pri modu preplanovani stavby pro konkretniho stavitele, jinak rovno (-1))
    @return                                 souradnice nejvhodnejsiho mista pro stavbu rpodukcni budovy
  */
  virtual int getBestProductionBuildingPlace(int playerID,int building_type,std::list<int>* planned_buildings_coordinates,int assigned_builder_position);

  /** 
    Metoda vrati pomyslny stred hracova uzemi.
    @param playerID                         World ID hrace
    @throw                                  E_8K_AI_MapAnalyzer_InvalidPlayerEmpireCenter
    @return                                 pomyslny stred hracova uzemi
  */
  virtual int findPlayerEmpireCenter(int playerID) throw(E_8K_AI_MapAnalyzer_InvalidPlayerEmpireCenter);

  /** 
    Metoda vrati prostou vzdalenost mezi dvema hexy.
    @param hexid1                           souradnice prvniho hexu
    @param hexid2                           souradnice druheho hexu
    @return                                 prostou vzdalenost mezi dvema hexy
  */
  virtual int hexDistance(int hexid1,int hexid2);

  /** 
    Metoda vrati ukazatel na strukturu informaci o meste na dane pozici.
    @param city_position                    souradnice hexu s mestem
    @throw                                  E_8K_AI_MapAnalyzer_CityNotFound
    @return                                 ukazatel na strukturu (TOWN*) informaci o meste
  */
  virtual World::TOWN* findCityByPosition(int city_position) throw(E_8K_AI_MapAnalyzer_CityNotFound);

  /** 
    Metoda vrati nahodny neprozkoumany hex pro daneho hrace.
    @param playerID                         World ID hrace
    @return                                 nahodny neprozkoumany hex pro daneho hrace
  */
  virtual int getRandomInvisibleHex(int playerID);
  
  /** 
    Metoda vrati akcni radius (kruhove okoli) zadaneho polomeru urciteho hexu jako seznam souradnic.
    @param hexid                         World ID hrace
    @param radius_size                   polomer kruhoveho okoli
    @return                              (TPacket_MapAnalyzer_ActionRadiusResponse*) obsahujici kontajner s hexy uvnitr kruhoveho okoli
  */
  virtual TPacket_MapAnalyzer_ActionRadiusResponse* getActionRadius(int hexid,int radius_size);
  
  /** 
    Metoda vrati nejvhodnejsi pozici a orientaci pro stavbu mostu pro daneho hrace.
    @param playerID                         World ID hrace
    @throw                                  E_8K_AI_MapAnalyzer_InvalidBridgeAnalysis
    @return                                 (TPacket_MapAnalyzer_BridgeBuildingResponse*) obsahujici pozici a orientaci mostu a navic i pozici stavitele pred stavbou mostu, nebo NULL, pokud neni zadny most potreba stavet
  */
  virtual TPacket_MapAnalyzer_BridgeBuildingResponse* bridgeBuildingAnalysis(int playerID) throw(E_8K_AI_MapAnalyzer_InvalidBridgeAnalysis);
  
  /** 
    Metoda vrati priznak, zdali je mozno vytrenovat jednotku urciteho typu v budove na zadanych souradnicich.
    @param building_placement               souradnice budovy
    @param unit_type                        typ trenovane jednotky
    @throw                                  E_8K_AI_MapAnalyzer_InvalidProductionBuildingPlacement
    @return                                 priznak, zdali je mozno vytrenovat jednotku urciteho typu v budove na zadanych souradnicich
  */
  virtual bool isUnitPossibleToTrainInBuildingOnHex(int building_placement,int unit_type) throw(E_8K_AI_MapAnalyzer_InvalidProductionBuildingPlacement);
  
  /** 
    Metoda vrati souradnice jednotky o zadanem World ID.
    @param unit_world_id                    World ID jednotky
    @throw                                  E_8K_AI_MapAnalyzer_UnitNotFoundOnMap
    @return                                 souradnice jednotky
  */
  virtual int findUnitPosition(int unit_world_id) throw(E_8K_AI_MapAnalyzer_UnitNotFoundOnMap);
  

  /** 
    Metoda vrati priznak, zdali nejake mesto ci budova sousedi, nebo primo stoji na danem hexu.
    @param hexid                            souradnice dotazovaneho hexu
    @return                                 priznak, zdali nejake mesto ci budova sousedi, nebo primo stoji na danem hexu
  */
  bool hexAdjacentToStatic(int hexid);

  /** 
    Metoda vrati priznak, zdali nejaka planovana budova sousedi, nebo primo stoji na danem hexu.
    @param planned_buildings_coordinates          kontajner se souradnicemi planovanych budov
    @param hexid                                  souradnice dotazovaneho hexu
    @return                                       priznak, zdali nejaka planovana budova sousedi, nebo primo stoji na danem hexu
  */
  bool hexAdjacentToPlannedBuildings(std::list<int>* planned_buildings_coordinates,int hexid);
  
  /** 
    Metoda vrati seznam sousednich hexu pro dany hex.
    @param hexid                    souradnice dotazovaneho hexu
    @return                         seznam sousednich hexu pro dany hex
  */
  COORDS_ARRAY getAdjacentHexes(int hexid);

  /** 
    Metoda vrati priznak, zdali je mozno postavit budovu na danem hexu.
    @param building_type                    typ budovy
    @param hexid                            souradnice dotazovaneho hexu
    @return                                 priznak, zdali je mozno postavit budovu na danem hexu
  */
  bool canBuildBuildingOnHex(int building_type,int hexid);
};

} // end of namespace "mapanalyzer_ns"
} // end of namespace "ai_ns"

#endif

