/**
  @file /ai/PathFind/pathenv.h
  @brief Header s makry a definicemi typu specifickych pro modul Pathfind.
  
  Header s makry a definicemi typu specifickych pro modul Pathfind.
  
  @author PZ
  @version 0.1
*/

#ifndef __PETR_ZITA_PATHFINDING_MACROS__
#define __PETR_ZITA_PATHFINDING_MACROS__

#include <deque>
#include <list>
#include <iterator>

#include "common/types.h"

/**
  Makra urcujici jaka verze pathfind/floodfill algoritmu bude pouzita.
  Mozne hodnoty PATHFIND_ASTAR_VERSION: astarSimple - zakladni verze A* pathfind algoritmu
                                        astarHeaps  - verze A* pathfind algoritmu vyuzivajici haldy
  Mozne hodnoty FLOODING_ASTAR_VERSION: astarFloodSimple - zakladni verze A* floodfill algoritmu
                                        astarFloodHeaps  - verze A* floodfill algoritmu vyuzivajici haldy
*/
#define PATHFIND_ASTAR_VERSION astarHeaps
#define FLOODING_ASTAR_VERSION astarFloodHeaps

/** Makra se specialnimi hodnotami pro floodfill. */
#define FLOOD_MASK_VALUE_CANT_SEE (-2)
#define FLOOD_MASK_VALUE_NOT_ACCESSIBLE (-1)

/** Makra pro typy STL kontajneru/iteratoru/datovych typu pouzitych v modulu Pathfind. */
#define TPATH_TYPE deque
#define TNEIGHBOURS_TYPE list
#define TPATH_FULL_TYPE std::TPATH_TYPE
#define TPATH TPATH_FULL_TYPE<PATH_COORDS>
#define TPATH_ITERATOR TPATH::iterator
#define TNEIGHBOURS_FULL_TYPE std::TNEIGHBOURS_TYPE
#define TNEIGHBOURS TNEIGHBOURS_FULL_TYPE<COORDS>
#define TNEIGHBOURS_ITERATOR TNEIGHBOURS::iterator
#define TLIST TNode*
#define THEAP THeap
#define PF_MAP TPFMap
#define PF_MARKMAP TPFMarkMap
#define FLOOD_MASK TFloodMask

/** Makro obsahujici porovnani konkretni pozice uzlu A* algoritmu. */
#define COORDSSAME(s1,s2) (((s1).x==(s2).x) && ((s1).y==(s2).y))

/** Makro transformujici souradnice pozice na linearni funkci. */
#define COORDS_TRANSFORM_FUNCTION(coords) (((coords).y*(getMapWidth()))+(coords).x)

/** Makro vracejici polozku v priznakove mape A* algoritmu na dane pozici. */
#define MARKMAP_CELL(mm,mmc) ((mm)[((mmc).y*(getMapWidth()))+(mmc).x])


namespace ai_ns {
namespace pathfind_ns {

/*
typedef struct type_linked_coordinates
{
  COORDS sou;
  struct type_linked_coordinates* next;
} LINKSOU;
*/

/** Definice datoveho typu uzlu A* algoritmu v jeho zakladni verzi. */
typedef struct type_node
{
  COORDS souuz;               ///< souradnice uzlu
  int g;                      ///< cena za dobyti tohoto uzlu
  int h;                      ///< heuristicky odhad z tohoto uzlu do cile
  int f;                      ///< soucet g + h
  struct type_node* next;     ///< ukazatel na dalsi uzel v listu
  struct type_node* parent;   ///< ukazatel na uzel, ze ktereho jsem do tohoto prisel
} TNode;

/** Definice polozek uzlu A* algoritmu ve verzi vyuzivajici haldy, ktere pri operacich v halde nemeni pozici. */
typedef struct type_heap_node_coords_and_values
{
  COORDS souuz;                                       ///< souradnice uzlu
  int g;                                              ///< cena za dobyti tohoto uzlu
  int h;                                              ///< heuristicky odhad z tohoto uzlu do cile
  int f;                                              ///< soucet g + h
  struct type_heap_node_coords_and_values* parent;    ///< ukazatel na uzel, ze ktereho jsem do tohoto prisel
  void* heap_node_pointer;                            ///< ukazatel na uzel v halde, kteremu hodnoty nalezi (pretypovavan na (THeapNode*))
} THeapNodeValues;

/** Definice datoveho typu uzlu A* algoritmu ve verzi vyuzivajici haldy. */
typedef struct type_heap_node
{
  THeapNodeValues* values;             ///< ukazatel na hodnoty nalezejici tomuto uzlu haldy
  struct type_heap_node* left;         ///< ukazatel na uzel leveho potomka v halde
  struct type_heap_node* right;        ///< ukazatel na uzel praveho potomka v halde
  struct type_heap_node* next;         ///< ukazatel na nasledujici uzel v halde v poradi prochazeni do sirky
  struct type_heap_node* prev;         ///< ukazatel na predchozi uzel v halde v poradi prochazeni do sirky
  struct type_heap_node* heap_parent;  ///< ukazatel na rodicovsky uzel v halde
} THeapNode;

/** Definice datoveho typu haldy pro A* algoritmus. */
typedef struct
{
  THeapNode* root;              ///< koren haldy
  THeapNode* last_added_node;   ///< posledni pridany uzel haldy
} THeap;

/** Definice datoveho typu s priznaky Pathfind transparence. */
typedef enum type_pftrnasparency
{
  PFT_UnitsNotTransparent,        ///< cesta nemuze prochazet skrze jakoukoliv jednotku
  PFT_IgnoreMyUnits,              ///< cesta muze prochazet pres jednotky hrace, pro jehoz jednotku se cesta hleda
  PFT_IgnoreTargetUnit,           ///< vyhledavani cesty ignoruje jakoukoliv pripadnou jednotku na cilovem hexu
  PFT_IgnoreMyUnitsAndTargetUnit, ///< kombinace dvou predchozich
  PFT_IgnoreAllUnits,             ///< cesta muze prochazet pres jakekoliv jednotky
  PFT_BridgeAnalysis              ///< cesta muze prochazet pres reky a more (berou se jako plan) a navic ignoruje vsechny jednotky
} TPathFindTransparency;

/** Definice datoveho typu s priznaky GetCost transparence. */
typedef enum type_gctrnasparency
{
  GCT_UnitsNotTransparent,        ///< jakakoliv jednotka na sousednim hexu se chova jako prekazka
  GCT_IgnoreMyUnits,              ///< jednotka daneho hrace na sousednim hexu neni brana jako prekazka
  GCT_IgnoreAllUnits,             ///< jakakoliv jednotka na sousednim hexu neni prekazkou
  GCT_IgnoreRivers,               ///< reka a more jsou pocitany jako louka
  GCT_SampleUnit                  ///< jednotka se chova jako lehka pechota, navic muze chodit pres reky a more
} TGetCostTransparency;

/** Definice datoveho typu priznakove masky A* algoritmu. */
typedef bool* TPFMarkMap;

/** Definice datoveho typu floodfill masky. */
typedef int* TFloodMask;

} // end of namespace "pathfind_ns"
} // end of namespace "ai_ns"

#endif
