/**
  @file /ai/Strategizer/planning.h
  @brief Header implementace planovacich struktur a interfacu pro komunikaci s WORLDem.
  
  Header implementace planovacich struktur a interfacu pro komunikaci s WORLDem.
  
  @author PZ
  @version 0.1
*/

#ifndef __PETR_ZITA_STRATEGY_PLANNING__
#define __PETR_ZITA_STRATEGY_PLANNING__

#pragma warning( disable : 4290 )

/** Makro se STL typem kontajneru pro akce v planu. */
#define PLAN_ACTIONS_CONTAINER list

/** Makro s hodnotou defaultni priority AI akce. */
#define DEFAULT_PLAN_ACTION_PRIORITY 5

#include <iostream>
#include <list>
#include "common/Msg.h"
#include "common/Log.h"
#include "common/Interface.h"
#include "common/exc.h"
#include "ai/ai_makra.h"
#include "ai/PathFind/pathenv.h"
#include "world/world_messages.h"
#include "world/history.h"

namespace ai_ns {
namespace strategizer_ns {

#ifdef AI_LOGGING
  /** Definice logovaciho souboru s AI akcemi. */
  static TLog LogAIActions("ai_actions.log");
#endif

/** 
  Zakladni trida pro jakoukoliv akci provadenou AI.
*/
class CPlanAction
{
public:
  /** Konstruktor. */
  CPlanAction();

  /** Destruktor. */
  // TODO: virtual?
  virtual ~CPlanAction();

  /** Funkce pretizeni operatoru "<" pro trideni akci v ramci planu dle priorit. */
  bool operator<(const CPlanAction* & opa);

  int id;                ///< id akce v planu
  int priority;          ///< priorita akce
  bool done;             ///< priznak, zdali byla akce uspesne provedena
  // TODO: odstranit pokud jde
  virtual void foo() {};
};

/** 
  Zakladni trida pro akci jednotky - je primo svazana s konkretni jednotkou zkrze jeji World ID.
*/
class CUnitPlanAction : public CPlanAction
{
public:
  /** Konstruktor. */
  CUnitPlanAction(int unit_id);

  /** Destruktor. */
  ~CUnitPlanAction();

  int unit_id;  ///< World ID jednotky
};

/** 
  Trida pro akci jednotky - pohyb.
*/
class CUnitPlanAction_Go : public CUnitPlanAction
{
public:
  /** Konstruktor. */
  CUnitPlanAction_Go(int unit_id);

  /** Destruktor. */
  ~CUnitPlanAction_Go();
  
  int path_target;                                          ///< souradnice cile pohybu jednotky
  ai_ns::pathfind_ns::TPathFindTransparency transparency;   ///< pozadovana transparence kterou World pouzije pri vyhledavani cesty
};

/** 
  Trida pro akci jednotky - utok.
*/
class CUnitPlanAction_Attack : public CUnitPlanAction
{
public:
  /** Konstruktor. */
  CUnitPlanAction_Attack(int unit_id);

  /** Destruktor. */
  ~CUnitPlanAction_Attack();
  
  int attack_target;      ///< souradnice hexu, na kterem se nachazi cil utoku
};

/** 
  Trida pro akci jednotky - leceni zranenych muzu.
*/
class CUnitPlanAction_HealWounded : public CUnitPlanAction
{
public:
  /** Konstruktor. */
  CUnitPlanAction_HealWounded(int unit_id);

  /** Destruktor. */
  ~CUnitPlanAction_HealWounded();
};

/** 
  Trida pro akci jednotky - elitni doplneni jednotek.
*/
class CUnitPlanAction_EliteRecruit : public CUnitPlanAction
{
public:
  /** Konstruktor. */
  CUnitPlanAction_EliteRecruit(int unit_id);

  /** Destruktor. */
  ~CUnitPlanAction_EliteRecruit();
};

/** 
  Trida pro akci jednotky - neelitni doplneni jednotek.
*/
class CUnitPlanAction_NonEliteRecruit : public CUnitPlanAction
{
public:
  /** Konstruktor. */
  CUnitPlanAction_NonEliteRecruit(int unit_id);

  /** Destruktor. */
  ~CUnitPlanAction_NonEliteRecruit();
};

/** 
  Trida pro akci jednotky - sebevrazda.
*/
class CUnitPlanAction_Release : public CUnitPlanAction
{
public:
  /** Konstruktor. */
  CUnitPlanAction_Release(int unit_id);

  /** Destruktor. */
  ~CUnitPlanAction_Release();
};

/** 
  Trida pro akci jednotky - stavba budovy.
*/
class CUnitPlanAction_Build : public CUnitPlanAction
{
public:
  /** Konstruktor. */
  CUnitPlanAction_Build(int unit_id);

  /** Destruktor. */
  ~CUnitPlanAction_Build();
  
  int build_target;                 ///< souradnice hexu, kde se bude budova stavet
  int building_type;                ///< typ budovy
  int building_orientation;         ///< orientace budovy
};

/** 
  Trida pro akci jednotky - oprava budovy.
*/
class CUnitPlanAction_Repair : public CUnitPlanAction
{
public:
  /** Konstruktor. */
  CUnitPlanAction_Repair(int unit_id);

  /** Destruktor. */
  ~CUnitPlanAction_Repair();
  
  int building_id;                  ///< World ID budovy, jenz ma byt opravena
};

/** 
  Trida pro akci jednotky - kouzleni.
  Tato trida neni vyuzivana - kouzleni nakonec neni jako samostatna akce jednotky, ale implementace je zde ponechana pro budouci rozsireni.
*/
class CUnitPlanAction_Spell : public CUnitPlanAction
{
public:
  /** Konstruktor. */
  CUnitPlanAction_Spell(int unit_id);

  /** Destruktor. */
  ~CUnitPlanAction_Spell();
  
  int spell_target;                 ///< hex, na ktery (ci na entitu na nem stojici) sesilam kouzlo
  int spell_type;                   ///< typ kouzla
};


/** 
  Zakladni trida pro akci budovy - je primo svazana s konkretni budovou zkrze jeji World ID.
*/
class CBuildingPlanAction : public CPlanAction
{
public:
  /** Konstruktor. */
  CBuildingPlanAction(int building_id);

  /** Destruktor. */
  ~CBuildingPlanAction();
  
  int building_id;                  ///< World ID budovy
};

/** 
  Trida pro akci budovy - vyrobeni jednotky.
*/
class CBuildingPlanAction_TrainUnit : public CBuildingPlanAction
{
public:
  /** Konstruktor. */
  CBuildingPlanAction_TrainUnit(int building_id);

  /** Destruktor. */
  ~CBuildingPlanAction_TrainUnit();

  int trained_unit_type;            ///< typ vyrabene jednotky
};

/** 
  Trida pro akci budovy - sebezniceni.
*/
class CBuildingPlanAction_Destroy : public CBuildingPlanAction
{
public:
  /** Konstruktor. */
  CBuildingPlanAction_Destroy(int building_id);

  /** Destruktor. */
  ~CBuildingPlanAction_Destroy();
};

/** 
  Trida pro akci budovy - prodej bonusu.
*/
class CBuildingPlanAction_SellBonus : public CBuildingPlanAction
{
public:
  /** Konstruktor. */
  CBuildingPlanAction_SellBonus(int building_id);

  /** Destruktor. */
  ~CBuildingPlanAction_SellBonus();

  int bonus_type;                   ///< typ prodavaneho bonusu
};


/** 
  Zakladni trida pro akci mesta - je primo svazana s konkretnim mestem zkrze jeho ID.
  Tato trida a vsechny od ni odvozene nejsou pouzivany, protoze mesta sama o sobe nakonec zadne akce nemaji. Implementace je zde vsak ponechana pro mozna budouci rozsireni.
*/
class CCityPlanAction : public CPlanAction
{
public:
  /** Konstruktor. */
  CCityPlanAction(int city_id);

  /** Destruktor. */
  ~CCityPlanAction();
  
  int city_id;                      ///< ID mesta
};

/** 
  Trida pro akci mesta - vyrobeni stavitele.
  Neni vyuzivana.
  @see CCityPlanAction
*/
class CCityPlanAction_TrainBuilder : public CCityPlanAction
{
public:
  /** Konstruktor. */
  CCityPlanAction_TrainBuilder(int city_id);

  /** Destruktor. */
  ~CCityPlanAction_TrainBuilder();
};


/** 
  Trida reprezentujici plan akci umeleho hrace - jedna se o posloupnost elementarnich akci, jenz jsou poslany na exekuci modulu World.
  Tato trida poskytuje jednoduche rozhrani pro tvorbu a provedeni takoveho planu.
*/
class CPlan
{
public:
  /**
    Konstruktor.
    @param player_id                        World ID (AI) hrace, pro ktereho se plan vytvari
  */
  CPlan(int player_id);

  /** Destruktor. */
  ~CPlan();

  /**
    Prida novou akci na zacatek planu.
    @param addedAction                      ukazatel na objekt akce, jenz je pridavana do planu
    @param process_plan_immediately         priznak, zdali ma po pridani akce byt i plan rovnou proveden (predan Worldu k exekuci)
  */
  int addActionToFront(CPlanAction* addedAction,bool process_plan_immediately);

  /**
    Prida novou akci na konec planu.
    @param addedAction                      ukazatel na objekt akce, jenz je pridavana do planu
    @param process_plan_immediately         priznak, zdali ma po pridani akce byt i plan rovnou proveden (predan Worldu k exekuci)
  */
  int addActionToBack(CPlanAction* addedAction,bool process_plan_immediately);

  /**
    Odstrani z planu akci daneho ID.
    @param actionID                         IDcko akce, jenz ma byt odstranena
  */
  void removeAction(int actionID);

  /**
    Odstrani z planu vsechny uspesne provedene akce (maji nastaveny priznak done na true).
  */
  void deleteAllDoneActions();

  /**
    Setridi akce v planu dle jejich priority (akce s nejvetsi prioritou budou na zacatku kontajneru).
  */
  void sortByPriorities();

  /**
    Provede postupne vsechny akce planu, respektive je preda Worldu k exekuci a nastavi done priznaky na true pro vsechny, jenz byly provedeny uspesne.
  */
  void processPlan();

private:
  int idcounter;          ///< ciselnik podle ktereho se generuji unikatni IDcka planu
  int player_id;          ///< World ID (AI) hrace, pro ktereho se plan vytvari (predane konstruktorem)

// pro Editor je pri testovani strategickeho grafu nutne i vypsat planovane akce - kontajner s akcemi musi mit public rozsah platnosti
#ifdef _K_EDITOR_
public:
#else
private:
#endif

  std::PLAN_ACTIONS_CONTAINER<CPlanAction*> actions;  ///< kontajner s planovanymi akcemi AI
};


} // end of namespace "strategizer_ns"
} // end of namespace "ai_ns"

#endif

