/**
*******************************************************************************

  @file /gui/common/ActiveObjects.h
  @brief okna, tlacitka, scrollbary,...
  
  @author Pavel
  @version 0.1
 
******************************************************************************/
#ifndef ACTIVEOBJECTS
#define ACTIVEOBJECTS

#include <SDL/SDL.h>
#include <vector>
#include "common/mm.h"
#include "common/rm/rminit.h"
#include "gui/common/KTime.h"

namespace gui {

/******************************************************************************/
/// Hodnoty INPUT::type.
#define INPUT_NO            0
#define INPUT_KEY_DOWN      1
#define INPUT_KEY_UP        2
#define INPUT_MOUSE_MOTION  3
#define INPUT_MOUSE_DOWN    4
#define INPUT_MOUSE_UP      5

/******************************************************************************/
/// Nastaveni.
#define LABEL_MAX_LENGTH  256

/******************************************************************************/

/// Struktura, kterou context (screen) posila ovladacim prvkum.
//@{

  /// Pozice mysi.
  struct MOUSEPOS{
    Sint16 x,y;
  };
  
  /// Klavesa (stiskla/pustena).
  struct KEY {
    /// Kod klavesy.
    SDLKey sym;
    /// Prelozeny unicode.
    Uint16 unicode;
  };
  
  /// Struktura obsahujici vstup z klavesnice nebo od mysi.
  struct INPUT {
    /// Typ vstupu.
    int type;
    /// S kazdym inputem se posila pozice mysi.
    MOUSEPOS mousepos;
    union {
      /// Klavesa (pro type == INPUT_KEY_UP, INPUT_KEY_DOWN).
      KEY key;
      /// Relativni pohyb mysi (pro type == INPUT_MOUSE_MOTION).
      MOUSEPOS relpos;
      /// Tlacitko mysi (pro type == INPUT_MOUSE_UP, INPUT_MOUSE_DOWN).
      Uint8 button;
    };
  };
//@}

/******************************************************************************/


class TBar;

/** Predek pro vsechny 2d objekty, ktere se v kontextu zobrazuji,
a zpracovavaji vstup. */
class TActiveObject //## 
{
protected:
  /// Tooltip s napovedou pro tenhle objekt.
  TBar *helpbar;
public:
  /// Indikator zda je prvek aktivni (mozne hodnoty AO_ACTIVE, AO_PASIVE).
  int active;

  /// Pozice a rozmery.
  //@{
  int x;
  int y;
  int width;
  int height;
  //@}

  /// Vnitrni stav (OR-kombinace AO_OVER, ...).
  int state;
  /// 32-bitova user data kazdeho prvku.
  int tag;

  
  /// Hlavni ridici funkce.
  //@{

  /** Zpracovava vstup.
    @param input je struktura upresnujici o jaky vstup se jedna.
  */
  virtual int workInput(INPUT *input)=0;
  /// Nakresli objekt.
  virtual void drawObject()=0;
  /// Konstruktor.
  TActiveObject();
  /// Virtualni destruktor (predevsim odalokovava pamet).
  virtual ~TActiveObject();
  //@}

  
  /// Pomocne ridici funkce.
  //@{
  /// Nastavi ze se bude buton kreslit a primat vstupy.
  virtual void show();
  /// Nastavi, ze button nebude primat vstupy ani se kreslit.
  virtual void hide();
  /// Prepne show/hide.
  virtual void showToggle();
  /// Obnovi stav prvku.
  virtual void restart();
  /// Vstup z klavesnice se nebude zpracovavat u nikoho jineho.
  void grabKeyboardInput();
  /// Uvolni vstup z klavesnice i pro ostatni.
  void releaseKeyboardInput();
  /// Odpovi zda mohu pouzit vstup z klavesnice.
  bool mayIUseKeyboardInput();
  //@}
  

  /// Funkce pro praci s tooltipy.
  //@{
  
  /** Prida tooltip ve forme radku textu.
    @param text je text ktery se bude zobrazovat.
  */
  void setHelp(wchar_t *text);
  /** Prida jeden radek textu tooltipu.
    @param text je text noveho radku.
  */
  void addHelpLine(wchar_t *text);
  /** Prida text na konec radku tooltipu.
    @param text je text na konci radku.
  */ 
  void addHelpWord(wchar_t *text);
  /** Prida libovolny tooltip.
    @param _helpbar je jakykoli TBar, ktery se bude zobrazovat.
  */
  void setHelp(TBar *_helpbar);
  /** Smaze tooltip.
  */
  void delHelp();
  //@}

};

/// vektor objektu:
typedef std::vector<TActiveObject*> TvActObj;
/******************************************************************************/

/// Mozne hodnoty promenne TActiveObject::active
//@{
#define AO_ACTIVE  1    ///< prvek se vykresluje a prima vstup
#define AO_PASIVE  0    ///< prvek spi
//@}

/******************************************************************************/

/** TActiveObject::state je | kombinaci nasledujicich:
  priklad: if (cudlik->state&AO_PUSHED==AO_PUSHED)
*/
//@{
#define AO_ALL                65535
#define AO_FOCUS              1
#define AO_OVER               2
#define AO_LEFT               4
#define AO_RIGHT              8
#define AO_CLICKED            16
#define AO_RIGHTCLICKED       32
#define AO_CHECKED            64
#define AO_SELECTED           128
#define AO_CHANGED            255
#define AO_NOT_FOCUS          (AO_ALL-AO_FOCUS)
#define AO_NOT_OVER           (AO_ALL-AO_OVER)
#define AO_NOT_LEFT           (AO_ALL-AO_LEFT)
#define AO_NOT_RIGHT          (AO_ALL-AO_RIGHT)
#define AO_NOT_CLICKED        (AO_ALL-AO_CLICKED)
#define AO_NOT_RIGHTCLICKED   (AO_ALL-AO_RIGHTCLICKED)
#define AO_NOT_SELECTED       (AO_ALL-AO_SELECTED)
#define AO_NOT_CHANGED        (AO_ALL-AO_CHANGED);
#define AO_CLEAR_CLICKS       (AO_ALL-AO_CLICKED-AO_RIGHTCLICKED)
#define AO_PUSHED             (AO_LEFT|AO_OVER|AO_FOCUS)
//@}

/// Dotazy na stav prvku.
//@{
#define PUSHED(x)       ((x->state&AO_PUSHED)==AO_PUSHED)
#define CLICKED(x)      ((x->state&AO_CLICKED)==AO_CLICKED)
#define MOUSEOVER(x)    ((x->state&AO_OVER)==AO_OVER)
#define CHECKED(x)      ((x->state&AO_CHECKED)==AO_CHECKED)
#define SELECTED(x)     ((x->state&AO_SELECTED)==AO_SELECTED)
#define FOCUSED(x)      ((x->state&AO_FOCUS)==AO_FOCUS)
#define CHANGED(x)      ((x->state&AO_CHANGED)==AO_CHANGED)
#define AO_MOUSEOVERME  ((input->mousepos.x>=x)&&(input->mousepos.y>=y)&&(x+width>input->mousepos.x)&&(y+height>input->mousepos.y))
//@}

/******************************************************************************/

/// Typicke hodnoty prvniho parametru u konstruktoru.
//@{
#define AO_NORMAL       0    ///< Objekt se normalne vykresluje.
#define AO_TOOLBAR      1    ///< Button se vykresluje jako button v toolbaru.
#define AO_HIDDEN       2    ///< Ojekt se nevykresluje, pouziva se u TBar a jeho potomku.
#define AO_VERTICAL     3    ///< Vertikalni Scrollbar.
#define AO_HORIZONTAL   4    ///< Horizontalni Scrollbar.
#define AO_UNPRESSABLE  5    ///< Button, ktery nemeni vzhled ani kdyz je zmackli (pouzivam ho pro scrollbar).
#define AO_UNSELECTABLE 5    ///< Item, ktera nelze oznacit.
#define AO_MOVABLE      0    ///< Window ktere jde premistovat.
#define AO_UNMOVABLE    6    ///< Window ktere nejde premistovat.
#define AO_NO           0    ///< CheckBox nezaskrtly.
#define AO_YES          1    ///< CheckBox zaskrtly.
#define AO_NOCHANGE		  -1   ///< Hodnota zustava stejna.
//@}

/******************************************************************************/

/** Jeden radek textu.
*/
class TLabel : public TActiveObject {
public:
  /// Barva textu.
  float light_col[4];
  /// Font textu.
  int font;
  /// Pamet pro text.
  wchar_t text[LABEL_MAX_LENGTH+1];

  /// Mozne zpusoby konstukce.
  //@{
  /**
    @param _text Retezec ktery se ma vykreslovat.
    @param _x Souradnice x.
    @param _y Souradnice y.
    @param _rl Mnozstvi cervene barvy z [0, 1].
    @param _gl Mnozstvi cervene barvy z [0, 1].
    @param _bl Mnozstvi cervene barvy z [0, 1].
  */
  TLabel(wchar_t *_text, int _x, int _y, float _rl, float _gl, float _bl);
  TLabel(wchar_t *_text, int _x, int _y);
  TLabel(wchar_t *_text, int _x, int _y, int _w, int _h, int _alignment);
  TLabel(int _x, int _y, wchar_t *format,...);
  //@}

  /// Zmena barvy textu.
  void setColor(float r, float g, float b, float alfa);
  
  /// Zpracovani vstupu a vystupu.
  //@{
  virtual void drawObject();
  virtual int workInput(INPUT *input);
  //@}
};

/** Maly obrazek (vyrez z bitmapy). */
class TIcon : public TActiveObject{
public:
  /// Data pro popis bitmapy.
  //@{
  rm::TSymbolData* data;
  HWTID hwtidIconTexture;       ///< textura ikony
  float tx1,ty1,tx2,ty2;        ///< poloha v texture
  float cr, cg, cb;             ///< barva materialu
  //@}

  /// Ridici funkce.
  //@{
  TIcon(rm::TSymbolData* _data, int _x, int _y);
  virtual void drawObject();
  virtual int workInput(INPUT *input);
  //@}
  
  /// Nastaveni.
  //@{
  void setIcon(rm::TSymbolData* _data);
  void setColor(float r, float g, float b);
  //@}
};

/** Maly obrazek (vyrez z bitmapy) s okrajem. */
class TIconExt: public TIcon{
public:
	/// Data popisujici okraj.
  //@{
  float bg[3];
  int border;
	float border_col[3];
  //@}

	/// Konstukce.
  TIconExt( rm::TSymbolData *_data, int _x, int _y, 
						float bg_r, float bg_g, float bg_b,
						int _border, float b_r, float b_b, float b_g);	
	
  /// Ridici funkce.
  //@{
  virtual void drawObject();
  //@}
};

/******************************************************************************/

/** Kurzor mysi. */
class TCursor : public TIcon {
public:
  /// Indikator, zda se jedna o kurzor, ktery na okraji obrazovky meni bitmapu na sipku.
  bool scrolling;
  /// Zobrazuji se presipaci hodiny/normalni kurzor
  bool wait;
  
  TCursor();
  /// Zpracovani vstupu.
  virtual int workInput(INPUT *input);

  /// Prepinani sipka/hodiny.
  //@{
  void setWait();
  void setBasic();
  //@}
};

/******************************************************************************/

/** Predek vseho na co lze kliknout.*/
class TButton : public TActiveObject {
protected:
  /// Klavesova zkratka.
  SDLKey shortcutKey;
  /// Modifikator klavesove zkratky (ctrl, ...)
  SDLMod modificator;
  /// Textura pozadi.
  HWTID hwtid;
public:
  /// Data urcujici vzhled.
  //@{
  /// False znamena, ze se na button neda klikat.
  bool enabled;
  int style;
  /// Jak se bude objekt vykreslovat, hodnota vypoctena na zaklade stavu.
  int appearence;
  float light_col[4];
  float dark_col[4];
  //@}
  
  /// Konstukce.
  TButton(int _style, int _x, int _y, int _width, int _height, float _rl, float _gl, float _bl);
  /// Konstukce II (pri pouziti konstruktoru bez barvy se pouzije standartni barva).
  TButton(int _style, int _x, int _y, int _width, int _height);
  /// Destruktor.
  ~TButton();

  /// Bezpecne premisteni prvku
  void move(int _x, int _y);
  /// Nastaveni klavesove zkratky
  void setShortcutKey(SDLKey _shortcutKey, wchar_t *shortcutKeyName = NULL, SDLMod _modificator = KMOD_NONE);

  /// Vstup/vystup
  //@{
  virtual void drawObject();
  virtual int workInput(INPUT *input);
  //@}
};

/** Tlacitko s textem. */
class TTextButton : public TButton {
public:
  TTextButton(int _style, wchar_t *_text, int _x, int _y, int _width, int _height, float _rl, float _gl, float _bl);
  TTextButton(int _style, wchar_t *_text, int _x, int _y, int _width, int _height);
  ~TTextButton();

  void move(int _x, int _y);
  TLabel *label;
  virtual void drawObject();
};

/// Tlacitko ktere pri kliknuti vyluzuje zvuk.
class TSoundedTextButton : public TTextButton {
public:
  int onclick_sample;
  TSoundedTextButton(int _style, wchar_t *_text, int _x, int _y, int _width, int _height, int _onclick_sample);
  virtual int workInput(INPUT *input);
};

/// Cudlik s obrazkem.
class TIconButton : public TButton {
public:
  TIcon *icon;
  TIconButton(int _style, rm::TSymbolData* _data, int _x, int _y, int _width, int _height, float _rl, float _gl, float _bl);
  TIconButton(int _style, rm::TSymbolData* _data, int _x, int _y, int _width, int _height);
  virtual void drawObject();
  ~TIconButton();
};
/******************************************************************************/

/// Predek pro vsechny panely (tj objekt ktery obsahuje nekolik objektu).
class TBar : public TButton{
public:
  /// Vektor objektu v nem.
  TvActObj vActObj;
  
  /// Konstrukce.
  //@{
  TBar(int _style, int _x, int _y, int _width, int _height, float _rl, float _gl, float _bl);
  TBar(int _style, int _x, int _y, int _width, int _height);
  //@}
  
  /// Ridici funkce.
  //@{
  /// Nakresli vsechno v sobe.
  virtual void drawObject();
  /// Vstup preda vsem potomkum.
  virtual int workInput(INPUT *input);
  /// Prida jeden objekt do seznamu objektu.
  virtual void Add(TActiveObject* ao);
  /// Smaze jeden objekt ze seznamu objektu.
  virtual int Del(TActiveObject* ao);
  /// Jako destruktor (smaze a odalokuje vsechno v sobe).
  virtual void DelNFreeAll();
  /// Restartuje sebe ale i potomky.
  virtual void restart();
  //@}

  /// Presune potomka do popredi.
  void setForground(TActiveObject* ao);
  /// Destrukce (smaze potomky).
  virtual ~TBar();
};
/** Pohyblivy/nepohyblivy Bar, ktery ma pozadi (implicitne polopruhledne, nebo
  se da zapnout textura. */
class TWindow : public TBar{
private:
  /// Pro pohybliva okna.
  MOUSEPOS click;
public:
  TWindow(int _style, int _x, int _y, int _width, int _height, float _rl, float _gl, float _bl);
  TWindow(int _style, int _x, int _y, int _width, int _height);
  virtual ~TWindow();
  
  /// Nakresli obrys a pozadi a potomky.
  virtual void drawObject();
  virtual int workInput(INPUT *input);
  virtual void hide();

  /// Nastavi texturu pozadi.
  void setTexture(HWTID _hwtidTex);
  void setBgImage(int bid);
  
  /// Pouze toto okno bude prijimat vstup.
  void showModal();
};

/// Predek pro kontext (osetruje preneseni okna do popredi, kdyz je na nej kliknuto).
class TFrame : public TBar{
private:
public:
  TFrame();
  virtual void drawObject();
  virtual int workInput(INPUT *input);
  virtual ~TFrame();
};
/******************************************************************************/

/** Panel ktery sdruzuje prvky ktere k sobe patri a udela kolem nich ramecek,
  vzdy jen jeden radioboxu v GroupBaru muze byt zaskrtly. */
class TGroupBar : public TBar {
public:
  TGroupBar(int _style, int _x, int _y, int _width, int _height, float _rl, float _gl, float _bl);
  TGroupBar(int _style, int _x, int _y, int _width, int _height);
  virtual void drawObject();
};

/// RadioBox, priklad pouziti: if SELECTED(radio) ...
class TRadioBox : public TIconButton {
public:
  /// Symboly pro jednotlive stavy radioboxu.
  //@{
  rm::TSymbolData *itselected, *itover, *itnone;
  //@}
  TRadioBox(int _selected, int _x, int _y, int _width, int _height);
  TRadioBox(int _selected, int _x, int _y, int _width, int _height, rm::TSymbolData* _itnone, rm::TSymbolData* _itselected, rm::TSymbolData* _itover);

  virtual void drawObject();
  virtual int workInput(INPUT *input);

  /// Explicitni nastaveni stavu AO_SELECTED/AO_NO.
  void setState(int newstate);
};

/******************************************************************************/

/** Vice radkovy text. */
class TMultiLabel : public TBar {
public:
  int rowcount;
  TMultiLabel(int _style, int _x, int _y, int _width, int _height, float _rl, float _gl, float _bl);
  TMultiLabel(int _style, int _x, int _y, int _width, int _height);

  /** Nastaveni textu
    @param _text Retezec ktery se ma zobrazit
    @param autoheight Rika zda ma Prvek prizpusobit svou vysku poctu radku textu.
  */
  void setText(wchar_t *_text, int autoheight = 0);
  void setColor(float r, float g, float b, float alfa);
};


/// CheckBox priklad pouziti: if CHECKED(combo) ...
class TCheckBox : public TIconButton {
public:
  TCheckBox(int _checked, int _x, int _y, int _width, int _height, float _rl, float _gl, float _bl);
  TCheckBox(int _checked, int _x, int _y, int _width, int _height);
  virtual void drawObject();
  virtual int workInput(INPUT *input);
};

/** Jednoradkovy vstup pro text. */
class TLineTextField : public TButton {
public:
  TLabel *label;
  /// Zda se ma okolo textu kreslit ramecek.
  bool hidden;
  /// Maximalni delka vstupu.
  int maxlength;
  /// Pozice kurzoru.
  int editorpos;

  void setText(wchar_t *newtext);
  wchar_t *getText();
  TLineTextField(int _x, int _y, int _maxlength = LABEL_MAX_LENGTH);
  TLineTextField(int _x, int _y, int _maxlength, float _rl, float _gl, float _bl);
  ~TLineTextField();

  /// Vstup/vystup
  //@{
  virtual void drawObject();
  virtual int workInput(INPUT *input);
  //@}
};
/******************************************************************************/

/// Jen pro vnitrni potrebu (jezdec na scrollbaru).
class TScrollButton : public TButton {
private:
  int orient;
  int winstart, winsize;
  MOUSEPOS click;
public:
  TScrollButton(int _type, int _x, int _y, int _width, int _height, int _winstart, int _winsize, float _rl, float _gl, float _bl);
  TScrollButton(int _type, int _x, int _y, int _width, int _height, int _winstart, int _winsize);
  virtual int workInput(INPUT *input);
};

/** Scrollbar.
*/
class TScrollBar : public TBar {
private:
  int orient;
  TIconButton *A,*B;
  TScrollButton *M;
  /// Representuje interval [0,Range) pri inicializaci Range = 100.
  int Range;
  /// Pozice jezdce.
  int Pos;
  /// Velikost jezdce z [1,Range], inicializace = 100.
  int Size;
  void CompSize();
  void CompPos();
public:
  
  /** Nastavi parametry scrollbaru.
    @param _Range [0, _Range) je interval, ve kterem se jezdec muze pohybovat.
    @param _Size  Velikost okenka/jezdce v tomto intervalu z [1, _Range].
    @param _Pos   pocatecni pozice jezdce v [0, _Range - _Size].
  */
  void ChangeTo(int _Range, int _Size, int _Pos);
  /// Vrati hodnotu Range.
  int range();
  /// Vrati hodnotu Pos.
  int pos();
  /// Vrati hodnotu size.
  int size();
  
  TScrollBar(int _type, int _x, int _y, int _width, int _height, float _rl, float _gl, float _bl);
  TScrollBar(int _type, int _x, int _y, int _width, int _height);  
  
  /// Vstup/vystup
  //@{
  virtual void drawObject();
  virtual int workInput(INPUT *input);
  //@}
};

/// Seznam polozek TItem se scrollbarem.
class TListBar : public TBar {
private:
  /// Vyska plochy pro itemy.
  int pixellength;
public:
  TScrollBar *scroll;
  TBar *bar;
  
  TListBar(int _style, int _x, int _y, int _width, int _height, float _rl, float _gl, float _bl);
  TListBar(int _style, int _x, int _y, int _width, int _height);
  
  /// Vraci ukazatel na prave vybranou polozku nebo NULL.
  TActiveObject* getSelectedObject();
  /// Vraci kolikata polozka je vybrana (pocitano od nuly) nebo -1.
  int getSelection();
 
  /// Pridavani/odebirani polozek
  //@{
  virtual void Add(TActiveObject* ao);
  virtual int Del(TActiveObject* ao);
  virtual void DelNFreeAll();
  //@}
  
  /// Vstup/vystup
  //@{
  virtual void drawObject();
  virtual int workInput(INPUT *input);
  void deselectAll();
  //@}
};

/// Polozka pro TListBar.
class TItem : public TBar {
public:
  TItem(int _selected, int _height, float _rl, float _gl, float _bl);
  TItem(int _selected, int _height);

  void setState(int state);
  virtual void Add(TActiveObject* ao);

  /// Vstup/vystup
  //@{
  virtual void drawObject();
  virtual int workInput(INPUT *input);
  //@}
};

/******************************************************************************/

/** Jadro ovladacich prvku. Existuje jedina instance, ktera
si pamatuje casto pouzivane/spolecne bitmapy. Stara se o
tooltipy, modalni okna, ...
*/
class TaoKernel //##
{
public:
  int ao_worked;
  /// true pokud jen jedno okno prijima vstup.
  bool modalmod;
  MOUSEPOS parentpos;
  int parentover;
  TCursor *cursor;
  TActiveObject *keyboardgraber;
  TBar *activeHelpBar;
  bool BlankedActiveHelpBar;
  TIME activeHelpBarSince;
  TWindow *modalWindow;

  /// Symbol interface.
  rm::TRM_symbol_i* rms;

  /// Prednahrane gui symboly
  //@{
  rm::TSymbolData*  sdCurBasic;
  rm::TSymbolData*  sdCurUp;
  rm::TSymbolData*  sdCurDown;
  rm::TSymbolData*  sdCurLeft;
  rm::TSymbolData*  sdCurRight;
  rm::TSymbolData*  sdCurWait;
  //
  rm::TSymbolData*  sdCross;
  rm::TSymbolData*  sdUpArrow;
  rm::TSymbolData*  sdDownArrow;
  rm::TSymbolData*  sdLeftArrow;
  rm::TSymbolData*  sdRightArrow; 
  //
  rm::TSymbolData*  sdRadioUnchecked;
  rm::TSymbolData*  sdRadioChecked;
  //@}
  
  /// Pomocnik pro nahravani symbolu.
  static rm::TSymbolData* getSymbol(int group, int sid); 
 
  /// Inicializuje kurzor a icony.
  void init();
  /// Odalokuje pamet.
  void done();

  /// Jedina instance.
  static TaoKernel* aoKernel;
  /// Inicializuje kurzor a icony.
  static void initAOKernel();
  /// Odalokuje pamet.
  static void freeAOKernel();

  /// Zjisti zda si nikdo nezamknul klavesnici.
  bool isKeyboardFree();

  /// Nastavi aktivni helpbar.
  void setActiveHelpBar(TBar *_activeHelpBar);
  /// Vynuluje helpbar.
  void setActiveHelpBarBlank();
};

/******************************************************************************/


/// Pomocne graficke funkce.
//@{
extern void glRectUp(int x, int y, int width, int height);
extern void glRectDown(int x, int y, int width, int height);
extern void glRectEmpty(int x, int y, int width, int height);
//@}

/// Do dest ulozi jmeno souboru ze source bez cesty a pripony.
extern void getRawFile(char *dest, char *source);

};  // namespace

#endif
