/**
******************************************************************************

  @file /gui/common/Screen.cpp
  @brief Trida reprezentujici obsah obrazovky

  Metody TScreen se musi volat synchronizovane.
  Makro FULLTROTTLE vyradi pasivni cekani ve vykreslovaci smycce.

  @author Vajicek
  @version 0.1

******************************************************************************/


#include "gui/common/Screen.h"
#include "gui/common/init_env.h"
#include "gui/engine/TextureManager.h"

/// perioda prekresleni
#define DELAY_FRAME		50

/// perioda zpracovani vstupu
#define DELAY_WORK		50

namespace gui{

void TScreen::sync_start(){
  if(screen_thread_id != SDL_ThreadID()){//nevolam ze stejneho vlakna
    SDL_LockMutex(thread_lock);
    SDL_LockMutex(screen_lock);
  }
}

void TScreen::sync_end(){
  if(screen_thread_id != SDL_ThreadID()){//nevolam ze stejneho vlakna
    SDL_UnlockMutex(screen_lock);
    SDL_UnlockMutex(thread_lock);
  }
}

void TScreen::selectContext(TContext* c, TContextParameter contextparameter){
  //nastav menici se kontext
  tconChangingContext = c;
  //zkopiruj parametr
  cpChangingContextParametr = contextparameter;
}

int TScreen::selectContext(char* szName, TContextParameter contextparameter){
  std::vector<TContext*>::iterator it;
  TContext* tc;
  //
  it=tconContexts.begin();
  while(it!=tconContexts.end()){
    tc=*it;
    if(!strcmp(tc->szName,szName))
      break;
    it++;
  }
  if(it==tconContexts.end())
    return -1;
  selectContext(tc, contextparameter);
  return 0;
}

void TScreen::addContext(TContext* c){
  tconContexts.push_back(c);
}

TContext* TScreen::getActiveContext(){
  //pokud probiha vymena, tak je za aktivni povazovan kontext, ktery se ma teprve aktivovat
  return tconChangingContext?tconChangingContext:tconActiveContext;
}

TContextParameter TScreen::enterMainLoop()
{
  SDL_Event event;
  Uint32 Tick;
  TContextParameter returnvalue;

  //
  iContinue=1;
  returnvalue.setDefault();

  LastDrawTick=0;
  LastWorkTick=0;

  //HLAVNI SMYCKA
  while(iContinue){

    //--- synchronizacni tanecek ---
    SDL_UnlockMutex(screen_lock);
    SDL_LockMutex(thread_lock);
    //tady hlavni vlakno stoji, kdyz probiha vymena dat na synchronizovanem vlakne
    SDL_LockMutex(screen_lock);
    SDL_UnlockMutex(thread_lock);

    //--- vymena kontextu se musi provest synchronne ---
    if(tconChangingContext){
      //DEACTIVATE & ACTIVATE CONTEXT
      if (tconActiveContext != NULL) tconActiveContext->deactivateContext();
      tconChangingContext->activateContext(cpChangingContextParametr);
      //
      tconActiveContext=tconChangingContext;
      tconChangingContext=NULL;
    }

    //--- nacitani textur na hlavnim vlakne ---
    TTextureManager::tm->loadTextures();

    //--- odchytnuti udalosti ---
    if (SDL_PollEvent(&event)){
      switch(event.type){
        case SDL_ACTIVEEVENT:{
        }break;
        case SDL_QUIT:{
          iContinue = 0;
        }break;
        case SDL_SYSWMEVENT:{
        }break;
        case SDL_VIDEORESIZE:{
        }break;
        case SDL_VIDEOEXPOSE:{
        }break;
        case SDL_USEREVENT:{
        }break;

        case SDL_KEYDOWN:
          input.type=INPUT_KEY_DOWN;
          input.key.sym=event.key.keysym.sym;
          input.key.unicode=event.key.keysym.unicode;
          processInput();
          break;          
        case SDL_KEYUP:
          // prozatim
          input.type=INPUT_KEY_UP;
          input.key.sym=event.key.keysym.sym;
          input.key.unicode=event.key.keysym.unicode;
          processInput();
          break;
        case SDL_MOUSEMOTION:
          input.type=INPUT_MOUSE_MOTION;
          input.mousepos.x = event.motion.x;
          input.mousepos.y = event.motion.y;
          input.relpos.x = event.motion.xrel;
          input.relpos.y = event.motion.yrel;
          processInput();
          break;
        case SDL_MOUSEBUTTONUP:
          input.type=INPUT_MOUSE_UP;
          input.mousepos.x = event.button.x;
          input.mousepos.y = event.button.y;
          input.button = event.button.button;
          processInput();
          break;
        case SDL_MOUSEBUTTONDOWN:
          input.type=INPUT_MOUSE_DOWN;
          input.mousepos.x = event.button.x;
          input.mousepos.y = event.button.y;
          input.button = event.button.button;
          processInput();
          break;
        break;
      } // switch(event.type)

    }// SDL_PollEvent(&event)
    else {
      // Kdyz se dlouho nic nestalo, tak dam contextu nahlidnout co je zmackly
      Tick = SDL_GetTicks();
      if (Tick - LastWorkTick >= DELAY_WORK) {
        LastWorkTick=Tick;
        input.type=SDL_NOEVENT;
        processInput();
      }

#ifdef FULLTROTTLE
      draw();
#else
      //vykresli kontext pokud je cas
      Tick = SDL_GetTicks();
      if (Tick - LastDrawTick >= DELAY_FRAME) {
        LastDrawTick=Tick;
        draw();
      }
      //nebo se uspi
      else{
        SDL_Delay(DELAY_FRAME - (Tick - LastDrawTick));
      }
#endif
    } // else
  } // while(iContinue)
  return returnvalue;
}

//vykresli obrazovku
int TScreen::draw()
{
  if(tconActiveContext){
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
    tconActiveContext->drawContext();
    SDL_GL_SwapBuffers();
  }
  return 0;
}

int TScreen::processInput()
{
  try{
    if (tconActiveContext) 
      tconActiveContext->workInput(&input);
    else 
      return 0;
  }
  catch(E_8K &e){
    KSendMessage(RQUEUE,MSG_EXCEPTION_IN_THREAD,MOD_GUI, BROADCAST, &e);
  }  
  return 1;
}

void TScreen::exitMainLoop(TContextParameter _returnvalue){
  iContinue=0;
}

TScreen::TScreen(int _w, int _h)
{
  //zamky
  screen_lock = SDL_CreateMutex();
  thread_lock = SDL_CreateMutex();
  screen_thread_id = SDL_ThreadID();

  //
  w = _w;
  h = _h;
  
  //
  tconActiveContext=NULL;
  tconChangingContext=NULL;
}

TScreen::~TScreen(){
  //zrusit zamky
  SDL_DestroyMutex(thread_lock);
  SDL_DestroyMutex(screen_lock);
}

void TScreen::switchResolution(int width, int height, int b, int fs){

  w = width;
  h = height;
  bpp = b;

  SwitchResolution(width, height, b, fs);
}

}//namespace
/*****************************************************************************/


