/**
******************************************************************************

  @file /gui/common/Screen.h
  @brief Trida reprezentujici obsah obrazovky
 

  @author Vajicek
  @version 0.1

******************************************************************************/
#ifndef TSCREEN_H
#define TSCREEN_H

#include <vector>
#include <SDL/SDL.h>

#include "gui/common/Context.h"
#include "common/Msg.h"
#include "common/Interface.h"

namespace gui{

/** Obrazovka.

  Reprezentuje obsah obrazovky, obsahuje hlavni smycku vykreslovani, zachytava 
  zpravy gui(klavesnice,mys), posila je do kontextu. Udrzuje kontexty. Registruje
  kontexty.Tato trida je final, nebude se nejspis dedit.
*/
class TScreen //## 
{
private:
  
  /// Data kontextu.
  //@{
  /// Vektor obsahujici pointery na dostupne kontexty.
  std::vector<TContext*> tconContexts;
  /// Aktivni kontext.
	TContext* tconActiveContext;
  /// Kontext, ktery se ma stat aktivnim, nebo NULL.
	TContext* tconChangingContext;
  /// Parametr pro kontext, ktery se ma stat aktivnim.
	TContextParameter cpChangingContextParametr;
  /// Priznak pokracovani hlavni smycky.
	int iContinue;
  //@}

  /// Zamky pro pripad pristupu z vlaken, zprav.
  //@{
  /// ID hlavniho vlakna
  Uint32 screen_thread_id;
  /// zamek obrazovky
  SDL_mutex *screen_lock;
  /// zamek pro pristupujici vlakno
  SDL_mutex *thread_lock;
  //@}

  /// Stavy vstupu.
  //@{
	Uint8 mbstate;  ///< mouse info
	INPUT input;    ///< informace o vstupu z klavesnice a mysi
  //@}

  /// Vnitrni promenne zajistujici pravidelne kresleni a osetrovani vstupu.
  //@{
	Uint32 LastDrawTick;
	Uint32 LastWorkTick;
  //@}
	
public:

  /// Synchronizace s hlavnim vlaknem.
  //@{
  void sync_start();
  void sync_end();
  //@}
	
  /// Manipulace s kontexty.
  //@{
  /** Prida kontext do seznamu.
    @param c ukazatel na pridavany kontext
  */
	void addContext(TContext* c);
  /** Odstrani ze seznamu. 
    @param c ukazatel na odstranovany kontext
  */
	void removeContext(TContext* c);
	/** Vrati aktivni kontext.
    Smi se volat jen na hlavnim vlakne, nebo v bloku "synchronizovanem" s hlavnim vlakne.
    @retval aktualne zvoleny kontext
  */
	TContext* getActiveContext();
  /** Aktivuje kontext.
    @param c pointer na kontext, ktery ma vyt aktivovan, nemusi byt v seznamu dostupnych
    @param contextparameter parametr se kterym se ma kontext spustit
  */
	void selectContext(TContext* c, TContextParameter contextparameter);
  /** Aktivuje podle jmena.
    @param szName jmeno kontextu ktery se ma kativovat, hleda se ve vektrou kontextu
    @param contextparameter parametr se kterym se ma kontext spustit
  */
	int selectContext(char* szName,  TContextParameter contextparameter);
  /** Vyhleda kontext.
    @param name jmeno kontextu, ktere se ma najit
    @retval kontex se jmenem szName, nebo NULL, kdyz neexistuje
  */
	TContext* findContext(char* name);
  //@}
 
	/** Hlavni smycka obrazovky. */
	TContextParameter enterMainLoop();

	/** Vykresli obrazovku. 
    Neosetruje vyjimky GUI, protoze by mohlo dojit k zacykleni.
  */
	int draw();
  
  /** Zpracuje vstup. 
    Osetruje vyjimky GUI.
  */
  int processInput();

	/** Preruseni hlavni smycky. 
    @param _returnvalue navratova hodnota enterMainLoop
  */
	void exitMainLoop(TContextParameter _returnvalue);

  /** Konstruktor. */
	TScreen(int _w, int _h);
  
	/** Destruktor. */
	~TScreen(); 
  
  /// Sirka
  int w;
  /// Vyska
  int h;
  /// Bitova hloubka.
  int bpp;

  /** Prepni rozliseni */  
  void switchResolution(int width, int height, int b, int fs);
  
};


}//namespace

#endif
/*****************************************************************************/
