/**
******************************************************************************

  @file /gui/common/Sounds.h
  @brief Modul pro prehravani zvuku
 

  @author Vajicek
  @version 0.1

******************************************************************************/
#ifndef SOUNDS_H
#define SOUNDS_H

#include "common/Interface.h"
#include "common/compatibility.h"
#include "common/rm/rminit.h"
#include "gui/common/mymath.h"
#include "gui/engine/Scene.h"
#include "gui/engine/Controllers.h"

#include <vector>
#include <SDL/SDL_mixer.h>

namespace gui{

/// Konstanta utlum.
#define DEFAULT_ATTENUATION 500.0f
/// Perioda obnoveni hlasitosti zvuku.
#define SOUNDINTERFACE_CONTROLER_UPDATE_PERIOD 100
/// Defaultni pocet kanalu.
#define DEFAULT_CHANNELS_COUNT 8
/// Maximalni pocet kanalu.
#define MAX_CHANNELS  32
/// Pocet urovni hlasitosti zvuku
#define SOUND_VOLUME_LEVELS	128
/// Pocet urovni hlasitost hudby
#define MUSIC_VOLUME_LEVELS	128

/** Nataveni kanalu.
*/
struct TChannel{
  /// identifikator kanalu
  int id;
  /// odkaz na prehravany sample
  Mix_Chunk* playing;
  /// priznak, jestli je kanal prirazen zdroji v prostoru
  int iSpatialControlled;
  /// hodnota utlumu
  float fLinearAttenuation;
  /// poloha zdroje v prostoru
  P3F p3fPosition;
  /// hlasitost
  int volume;
};

/** Rozhrani pro prehravani zvuku.
*/
class TSoundInterface //##
{
private:

  /// Data hudby.
  //@{
  /// hlasitos
  float musicVolume;
  /// id skladby, ktera se prehrava
  int playing_track_id;
  /// odkaz na reprezentaci prehravane skladby
  Mix_Music *playing_track;
  //@}

  /// Data kanalu
  //@{
  /// hlavni hlasitost zvuku
  float soundVolume;
  /// buffer na kanaly
  TChannel channels[MAX_CHANNELS];
  /// pocet pouzitych kanalu
  int channels_used;
  //@}

public:

  /// odkaz na interface pro RM
  rm::TRM_sound_i* SRM;

  /// Zvuky
  //@{
  /** Prehrej zvuk.
    @param rid id zvuku;
  */
  void playSample(int rid);
  /** Prehrej zvuk ze zdroje v prostoru.
    @param rid id zvuku.
    @param position Souradnice zvuku v prostoru.    
    @param latt Konstanta pro linearni utlumeni vzdalenosti.    
  */
  void playSampleSpatialConrolled(int rid, P3F* position, float latt = DEFAULT_ATTENUATION);
  /** Nastaveni hlasitosti pro vsechny zvuky.
    @param vol Hlasitost mezi 0 a 1.    
  */  
  void setSoundVolume(float vol);
  /** Ziskani hlasitosti zvuku.
    @return Hlasitost mezi 0 a 1.    
  */    
	float getSoundVolume();
  /** Stop prehravani.
  */      
  void stopAllChannels();
  //@}

  /// Hudba.
  //@{
  /** Pocet skaldeb.
    @return Pocet skladeb v RM.
  */
  int getMusicCount();
  /** Jmeno skladby.
    @param rid id skladby.
    @return Jmeno skladby.
  */
  char* getMusicName(int rid);
  /** Prehrani skladby.
    @param rid id skladby.  
  */
  void playMusic(int rid);
  /** Nastaveni hlasitosti hudby.
    @param vol Hlasitost mezi 0 a 1.    
  */  
  void setMusicVolume(float vol);
  /** Ziskani hlasitosti hudby.
    @return Hlasitost mezi 0 a 1.    
  */    
	float getMusicVolume();
  /** Okamzite zastaveni prehravani.
  */    
  void stopMusic();
  /** Postupne zastaveni prehravani.
    @param ms Doba utlumu na nulu v ms.    
  */    
  void fadeoutMusic(int ms);
  //@}

  //
  /** Inicializuje kanaly.
    @param chc Pocet kanalu.
  */    
  void initChannels(int chc);
  /** Konstruktor.
  */    
  TSoundInterface();
  /** Destruktor.
  */    
  ~TSoundInterface();

  /// Staticke rozhrani.
  //@{
  /// Jedina globalni instance.
  static TSoundInterface* si;
  /// Inicializace rozhrani.
  static void initSound();
  /// Zruseni rozhrani.
  static void uninitSound();
  //@}

  friend void channelDone_callback(int channel);
  friend void musicFinished_callback();
  friend class TSoundInterfaceController;
};

/** Kontroler manipulaci se zvukem
*/
class TSoundInterfaceController: public TController{
protected:
  /// Odkaz na rozhrani.
  TSoundInterface* si;
  /// Odkaz na kameru
  TCamera* cam; 
public:
  /** Update kontroleru.
    @see TController
  */
  void update(int absoluteTime);
  /** Konstruktor.
    @param _cam Kamera vzhledem ke ktere se ma pocitat utlum zdroju.
    @param _si  Zvukove rozhrani.
  */
  TSoundInterfaceController(TCamera* _cam, TSoundInterface* _si);
};

}//namespace
#endif
/*****************************************************************************/
