/**
*******************************************************************************

  @file /gui/model/hexy.h
  @brief Reprezentace hexu a proxyobjektu
  

  @author Pavel
  @version 0.1
 
******************************************************************************/

#ifndef HEXY_H
#define HEXY_H

#include "gui/engine/Scene.h"
#include "gui/model/selection.h"

namespace gui{

class THex;
class TguiMap;

/// Typ proxyobjektu.
enum TProxyType {ptStatic, ptDynamic, ptTerrain, ptBuilding};

/** Proxyobjekt.

  Proxyobjekt je predek pro kazdy logicky herni objekt, ktery se muze 
  vyskytnout na hexu. Potomci jsou teren, budova, jednotka. 

*/
class TProxyObject //##
{
protected:
  /// mapa
  TguiMap* map;
public:
	/// ukazatel na hex, ve kterem ProxyObject je
	THex *hexWhereIam;
	/// typ proxy objectu
	TProxyType type;
	/// naplni proxy object podle toho co je zac
	TProxyObject(TguiMap* m, TProxyType t);	
  /// presune ProxyObject do jineho hexu (vraci 0 pokud takovy hex neexistuje)
	int moveToHex(int newindex);
	/// presune ProxyObject do jineho hexu
	void moveToHex(THex *newHex);
	/// prida ProxyObject k danemu hexu, pokud to lze, jinak vraci 0
	virtual int addToHex(int newindex)=0;
  /// zjisti vysku na povrch prekazek, odliseni od map->altitude
  virtual void surfaceAltitude(float x, float y, float *z);
	/// zjisti, zda je na souradnicich [x, y], ktere spadaji do hexu, na kterem je tento proxyObject, prekazka
	virtual ~TProxyObject() {};
	/// nastavi oslabeni objektu
	virtual void setFade(int newfade) = 0;
  /// 
  virtual void dump2D(FILE* f) = 0;
};

/// Typ vektoru proxyobjektu.
typedef std::vector<TProxyObject*> Tproxyobjs;

/** Priznaky osvetleni hexu.  
  Osvetleni hexu(highlightState) se sklada z dvou casti
    -Tag - hornich 16bitu(uziv data)
    -Typ - dolnich 16bitu
*/
//@{
/// hodnota Typu, zadne osvetleni
#define hhNoHighlight       0
/// hodnota Typu, osvetleni obvodu hexu
#define hhHexHighlight      1 
/// hodnota Typu, osvetleni plochy hexu
#define hhHexAreaHighlight  2
/// makro nastavujici Tag osvetleni
#define hhtSetTag(x,t) ( (x&0xFFFF) | ((t)<<16) )
/// makro vracejici hodnotu Tagu osvetleni
#define hhtGetTag(x) ( (x)>>16 )
//@}


///indexy sousednich hexu
//@{
#define HN(id,w)  (id+w*2)
#define HS(id,w)  (id-w*2)
#define HNW(id,w) (((id/w)%2)?(id+w-1):(id+w))
#define HSE(id,w) (((id/w)%2)?(id-w):(id-w+1))
#define HNE(id,w) (((id/w)%2)?(id+w):(id+w+1))
#define HSW(id,w) (((id/w)%2)?(id-w-1):(id-w))
//@}

//FWD
class THex;

///hranice
//@{
#define STONES_PER_EDGE     3
#define BORDEREDGE_LENGTH   (HEX_EDGE_LENGTH-10)
//@}

/** Hranicni kamen */
class TBorderStone: public TModelSceneObject{
public:
  /// Odkaz na hex jehoz okrajem hranice prochazi.
  THex* hex;

	/** Vykresleni povrchu.
    @param h Hex.
    @param mod Model.
  */
  TBorderStone(THex* h, TModel* mod);
	/** Vykresleni povrchu.
    @see TModelSceneObject
  */
	virtual void drawObject();
};

/** Hex */
class THex: public TModelSceneObject, public THighlighting
{
friend class TProxyObject;
private:
	/// index hexu v mape
	int index;
	///model hexu
	TModel* model;
	//
	TguiMap* map;
public:
  /// Priznak hranic.
  int border_flag;
  /// Cislo kralovstvi
  int kingdom;
  /// Cislo hrace vlastniciho kralovstvi hexu.
  int owner;
	/// vektor proxy objektu
	Tproxyobjs objs;
	/// model vodni hladiny, pokud je nejaka
	TModel* watermodel;
	/// sceneobject vodni hladiny
	TModelSceneObject* waterso;
  /// scene object hranicnich kamenu
	std::vector<TModelSceneObject*> borderstones;  

	/** Konstruktor.
    @param ind Index.
    @param mod Model hexu.
    @param m Mapa.
  */
	THex(int ind, TModel* mod, TguiMap* m);
	/** Destruktor.
  */
	virtual ~THex(); 

  /// manipulace s proxyobjekty
  //@{
  
	/** Smaze z hexu jeden proxy object.
  
    @param po objekt ktery ma byt smazan
    @return smazany objekt
  */
	TProxyObject *remProxy(TProxyObject* po);
  
	/** Prida ukazatel na proxy.
  
    Prida ukazatel na proxyobejekt do seznamu hexu, ale neprida nic do sceny.

    @param po ukazatel na proxyobjekt ktery se ma pridat
  */
	void addProxy(TProxyObject* po);
  
  /** Vrati prvni proxyobjekt daneho typu.
  
    Predpokla se, ze hex obsahuje nanejvys jeden objekt daneho typu.
  
    @param t typ proxyobjektu
    @see TProxyType
    @return objekt daneho typu 
  */
	TProxyObject* getProxy(TProxyType t);
  //@}
  
  /** Zjisti stred hexu.
    @param hc ukazatel na bod, do ktereho se ulozi souradnice stredu hexu
  */  
  void getHexCenter(P3F* hc);

  /** Zjisti polohu na povrchu proxyobjektu na hexu
    
    Vola korekci na vysku na vsechn proxyobjektech a vrati nejvyssi.
    
    @param x vstupni souradnice x
    @param y vstupni souradnice y
    @param z vystupni souranice z
  */
  void altitudeCorection(float x, float y, float *z);


  /// Dedictvi po TSelectableObject.
  //@{
	/** Zrusi preselekci.
    @see TSelectableObject
  */
	virtual void preunselect();
	/** Nastavi preselekci.
    @see TSelectableObject
  */
	virtual void preselect();
	/** Zachyti kliknuti.
    @see TSelectableObject
  */
	virtual void onClick();
  //@}

  /** Prepni zvyrazneni.
    Nastavi barvu materialu hexu.
    @param on Priznak zvyrazneni.
  */
  void highlight(int on);

	/** Vykresleni zvyrazneni.
    @see THighlighting
  */
	virtual void drawHighlight();

  /** Vytvori hranice. 
  */
  virtual void createBorders();

	/** Vykresleni Hranice.
  */
	virtual void drawLineBorders();

	/** Vykresleni povrchu.
    @see TModelSceneObject
  */
	virtual void drawObject();

  /** Index hexu.
    @return index hexu.
  */
  int getIndex();

	/** Nastavi oslabeni objektu */
	void setFade(int newfade);

};

/// Typ ukazatele na hex.
typedef THex* PHex;

} // namespace;
#endif
/*****************************************************************************/
