/** @file /world/buildings/building.h
    Zakladni spolecne datove struktury pro vlastnosti staveb vyskytujicich se ve hre.

    Tyto struktury slouzi pro nacitani a ukladani dat z a do XML souboru a pro
    urceni spolecneho rozhrani mezi moduly World, GUI a AI.
  
    @author Petr Wolf
*/

#ifndef WORLD_TYPEDEF_BUILDING_H
#define WORLD_TYPEDEF_BUILDING_H

#include "common/mutex.h"
#include "common/TCL/tcl_script.h"
#include "world/common.h"
#include "world/plan/plan.h"
#include "world/struct_holder.h"
#include "world/tcl_xml.h"
#include "world/units/unit.h"

namespace World
{    
    /** Pozice, ze kter se budova stav/bour.
        Resi stavbu mostu, ktery by jinak neslo postavit, nebot
        vstup do reky je zakazan. 
    */
    enum CONSTRUCTION_POSITION
    {
        /** Stavba/bourn probh na hexu s budovou */
        CP_SAME_HEX

        /** Stavba/bourn probh na libovolnm sousednm
            hexu.
        */
       ,CP_ADJACENT_HEX
    };

    /// pocet moznych orientaci budovy
    #define BUILDING_ORIENTATION_COUNT 12

    /** Tvar budovy.
        Rozliseni, jestli se jedna o uzavrenou budovu, do ktere jednotky
        vstupuji (obvykla varianta), nebo o stavbu, po ktere se chodi (most)
    */
    enum STRUCTURE_TYPE
    {
        /** uzavrena budova, jednotky do ni vstupuji */
        ST_COVERED 

        /** otrevrena stavba, jednotky po ni chodi (most) */
       ,ST_OPEN
    };
    
    /** 
        Seznam moznych hernich akci budovy.
    */
    enum BUILDING_ACTION
    {
        /// dobrovolne zniceni budovy
        BA_SUICIDE

        /// vyroba jednotky v budove
       ,BA_RECRUIT_UNIT

        /// prodej bonusu jednotce v budove
       ,BA_SELL_BONUS
    };

    // pocet moznych akci budovy
    #define BUILDING_ACTION_COUNT 3

    /** Priznaky urcujici, co se muze o budove zobrazit.
    */
    enum BUILDING_KNOWLEDGE 
    {
        /// o budove nesmim zobrazit nic
        BK_NONE

        /// budova patri nepriteli, zobrazuji jen nektere polozky
       ,BK_ENEMY

        /// budova je moje, zobrazuji vsechno
       ,BK_MINE
    };

	/** Budova a jeji zakladni "mrtva" reprezentace.
	    Polozky podchycuji vlastnosti dane z pravidel pro dany typ budovy.
	*/
    struct BUILDING
    {
        /** Vojtovo ID */
        int id;

        /** ID nazvu budovy */
        int nameid;

        /** ID modelu */
        int modelid;

        /** ID modelu rozestavene budovy */
        int construction_modelid;

        /** ID zvuku */
        int soundsid[MAX_ACTION_SOUNDS];

        /** ID ikony */
        int symbolid;

        /** Pravdepodobnost zasahu strojem
            Vztahuje se i na jednotky v budov ukryte. Mela by primo odpovidat
			rozloze budovy.
        */
        float hit_probability;
        
        /** Pocet "zivotu" (maximalni) stavby */
        int lives;

        /** Cena za zahajen stavby */
        int base_cost;

        /** Cena za opravu jednoho "zivota" */
        int repair_cost;

        /** Doba stavby 
            Pocet kol, po ktere trva vybudovani stavby
        */
        int construction_duration;

        /** Pocet orientaci */
        int orientation_count;
               
        /** Pozice jednotky pro stavbu budovy */
        CONSTRUCTION_POSITION construction_position;
        
        /** Struktura stavby */
        STRUCTURE_TYPE structure_type;

        /** Bonus k obrane pro jednotku uvnitr */
        int defense_bonus;

        /** Sila protiutoku */
        int counterattack;

        /** Dosah videni budovy samotne */
        int eyesight_distance;

        /** Seznam terenu, na kterych lze budovu stavet */
        int terrain[TERRAIN_TYPES_COUNT+1];        

        /** Vroba jednotek 
            Seznam jednotek, kter lze v dan budov rekrutovat, lecit nebo doplnovat

            Zda bude mozne v dane budove skutecne jednotku koupit zalezi
            take na tom, jestli se dany typ jednotky vyskytuje na hexu,
            na kterem je budova postavena. Obe podminky pak funguji 
			v pruniku.

            @see HEX.unit_occurence
        */
		K8_ARRAY<UNIT_TYPE_ID> unit_recruitment;
		
		/** Bonusy
		    Navyseni vlastnosti nebo uceni kouzel pro jednotky, ktere
		    je mozne v budove kupovat.
		*/
		K8_ARRAY<UNIT_PROPERTY_BONUS_ID> bonuses;
    };

    /** reprezentace "zivouci" budovy v mape */
    struct LIVING_BUILDING
    {
        /** Typ budovy */
        BUILDING_TYPE_ID type;

		/** hrac, ke kteremu budova patri */
		PLAYER_ID player;

		/** jmeno budovy */
		K8_CHAR name[MAX_STRLEN];

		/** Pocet "zivotu" budovy */
		int lives;

        /** U rozestavene budovy pocet kol, po ktere je jiz budova stavena.
            U dostavene je hodnota vzdy 0 
        */
        int construction_duration;

        /** orientace budovy */
		int orientation;		
    };

    /** Typ budovy
        Objektovy obal struktury BUILDING
    */
    class TBuilding : public TStructHolder<BUILDING>, public TTCL_Struct
    {
    public:
        /// Konstruktor
        TBuilding();

        /// Konstruktor prejimajici nactena data
        TBuilding(BUILDING * data, int id = 0);

        /// nacteni z TCL
        void readFromTCL(TTCL_Interpreter & interpreter, const char * varName);
        
        /// zapis do TCL
        void writeToTCL(TTCL_Interpreter & interpreter, const char * varName);
    };
    
    /** Zivouci budova
        Objektovy obal struktury LIVING_BUILDING
    */        
    class TLivingBuilding : public TStructHolder<LIVING_BUILDING>, public TTCL_XML_Struct, public TMutex 
    {
    private:
        /// hex, na kterem budova stoji
        HEX_ID _hex;

        /// jednotka, ktera je v budove
        UNIT_ID _unit;

        /// odkaz na reprezentaci sveta
        World::TWorld * _world;
        
        friend class TWorld;
        
        friend class TWorldEngine;
    public:
        /// Konstruktor
        TLivingBuilding();

        /// Konstruktor prejimajici nacnena data
        TLivingBuilding(LIVING_BUILDING * data, int id = 0);

        /// Konstruktor prejimajici ID budovy
        TLivingBuilding(int id);

        /// nacteni z XML
        void readFromXML(TPackage * package);

        /// zapis do XML
        void writeToXML(TPackage * package);

        /// nacteni z TCL
        void readFromTCL(TTCL_Interpreter & interpreter, const char * varName);
        
        /// zapis do TCL
        void writeToTCL(TTCL_Interpreter & interpreter, const char * varName);

        /// Cislo hexu, na kterem budova stoji
        HEX_ID hex();

        /// Cislo jednotky, ktera je v budove 
        UNIT_ID unit();
    };
}

#endif
/*****************************************************************************/
