/** @file /world/da_holder.h

    Obecna sablona pro kontejner dat pracujici nad tridou std::map,
    zvladajici spolupraci s Vojtovou tridou DA.

    @author Petr Wolf
*/

#ifndef WORLD_DA_HOLDER_H
#define WORLD_DA_HOLDER_H

#include <map>
#include "common/xml/package.h"
#include "common/xml/packsmall.h"
#include "common/TCL/tcl_script.h"
#include "common/mm.h"
#include "common/da.h"
#include "common/mutex.h"
#include "world/struct_holder.h"

namespace World
{
    /** Obecna sablona pro kontejner dat pracujici nad tridou std::map,
        zvladajici spolupraci s Vojtovou tridou DA.
    */
    template<class R, class S, class T>
    class TDAHolder : public std::map<S, T *> 
    {
    protected:
        S _increment;
    public:
        /** Konstruktor */
        TDAHolder();

        /** Vyprazdneni kontejneru dat */
        void reset();

        /** Nacteni z dynamickeho pole */
        void load(DA<R *> * data);

        /** Export do dynamickeho pole 
            @param dealloc Priznak, maji-li se data pri ulozeni rovnou vymazat z worldu.
                           \c True je standardni hodnota, nebot pri ukladani do RM pak 
                           dojde k opetovnemu nahrani dat. Zato pri ukladani z klienta
                           je nutne nastavit parametr na \c false, jinak se nepovede ani
                           samotne ulozeni.
        */
        DA<R *> * save(bool dealloc = true);

        S auto_increment();

        /** Zapis dat do XML.
            @param package  XML objekt, do ktereho se zapisuje
            @param itemTag  nazev tagu pro prvky seznamu
        */
        void readFromXML(TPackage * package, char * itemTag);
        
        /** Nacteni dat z XML.
            @param package  XML objekt, do ktereho se zapisuje
            @param itemTag  nazev tagu pro prvky seznamu
        */
        void writeToXML(TPackage * package, char * itemTag);

        /** Zapis dat do prostredi TCL interpretu.
            @param interpreter  interpret jazyka TCL
            @param name         nazev promenne odpovidajici seznamu
        */
        void writeToTCL(TTCL_Interpreter & interpreter, const char * name);
    };

    template<class R, class S, class T>
    TDAHolder<R, S, T>::TDAHolder()
    {
        _increment = (S)1;
    }

    template<class R, class S, class T>
    void TDAHolder<R, S, T>::reset()
    {
        typename TDAHolder<R,S,T>::iterator i;

        for (i = this->begin(); i != this->end(); i++) 
            delete i->second;
        this->clear();
        
        _increment = (S)1;
    }

    template<class R, class S, class T>
    S TDAHolder<R, S, T>::auto_increment()
    {
        return _increment++;
    }
    
    template<class R, class S, class T>
    void TDAHolder<R, S, T>::load(DA<R *> * data)
    {
        for (int i = data->getNext(-1); i != -1; i = data->getNext(i)) {
            (*this)[i] = new T((*data)[i], i);
            _increment = MAX(_increment, i + 1);
        }      
    } 

    template<class R, class S, class T>
    DA<R *> * TDAHolder<R, S, T>::save(bool dealloc)
    {
        DA<R *> * result = new DA<R *>;
        
        typename TDAHolder<R,S,T>::iterator i;

        for (i = this->begin(); i != this->end(); i++) {
            (*result)[i->first] = &i->second->data();
            if (dealloc) {
                delete i->second;
                (*this)[i->first] = NULL;
            }
        }

        if (dealloc) {
            this->clear();
        }

        return result;
    }

    template<class R, class S, class T>
    void TDAHolder<R, S, T>::readFromXML(TPackage * package, char * itemTag)
    {     
        // nacteni prvku
        TXMLdata * xml_data = package->getXML();
        DA<int> ids;
        xml_data->getListOfIds(&ids, NULL, 0, NULL);

        for (int i = ids.getNext(-1); i != -1; i = ids.getNext(i)) {
            if (ids[i] == 1) {       
                (*this)[i] = new T(i);
                (*this)[i]->readFromXML(new TPackSmall(package->getSubtag(itemTag, i, NULL)));
                _increment = MAX(_increment, i + 1);
            }
        }
    }

    template<class R, class S, class T>
    void TDAHolder<R, S, T>::writeToXML(TPackage * package, char * itemTag)
    {     
      // ulozeni prvku
        TPackSmall * xml_item;
        typename TDAHolder<R,S,T>::iterator i;

        for (i = this->begin(); i != this->end(); i++) {
            package->setEmpty(itemTag, i->first, NULL);
            xml_item = new TPackSmall(package->getSubtag(itemTag, i->first, NULL));            
            i->second->writeToXML(xml_item);
            delete xml_item;
        }
    }

    template<class R, class S, class T>
    void TDAHolder<R, S, T>::writeToTCL(TTCL_Interpreter & interpreter, const char * name)
    {
        typename TDAHolder<R, S, T>::iterator i;
        char itemName[MAX_STRLEN];
        
        for (i = this->begin(); i != this->end(); i++) {
            snprintf(itemName, MAX_STRLEN, "%s(%d)", name, i->first);
            i->second->writeToTCL(interpreter, itemName);
        }
    }    
    
    /** Vlaknove bezpecna varianta tridy TDAHolder.
        Funkcne ekvivalentni tride TDAHolder, dedi ale navic po tride TMutex umoznujici
        zamykani a odemykani.
    */
    template<class R, class S, class T>
    class TDAHolder_Mutex : public TDAHolder<R, S, T>, public TMutex
    {

    };
};

#endif 
/*****************************************************************************/
