/** @file /world/history.h
    Ukladani posloupnosti hernich akci, ktere se udaly.

    @author Petr Wolf
*/

#ifndef WORLD_HISTORY_H
#define WORLD_HISTORY_H

#include <map>
#include "common/mm.h"
#include "world/typedefs.h"

namespace World
{
    /// typ herni udalosti
    enum ACTION_TYPE
    {
        /// nezadano
        AT_NONE

        /// zacatek hry
       ,AT_GAME_START

        /// hrac porazen
        /// @see TPacket_PLAYER_DEFEATED
       ,AT_PLAYER_DEFEATED       

        /// konec hry
        /// @see TPacket_END_GAME
       ,AT_ENDGAME       

        /// pohyb jednotky
        /// @see TPacket_RCT_UNIT_MOVE
       ,AT_UNIT_MOVE

        /// utok jednotky
        /// @see TPacket_RCT_UNIT_ATTACK
       ,AT_UNIT_ATTACK

        /// dobrovolna smrt jednotky
        /// @see TPacket_RCT_UNIT_SUICIDE
       ,AT_UNIT_SUICIDE

        /// dobrovolne sebezniceni budovy
        /// @see TPacket_RCT_BUILDING_SUICIDE
       ,AT_BUILDING_SUICIDE

        /// vyroba jednotky
        /// @see TPacket_RCT_BUILDING_RECRUIT_UNIT
       ,AT_BUILDING_RECRUIT_UNIT

        /// prodej bonusu
        /// @see TPacket_RCT_SELL_BONUS
       ,AT_BUILDING_SELL_BONUS       

        /// zacatek stavby budovy
        /// @see TPacket_RCT_START_BUILDING
       ,AT_UNIT_BUILD_START       

        /// preruseni stavby budovy
        /// @see TPacket_RCT_STOP_BUILDING
       ,AT_UNIT_BUILD_STOP       

        /// zapoceti opravy budovy
        /// @see TPacket_RCT_REPAIR_BUILDING
       ,AT_UNIT_REPAIR_START       

        /// leceni jednotky
        /// @see TPacket_RCT_UNIT_HEAL
       ,AT_UNIT_HEAL
              
        /// doplneni jednotky (neelitni)
        /// @see TPacket_RCT_UNIT_RECRUIT
       ,AT_UNIT_RECRUIT

        /// doplneni jednotky (elitni)
        /// @see TPacket_RCT_UNIT_RECRUIT_ELITE
       ,AT_UNIT_RECRUIT_ELITE

        /// konec tahu
        /// @see TPacket_RCT_NEXT_TURN
       ,AT_NEXT_TURN       

        /// konec kola
        /// @see TPacket_RCT_NEXT_TURN
       ,AT_NEXT_ROUND

        /// obsazeni budovy
        /// @see TPacket_BUILDING_OCCUPIED
       ,AT_BUILDING_OCCUPIED
       
        /// obsazeni mesta
        /// @see TPacket_TOWN_OCCUPIED
       ,AT_TOWN_OCCUPIED

        /// obsazeni kralovstvi
        /// @see TPacket_KINGDOM_OCCUPIED
       ,AT_KINGDOM_OCCUPIED

        /// zmena viditelnosti
        /// @see TPacket_VISIBILITY_CHANGE
       ,AT_VISIBILITY_CHANGE

        /// zmena v diplomatickem vztahu
        /// @see TPacket_DIPLOMACY_CHANGE
       ,AT_DIPLOMACY_RELATION_CHANGE
    };

    /** Herni udalost */
    class TAction
    {
    private:
        ACTION_TYPE _type;
        void *      _data;
    public:
        TAction();
        TAction(ACTION_TYPE type, void * data = NULL);

        ACTION_TYPE type();
        void * data();
    };
    
    /// seznam ID hracu
    typedef std::vector<int> TPlayerIdContainer;

    /** Obalka herni udalosti obsahujici navic informaci o poctu hracu, 
        kteri si ji jiz zaznamenali 
    */
    class TActionContainer 
    {
    private:
        TAction * _data;
        bool      _synchronized[MAX_PLAYERS];
    public:        
        TActionContainer();
        TActionContainer(TPlayerIdContainer & players);
        TActionContainer(TAction * data, TPlayerIdContainer & players);

        TAction & data();
        bool isPlayerSynchronized(const PLAYER_ID player_id);
        void setPlayerSynchronized(const PLAYER_ID player_id);
        bool arePlayersSynchronized();
    };

    /** Fronta hernich udalosti.

        Trida slouzi k ukladani historie hernich akci v poradi,
        v jakem sly za sebou.  

        Z tvorene fronty se odesilaji aktualizace klientum pri synchronizaci,
        stejne jako se ji vyuziva pri generovani animovaneho prehledu udalosti
        z minuleho kola. 

        Udalosti a akce, ktere zaznamenali jiz vsichni klienti jsou z konce
        fronty opet odstranovany, takze jeji velikost by se mela drzet na
        stabilni urovni.
    */
    class THistory : public std::map<ACTION_ID, TActionContainer *>
    {
    private:
        TPlayerIdContainer _players;
        int _maxId;
    public:
        /// Konstruktor
        THistory();
        
        /// Konstruktor udavajici pocet hracu
        THistory(TPlayerContainer & players);
        
        /// Inicializace nastavujici pocet hracu
        void init(TPlayerContainer & players);

        /// vyprazdneni zasobniku historie
        void reset();
        
        /// Pridani akce do historie
        ACTION_ID add(TAction * action);

        /// Pridani akce do historie.
        ACTION_ID add(ACTION_TYPE type, void * data);
        
        /// Ziskani akce o danem ID
        TActionContainer * get(ACTION_ID action_id);

        /// Zjisteni maximalniho dosud pouziteho ID
        int maxId();
    };

    /// trida pro vkladani zaznamu do seznamu udalosti
    typedef std::pair<ACTION_ID, TActionContainer *> THistoryPair;

    /// iterator pres seznam udalosti
    typedef std::map<ACTION_ID, TActionContainer *>::iterator THistoryIterator;

    /// Seznam hernich udalosti pro herni klienty
    class TClientHistory : public std::map<ACTION_ID, TAction *>
    {
    private:
        int _maxId;
    public:
        /// Konstruktor
        TClientHistory();

        /// Destruktor
        ~TClientHistory();
        
        /// Inicializace
        void init();        

        /// vyprazdneni seznamu
        void reset();

        /// Pridani akce do historie
        ACTION_ID add(ACTION_ID action_id, TAction * action);

        /// kontrola existence akce s danym id
        bool synchronized(ACTION_ID action_id);
    };
}

#endif
/*****************************************************************************/
