/** @file /world/plan/plan.h
	Reprezentace hern mapy

    Zakladni spolecne datove struktury pro herni mapu a jeji hexy a jejich
    objektove nastavby obsahujici metody pro prepocet souradnic apod.

    @author Petr Wolf
*/

#ifndef WORLD_PLAN_PLAN_H
#define WORLD_PLAN_PLAN_H

/// Maximalni rozmer herni mapy
#define MAX_MAP_SIZE 100

/// pocet typu terenu 
#define TERRAIN_TYPES_COUNT 9

/// pocet vyskovych typu
#define ELEVATION_TYPES_COUNT 5

/// sirka "nedostupneho" okraje mapy, pridavaneho kvuli hezcimu zobrazeni
#define UNACCESSIBLE_BORDER_WIDTH 3

/// vyska "nedostupneho" okraje mapy, pridavaneho kvuli hezcimu zobrazeni
#define UNACCESSIBLE_BORDER_HEIGHT 9

/// Maximalni velikost mesta
#define MAX_CITY_SIZE 7      

// dopredne deklarace
namespace World
{
    struct TERRAIN;
};

#include "world/common.h"
#include "common/TCL/tcl_script.h"
#include "common/mutex.h"
#include "common/rm/rm.h"
#include "world/typedefs.h"
#include "world/struct_holder.h"
#include "world/tcl_xml.h"

namespace World
{
    // FWD
    class TWorldEngine;

    /** Typ ternu - povrch */
    // poradi polozek nemenit, zalezi na tom spravne fungovani tlacitek v editoru !!!
    enum TERRAIN_TYPE 
    {
        TT_PLAIN = 1   ///< plan
       ,TT_FOREST      ///< les
       ,TT_DEEP_FOREST ///< hvozd
       ,TT_ROCKS       ///< skaly
       ,TT_SWAMP       ///< bazina
       ,TT_RIVER       ///< reka
       ,TT_SEA         ///< more
       ,TT_SHALLOW     ///< brod
       ,TT_TOWN        ///< mesto
    };
  
    /** Vyskovy typ */
    enum ELEVATION_TYPE
    {
        ET_LEVEL0 = 0 ///< stupen 0 (morska hladina)
       ,ET_LEVEL1     ///< stupen 1
       ,ET_LEVEL2     ///< stupen 2
       ,ET_LEVEL3     ///< stupen 3 
       ,ET_LEVEL4     ///< stupen 4 (nad 1000 m.n.m.)
    };

    /** Teren
        Vlastnosti spolecne pro vsechny hexy daneho terenu. Jeden takovyto
		zaznam musi existovat pro kazdy typ \c TERRAIN_TYPE
    */
    struct TERRAIN
    {      
        /** Snizeni pravdepodobnosti zasahu strojem.
            Pri vypoctu zasahu jednotky strojem se normalne ziskana pravdepodobnost
            nasobi koeficientem (1 - hit_probability_reduction)
        */
        float hit_probability_reduction; 
        
        /** Omezeni viditelnosti
            Je-li tato hodnota nastavena na \c 1, nen pes hex s danm ternem
            vidt
        */
        int visibility_reduction;
       
        /** Leceni jednotek
            Obsahuje koeficient, ktery pripadne snizuje pocet vylecenych muzu, bylo-li
            leceni provadeno na poli daneho terenu (9.2.1).
        */
        float healing_reduction;
    };

	/** Hex na mape 
		Reprezentace jednoho herniho policka. Obsahuje jak staticke informace (moznost 
		nakupu jednotek, teren, vyska), tak i dynamicke informace (budova na hexu, jednotka
		na hexu)
	*/
    struct HEX 
    {
        /** Vyskyt jednotek na hexu.
            Takove typy jednotek je na danem hexu mozno rekrutovat,
            postavi-li tam hrac patricnou budovu (zbrojnici, vez..)
            
            Jedna se o pole ID typu jednotek (UNIT_TYPE_ID).
            
            Krome toho, existuji jednotky s implicitnim vyskytem vsude (UNIT.default_occurrence).

            @see World::BUILDING.unit_recruitment
        */
		K8_ARRAY<UNIT_TYPE_ID> unit_occurrence;  

		/// teren hexu
		TERRAIN_TYPE terrain;

		/// vyskovy typ hexu
		ELEVATION_TYPE elevation;

		/// jednotka na poli 
		UNIT_ID unit;

		/// budova na poli 
		BUILDING_ID building;

        /// kralovstvi, do ktereho hex nalezi
        KINGDOM_ID kingdom;
    };

    /// Implicitni inicializace (vynulovani hodnot) struktury HEX
    void HEX_Init(World::HEX * hex);

	/** Kopie hodnot struktury HEX z jineho HEXu 
		@param hex cilovy hex, ktery ma byt naplnen
		@param src zdrojovy hex
	*/
	bool HEX_Copy(HEX * hex, const HEX * src);

	/** Herni mapa
		Celkova "mrtva" reprezentace herni mapy jako pole hexu.

		Ve hre ve skutecnosti neexistuje, pouziva se jako nosic dat pri 
		nacitani ci ukladani hry.
	*/
	struct MAP 
	{
		/** Sirka mapy (pocet hexu) 
		    Vcetne "nedostupneho okraje"
		*/ 
		int width;

		/** Vyska mapy (pocet hexu)
		    Vcetne "nedostupneho okraje"
		*/ 
		int height;
		
		/** Policka mapy
		    Pole usporadane podle logiky cislovani hexu, obsahujici i "nedostupny"
		    okraj
		*/ 
		HEX ** hexes;
	};

	/** Viditelnost hexu 
		Viditelnost z pohledu hrace 
	*/
	enum VISIBILITY
	{
	    /// nikdy nevideno
		VI_NEVER_SEEN

		/// jiz nekdy videno, ale mimo aktualni dohled
	   ,VI_SEEN

	    /// v aktualnim dohledu
	   ,VI_IN_SIGHT
	};

	/** Historicky zaznam hexu
		Reprezentace hexu v takove podobe, v jake jej hrac naposledy videl.
		Struktura neuchovava vsechny udaje hexu.
	*/
	struct HISTORY_HEX
	{
        /// cislo kola, ve kterem jej hrac naposedy videl
        ROUND_ID turn_id;

		/// budova, ktera na poli stala (pouze typ)
		BUILDING_TYPE_ID building; 
	};

    /// Struktura predstavujici instanci kralovstvi - prevzato z RM
    struct KINGDOM
    {
        int id;      ///< Index kralovstvi
        int roleid;  ///< Role vlastnici kralovstvi
        int centerx; ///< X-ova souradnice stredu kralovstvi
        int centery; ///< Y-ova souradnice stredu kralovstvi
    };

    /// Implicitni inicializace (vynulovani hodnot) struktury KINGDOM
    void KINGDOM_Init(World::KINGDOM * kingdom);

    /// Struktura predstavujici instanci mesta - prevzato z RM
    struct TOWN
    {
        char name[K8_MAX_NAME_LEN];  ///< Jmeno mesta
        int owner;                   ///< Vlastnik mesta
        int citysize;                ///< Velikost mesta
        int position[MAX_CITY_SIZE]; ///< Umisteni jednotlivych hexu mesta
    };

    /// Implicitni inicializace (vynulovani hodnot) struktury TOWN
    void TOWN_Init(World::TOWN * town);

    /** Teren.
        Objektova nadstavba struktury TERRAIN reprezentujici vlastnosti
        terenu.
    */
    class TTerrain : public TStructHolder<TERRAIN>, public TTCL_Struct
    {
    public:
        TTerrain(TERRAIN * data, int id = 0);

        void readFromTCL(TTCL_Interpreter & interpreter, const char * varName);

        void writeToTCL(TTCL_Interpreter & interpreter, const char * varName);
    };

    /** Kralovstvi.
        Objektova nadstavba struktury KINGDOM, resp. TKingdomInfo, reprezentujici
        kralovstvi jako celek.
    */
    class TKingdom : public TStructHolder<KINGDOM>, public TTCL_XML_Struct
    {
    private:
        friend class TWorldEngine;
    public:
        TKingdom(int id);

        TKingdom(KINGDOM * data, int id = 0);

        void readFromXML(TPackage * package);

        void writeToXML(TPackage * package);

        void readFromTCL(TTCL_Interpreter & interpreter, const char * varName);

        void writeToTCL(TTCL_Interpreter & interpreter, const char * varName);
    };

    /** Mesto.
        Objektova nadstavba struktury TOWN, resp. TCityInfo, reprezentujici
        mesto jako celek.
    */
    class TTown : public TStructHolder<TOWN>, public TTCL_XML_Struct
    {
    private:
        friend class TWorldEngine;
    public:
        TTown(int id);

        TTown(TOWN * data, int id = 0);

        void readFromXML(TPackage * package);

        void writeToXML(TPackage * package);

        void readFromTCL(TTCL_Interpreter & interpreter, const char * varName);

        void writeToTCL(TTCL_Interpreter & interpreter, const char * varName);
    };

    /** Hex */
    class THex : public TStructHolder<HEX>, public TTCL_XML_Struct, public TMutex
    {
    private:
        int _town;

        friend class TWorld;
        friend class TWorldEngine;
        friend class TMap;
    public:
        THex(HEX * data, int id = 0);

        ~THex();

        void readFromXML(TPackage * package);

        void writeToXML(TPackage * package);

        void readFromTCL(TTCL_Interpreter & interpreter, const char * varName);

        void writeToTCL(TTCL_Interpreter & interpreter, const char * varName);

        /** Skutecny seznam typu jednotek, ktere lze na danem policku kupovat.
            Seznam kombinuje vyskyt jednotek dany ve strukture HEX s implicitnim
            vyskytem ze struktury UNIT. Pro koupi jednotky je vsak nutne postavit
            na hexu odpovidajici budovu.
            
            @see TWorld::fillRelations()
            @see TLivingBuilding::fillActions()
            */
        TIntContainer unit_occurrence;

        UNIT_ID unit_id();

        BUILDING_ID building_id();

        int town();
    };

    /// seznam sourednic
    typedef std::vector<COORDS> TCoordsContainer;

    /// iterator pres seznam souradnic
    typedef std::vector<COORDS>::iterator TCoordsIterator;

	/** Hern mapa */
    class TMap : public TStructHolder<MAP>, public TXML_Struct, public TMutex
	{
    private:
        bool _hexes_self_allocated;
        THex *** _thexes;
        friend class TWorldEngine;
        friend class TVisibilityMap;
	public:
        /** Implicitni konstruktor */
        TMap();

        /** Pretizeni konstruktoru TStructHolder */
        TMap(MAP * data, int id = 0);

        /// Destruktor
        ~TMap();

        void reset();

        /// Nacteni ze struktury \c MAP
        void load(MAP * data);

        void reload(HEX ** hexes);

        HEX ** save(bool dealloc = true);

        /** Kontrola platnych souradnic */
        bool validCoords(const int x, const int y);
        
        /** Kontrola platneho ID hexu */
        bool validHexId(const int id);

        /** Kontrola, je-li hex v hratelne casti mapy (mimo okraje) 
            @retval false hex je mimo hratelnou cast mapy
            @retval true hex je v hratelne casti
        */
        bool insideBorders(const int id);

        /** prevod ID hexu na souradnice 
            @retval true pri uspesnem prevodu
            @retval false je-li id neplatne
        */ 
	    bool idToCoords(const int id, int * x, int * y);
    	
	    /** prevod souradnic hexu na ID 
            @retval true pri uspesnem prevodu
            @retval false jsou-li souradnice neplatne	
	    */
	    bool coordsToId(const int x, const int y, int * id);

		/** Vraci hex na danych souradnicich */
		THex * getHexAt(const int x, const int y);

		/** Vraci hex podle ID */
		THex * getHexById(const int id);

        /** Vraci sousedni hex na konkretni orientaci
            @param hex_id startovni hex
            @param orientation smer orientace
            @return ID sousedniho hexu v danem smeru, nebo -1 pri chybnych datech
        */
        TIntContainer getNeighboursOrientation(const int hex_id, const int orientation);

        /// Vraci seznam sousedu hexu 
        TIntContainer * getNeighbours(const int x, const int y);
        
        /// Vraci seznam sousedu hexu 
        TIntContainer * getNeighbours(const int id);

        /** Vypocet seznamu hexu v kruhovem okoli zvoleneho hexu.
            @param  center stren kruhoveho okoli
            @param  perimeter  polomer
            @return seznam hexu do vzdalenosti \c perimeter od hexu \c center
        */         
        TIntContainer getCircle(HEX_ID center, int perimeter);

        /** Sirka mapy */
        int width();

        /** Vyska mapy */
        int height();

        /** Maximalni ID hexu obsazene v mape */
        int maxId();

        /// nacteni z XML
        void readFromXML(TPackage * package);

        /// ulozeni do XML
        void writeToXML(TPackage * package);
	};

    /** Historicky hex */
    class THistoryHex : public TStructHolder<HISTORY_HEX>, public TMutex
    {
    private:
        friend class TVisibilityHex;
    public:
        THistoryHex(HISTORY_HEX * data, int id = 0);
    };

    /** Policko viditelnosti.
        Dilek mapy viditelnosti nejakeho hrace, ktery rozlisuje,
        je-li odpovidajici hex na mape videt v aktualnim vyhledu,
        nebo si jej hrac pamatuje z drivejska anebo jej
        jeste nevidel.
        */
    class TVisibilityHex : public TXML_Struct, public TMutex
    {
    private:
        World::TWorld * _world;

        /// viditelnost hexu
		VISIBILITY _visibility;
        
        /// ukazatel na zaznam historie hexu 
        THistoryHex * _history;

        /// ukazatel na odpovidajici hex z mapy TMap
		THex * _actual;

        /// seznam budov, ktere na dany hex vidi
        TIntContainer _buildings;

        /// seznam jednotek, ktere na dany hex vidi
        TIntContainer _units;

        /// seznam mest, ktere na dany hex vidi
        TIntContainer _towns;

        friend class World::TWorld;   

        friend class World::TWorldEngine;
    public:
        /** Konstruktor.
            @param  actual  ukazatel na odpovidajici hex ze skutecne mapy
            */
        TVisibilityHex(THex * actual);

        /// Destruktor
        ~TVisibilityHex();

        /** Presun hexu do oblasti mimo aktualni vyhled.
            @param  history  ukazatel na zaznam o hexu, ktery si bude hrac pamatovat
            */
        void setHistory(HISTORY_HEX * history);

        /** Presun hexu do aktualniho vyhledu.
            @param  unit_id      ID jednotky, ktera na hex vidi (nebo 0, neni-li to relevantni)
            @param  building_id  ID budovy, ktera na hex vidi (nebo 0, neni-li to relevantni)
            @param  town_id      ID mesta, ktere na hex vidi (nebo 0, neni-li to relevantni)

            @retval  true   teprve touto zmenou se hex dostava do aktualniho vyhledu
            @retval  false  na hex jiz predtim bylo videt, zadna zmena se nekona
            */
        bool setActual(int unit_id = 0, int building_id = 0, int town_id = 0);

        /** Odstraneni jednotky ci budovy ze seznamu entit, ktere na hex vidi.
            Odchazi-li jednotka s dohledu hexu, nebo je-li budova znicena apod. je treba
            odstranit ID jednotky, resp. budovy, ze seznamu.
            @param  unit_id      ID jednotky, ktera ztraci na hex dohled (nebo 0, neni-li to relevantni)
            @param  building_id  ID budovy, ktera ztraci na hex dohled (nebo 0, neni-li to relevantni)
            @param  town_id      ID mesta, ktere ztraci na hex dohled (nebo 0, neni-li to relevantni)

            @retval  true   Jednotka, resp. budova, kterou jsem odstranil, byla posledni entitou
                            ktera na hex videla -> hex se dostava do historie.
            @retval  false  I po odstraneni jednotky ci budovy je stale nekdo, kdo na nej
                            vidi, tedy zadna zmena viditelnosti se nedeje.
            */
        bool removeActual(int unit_id = 0, int building_id = 0, int town_id = 0);

        /// Ziskani viditelnosti hexu 
        VISIBILITY getVisibility();

        /// Ziskani reference na odpovidajici hex z mapy, je-li hex v aktualnim vyhledu
        THex & getActual();

        /// Ziskani reference na zaznam historie hexu, byl-li jiz hex drive viden a nyni je 
        /// mimo aktualni vyhled.
        THistoryHex & getHistory();

        /// nacteni z XML
        void readFromXML(TPackage * package);

        /// zapis do XML
        void writeToXML(TPackage * package);
    };

    /** Mapa viditelnosti */
    class TVisibilityMap : TXML_Struct, public TMutex
    {
    private:
        TVisibilityHex *** _hexes;

        TMap & _map;

        friend class TVisibilityMapsContainer;
    public:
        TVisibilityMap(TMap & whole_map);

        ~TVisibilityMap();  

        void reset();

        /** Vraci hex na danych souradnicich */
		TVisibilityHex & getHexAt(const int x, const int y);

		/** Vraci hex podle ID */
		TVisibilityHex & getHexById(const int id);       

        void writeToXML(TPackage * package);

        void readFromXML(TPackage * package);
    };

    /// seznam historickych map
    class TVisibilityMapsContainer : public std::map<int, TVisibilityMap *>, public TXML_Struct, public TMutex
    {
    public:
        void load(HISTORY_HEX *** hm, TMap & map);

        HISTORY_HEX *** save(bool dealloc = true);

        void reset();

        void writeToXML(TPackage * package);

        void readFromXML(TPackage * package);
    };

    /// iterator pres seznam historickych map
    typedef TVisibilityMapsContainer::iterator TVisibilityMapsIterator;

    // testovaci mapa
    TMap * _test_map();
};

#endif
