/** @file /world/players/player.h
	
	Reprezentace hrace, predavani tahu

	@author Petr Wolf
*/

#ifndef WORLD_TYPEDEF_PLAYER_H
#define WORLD_TYPEDEF_PLAYER_H

// dopredne deklarace 
namespace rm
{
  struct TRoleInfo;
  struct TAIPlan;
};

#include <map>
#include "common/Msg.h"
#include "common/mutex.h"
#include "world/da_holder.h"
#include "world/typedefs.h"
#include "world/struct_holder.h"
#include "world/tcl_xml.h"
#include "ai/Diplomacy/diptypes.h"

typedef int K8_COLOR; 

namespace World 
{
  /// typ ukonceni hry
  enum ENDGAME_REASON
  {
    /// dobrovolne zruseni hry
    ER_CANCEL = 1

    /// vynucene ukonceni hry (vypnuti serveru)
    ,ER_SHUTDOWN

    /// remiza. Hra skoncila, ale neni mozne urcit viteze
    ,ER_DRAW
    
    /// vitezstvi zabitim vsech nepratelskych jednotek
    ,ER_VICTORY_DESTROY

    /// vitezstvi dobytim vetsiny mest
    ,ER_VICTORY_CONQUER

    /// vitezstvi dobytim nejvice mest do casoveho limitu
    ,ER_VICTORY_TIMEOUT
  };

  /// typ hrace (clovek x pocitac)
  enum PLAYER_TYPE
  {
    /// lidsky hrac
    PT_HUMAN

    /// umela inteligence - zacatecnik
    ,PT_AI_1

    /// umela inteligence - pokrocily
    ,PT_AI_2

    /// umela inteligence - expert
    ,PT_AI_3
  };

  /// commands the player can give
  enum PLAYER_ALLOWED_COMMANDS
  {
    /// Player can do anything
    PAC_ALL,
    
    /// Player cannot move
    PAC_NOT_MOVE
  };

  /// Stav hrace
  enum PLAYER_STATE
  {
    /// aktivni hrac
    PS_ACTIVE

    /// porazeny hrac
    ,PS_DEFEATED
  };

  /// penize - pocitano v celych jednotkach
  typedef int MONEY;

  /// herni data hrace
	struct PLAYER_GAME_DATA
	{
    /// jmeno role hrace (napriklad Vevoda z Nottinghamu)
    char name[MAX_STRLEN];

    /// stav hrace
    World::PLAYER_STATE state;

    /// penize
    World::MONEY money;

    /// vztahy k ostatnim hracum
    ai_ns::diplomacy_ns::TDipRelation *(relations[MAX_PLAYERS + 1]);
	};

  /// Charakteristika pripojeni hrace
  struct PLAYER_CLIENT_DATA
  {
    /// jmeno hrace (napriklad Karel)
    char name[MAX_STRLEN];

    /// in case of client reconnection this value will be set to his clientId, otherwise will be set to 0 - information, that there is client waiting for entering the game
    int newclient_id;

    /// id sitoveho klienta, ktery se odpojil, nebo 0, hraje-li lokalne
    int oldclient_id;
    
    /// id sitoveho klienta, je-li hrac pripojen po siti, nebo 0, hraje-li lokalne
    int client_id;

    /// typ hrace (clovek/AI)
    PLAYER_TYPE type;

    /// commands the player can give (all/cannot move)
    PLAYER_ALLOWED_COMMANDS acommands;

    /// pripojen?
    bool connected;

    /// ma nahranou mapu?
    bool guimap_ready;
  };

  /// Statistiky hrace
  struct PLAYER_STATISTICS
  {
    /// pocet zabitych jednotek
    int units_killed;

    /// pocet ztracenych jednotek
    int units_lost;

    /// pocet vytvorenych jednotek
    int units_created;

    /// pocet postavenych budov
    int buildings_built;

    /// pocet obsazenych budov
    int buildings_captured;

    /// pocet ztracenych budov
    int buildings_lost;

    /// pocet znicenych budov
    int buildings_destroyed;

    /// pocet obsazenych mest
    int towns_captured;

    /// pocet ztracenych mest
    int towns_lost;

    /// pocet obsazenych kralovstvi
    int kingdoms_captured;

    /// pocet ztracenych kralovstvi
    int kingdoms_lost;

    /// pocet ziskanych penez
    int money_gained;
  };

  /// Reprezentace hrace
  struct PLAYER
  {
    /// herni data
    PLAYER_GAME_DATA game_data;

    /// informace o pripojeni
    PLAYER_CLIENT_DATA client_data;

    /// statistiky herni uspesnosti
    PLAYER_STATISTICS statistics;
  };

  /// Objektova reprezentace hrace
  class TPlayer: public TStructHolder<PLAYER>, public TTCL_XML_Struct, public TMutex
  {
  private:
    /// byla data alokovana az behem hry?
    bool _relations_self_allocated;

    friend class TWorldServer;
  public:
    /// Konstruktor
    TPlayer();

    /// Konstruktor prejimajici data
    TPlayer(PLAYER * data, int id = 0);

    /// Konstruktor prejimajici id
    TPlayer(int id);

    /// Destruktor
    ~TPlayer();

    /// Zapis do XML
    void writeToXML(TPackage * package);

    /// Nacteni z XML
    void readFromXML(TPackage * package);
    
    /// Nacteni z TCL
    void readFromTCL(TTCL_Interpreter & interpreter, const char * varName);
    
    /// Zapis do TCL
    void writeToTCL(TTCL_Interpreter & interpreter, const char * varName);
  };

  /// Seznam hracu
  class TPlayerContainer : public TDAHolder<PLAYER, PLAYER_ID, TPlayer>, public TMutex
  {
  private:
    /// ID hrace na tahu
    PLAYER_ID player_on_turn;

    TPlayer * _dummy;
  public:
    /// Data umele inteligence
    rm::TAIPlan * ai_data;

    /// Konstruktor
    TPlayerContainer();

    /// Destruktor
    ~TPlayerContainer();

    /// Pristup k jednotlivym hracum
    TPlayer * & operator[](const PLAYER_ID _Key); 

    /// Vyprazdneni seznamu
    void reset();

    /// Ulozeni do pole TRoleInfo
    void save(rm::TRoleInfo ** info);

    /// Hrac na tahu
    TPlayer * playerOnTurn();

    /// Cislo hrace na tahu
    PLAYER_ID playerIdOnTurn();

    /// nastaveni hrace na tahu
    PLAYER_ID setPlayerOnTurn(PLAYER_ID player_id);

    /// Cislo nejblizsiho lidskeho hrace (0, neni-li zadny takovy)
    PLAYER_ID nearestHumanPlayerId();

    /** Pocet lidskych hracu ve hre
      @param ignore cislo hrace, ktery ma byt pri scitani ignorovan,
              nebo 0, maji-li se zapocitat vsichni
    */
    int numberOfHumanPlayers(PLAYER_ID ignore = 0);

    /** Posun tahu na dalsiho hrace v rade.
      Posune interni ukazatel na dalsiho hrace na rade.
      @retval 0 Pokud nedoslo k posunu do noveho kola (nejednalo se o posledniho
           hrace v rade).
      @retval 1 Pokud doslo k posunu do noveho kola (jednalo se o posledniho
           hrace v rade).      
    */
    int next();   

    // Zaznamy statistik

    /** STATISTIKA - Zabiti jednotky */
    void unitKilled(PLAYER_ID owner, PLAYER_ID killer = 0);
    
    /** STATISTIKA - Vytvoreni jednotky */
    void unitCreated(PLAYER_ID owner);

    /** STATISTIKA - Postaveni budovy */
    void buildingBuilt(PLAYER_ID owner);

    /** STATISTIKA - Oobsazeni budovy */
    void buildingCaptured(PLAYER_ID owner, PLAYER_ID previous_owner);

    /** STATISTIKA - Zniceni budovy */
    void buildingDestroyed(PLAYER_ID owner, PLAYER_ID destroyer = 0);

    /** STATISTIKA - Obsazeni mesta */
    void townCaptured(PLAYER_ID owner, PLAYER_ID previous_owner);

    /** STATISTIKA - Obsazeni kralovstvi */
    void kingdomCaptured(PLAYER_ID owner, PLAYER_ID previous_owner);

    /** STATISTIKA - Zisk penez */
    void moneyGained(PLAYER_ID owner, MONEY amount);
  };
  
  /// iterator pres seznam hracu
  typedef std::map<PLAYER_ID, TPlayer *>::iterator TPlayerIterator;

  /// struktura pro vkladani novych hracu
  typedef std::pair<PLAYER_ID, TPlayer *> TPlayerPair;
}

#endif
/*****************************************************************************/
