/** @file /world/sa_holder.h

    Obecna sablona pro kontejner dat pracujici nad tridou std::map,
    zvladajici nacitani a ukladani i do staticky alokovanych
    poli pointeru.

    @author Petr Wolf
*/

#ifndef WORLD_SA_HOLDER_H
#define WORLD_SA_HOLDER_H

#include <map>
#include "common/xml/package.h"
#include "common/xml/packsmall.h"
#include "common/TCL/tcl_script.h"
#include "common/mm.h"
#include "common/mutex.h"
#include "world/struct_holder.h"

namespace World
{
    /** Obecna sablona pro kontejner dat pracujici nad tridou std::map,
        zvladajici nacitani a ukladani i do staticky alokovanych
        poli pointeru.
    */
    template<class R, class S, class T>
    class TSAHolder : public std::map<S, T *> 
    {
    public:
        /** Nacteni ze statickeho pole pointeru */
        int load(R ** data, const int length);

        /** Vyprazdneni pole */
        void reset();

        /** Export do statickeho pole */
        R ** save(const int length, bool dealloc = true);

        /** Zapis dat do XML.
            @param package  XML objekt, do ktereho se zapisuje
            @param itemTag  nazev tagu pro prvky seznamu
        */
        void readFromXML(TPackage * package, char * itemTag);
        
        /** Nacteni dat z XML.
            @param package  XML objekt, do ktereho se zapisuje
            @param itemTag  nazev tagu pro prvky seznamu
        */
        void writeToXML(TPackage * package, char * itemTag);

        /** Zapis dat do prostredi TCL interpretu.
            @param interpreter  interpret jazyka TCL
            @param name         nazev promenne odpovidajici seznamu
        */
        void writeToTCL(TTCL_Interpreter & interpreter, const char * name);
    };
    
    template<class R, class S, class T>
    int TSAHolder<R, S, T>::load(R ** data, const int length)
    {
        for (int i = 0; i < length; i++) {
            if (data[i])
                (*this)[i] = new T(data[i], i);
        }

        return 0;
    } 

    template<class R, class S, class T>
    void TSAHolder<R, S, T>::reset()
    {
        typename TSAHolder<R,S,T>::iterator it;
        for (it = this->begin(); it != this->end(); it++) 
            delete it->second;

        this->clear();
    }

    template<class R, class S, class T>
    R ** TSAHolder<R, S, T>::save(const int length, bool dealloc)
    {
        R ** result = (R **)KMemAlloc(length * sizeof(R *));
        
        for (int i = 0; i < length; i++)
            result[i] = NULL;

        typename TSAHolder<R,S,T>::iterator it;
        for (it = this->begin(); it != this->end(); it++) { 
            result[it->first] = &it->second->data();
            
            if (dealloc) {
                delete it->second;
                it->second = NULL;
            }
        }

        return result;
    }

    template<class R, class S, class T>
    void TSAHolder<R, S, T>::readFromXML(TPackage * package, char * itemTag)
    {     
        // nacteni prvku
        TXMLdata * xml_data = package->getXML();
        DA<int> ids;
        xml_data->getListOfIds(&ids, NULL, 0, NULL);

        for (int i = ids.getNext(-1); i != -1; i = ids.getNext(i)) {
            if (ids[i] == 1) {       
                (*this)[i] = new T(i);
                (*this)[i]->readFromXML(new TPackSmall(package->getSubtag(itemTag, i, NULL)));
            }
        }
    }

    template<class R, class S, class T>
    void TSAHolder<R, S, T>::writeToXML(TPackage * package, char * itemTag)
    {     
      // ulozeni prvku
        TPackSmall * xml_item;
        typename TSAHolder<R,S,T>::iterator i;

        for (i = this->begin(); i != this->end(); i++) {
            package->setEmpty(itemTag, i->first, NULL);
            xml_item = new TPackSmall(package->getSubtag(itemTag, i->first, NULL));            
            i->second->writeToXML(xml_item);
            delete xml_item;
        }
    }

    template<class R, class S, class T>
    void TSAHolder<R, S, T>::writeToTCL(TTCL_Interpreter & interpreter, const char * name)
    {
        typename TSAHolder<R, S, T>::iterator i;
        char itemName[MAX_STRLEN];
        
        for (i = this->begin(); i != this->end(); i++) {
            snprintf(itemName, MAX_STRLEN, "%s(%d)", name, i->first);
            i->second->writeToTCL(interpreter, itemName);
        }
    }    
    
    /** Vlaknove bezpecna varianta tridy TSAHolder.
        Funkcne ekvivalentni tride TSAHolder, dedi ale navic po tride TMutex umoznujici
        zamykani a odemykani.
    */
    template<class R, class S, class T>
    class TSAHolder_Mutex : public TSAHolder<R, S, T>, public TMutex
    {

    };
};

#endif 
/*****************************************************************************/
