/** @file /world/units/unit.h
    Zakladni spolecne datove struktury pro vlastnosti jednotek vyskytujc se ve hre.

    Tyto struktury slouzu pro nacitani a ukladani dat z a do XML souboru a pro
    urceni spolecneho rozhrani mezi moduly World, GUI a AI.

    @author PW
*/

#ifndef WORLD_TYPEDEFS_UNIT_H
#define WORLD_TYPEDEFS_UNIT_H

#include <map>
#include "common/mutex.h"
#include "common/TCL/tcl_script.h"
#include "common/da.h"
#include "common/mm.h"
#include "world/common.h"
#include "world/weather.h"
#include "world/plan/plan.h"
#include "world/struct_holder.h"
#include "world/tcl_xml.h"

/// minimalni pouzite ID bonusu
#define BONUS_MIN_ID 1 

/// maximalni pouzite ID bonusu
#define BONUS_MAX_ID 15

/// pocet stavu pocasi
#define WEATHER_STATES_COUNT 3

namespace World
{
    // dopredna deklarace
    class TLivingBuilding;
    class THex;

    /** Rozliseni jednotek podle typu */
    enum UNIT_TYPE
    {
        UT_INFANTRY   ///< pechota
       ,UT_CAVALRY    ///< jizda
       ,UT_VEHICLE    ///< stroj
       ,UT_INDIVIDUAL ///< jednotlivec (hrdina, kouzelnik)
       ,UT_BEAST      ///< netvor (drak, obr)
       ,UT_FLYING     ///< letajici jednotka
    };
    
    /** Moznosti dostupnosti terenu pro jednotky */
    enum TERRAIN_ACCESSIBILITY
    {
        /** Neomezen vstup 
            Na pole takovehoto typu muze jednotka bez omezeni vstupovat, utocit atd. 
        */
        TA_NORMAL

        /** Vstup zakazan 
            Na pole takoveho typu jednotka nemuze vstoupit ani utocit zteci.
        */
       ,TA_CANNOT       
    };
    
    /** Kombinace pohybu a boje */
    enum MOVEMENT_AND_FIGHT_COMBINATION
    {
        MFC_ANYTHING       ///< libovolna kombinace pohybu a boje
       ,MFC_MOVEMENT_FIRST ///< jednotka se muze pohnout a pote bojovat
       ,MFC_ATTACK_FIRST   ///< jednotka muze utocit pouze pred pohybem, pak uz ne
       ,MFC_SINGLE         ///< jednotka muze pouze zautocit nebo se pouze pohnout 
    };

    /** Styl boje
        Rozliseni typu boje, kterym jednotka utoci.
    */
    enum FIGHT_TYPE
    {
        FT_NONE         ///< jednotka nedokaze bojovat
       ,FT_FACE_TO_FACE ///< souboj tvari v tvar, klasicka ztec
       ,FT_BLAST        ///< utok vybusninami, pri nemz sama utocici jednotka umira
       ,FT_FIRE         ///< strelba - balisticky oblouk (sipy)
       ,FT_CATAPULT     ///< vrhani kamenu (katapult)
       ,FT_MAGIC        ///< utok pomoci kouzel
    };

    /** Vlastnost jednotky, kterou lze zlepsit bonusem */
    enum UNIT_PROPERTY
    {
        UP_ATTACK       ///< navyseni utocneho cisla
       ,UP_DEFENSE      ///< navyseni obranneho cisla
       ,UP_LIVES        ///< navyseni maximalniho poctu zivotu
       ,UP_MOVEMENT     ///< navyseni maximalniho poctu bodu pohybu
       ,UP_ATTACK_RANGE ///< navyseni dostrelu
    };

    /** Pocet hodnot v UNIT_PROPERTY */
    #define UNIT_PROPERTY_COUNT 5

    /** Maximalni pocet ruznych bonusu, ktere jednotky mohou vlastnit */
    #define UNIT_PROPERTY_IMPROVE_MAX_COUNT 3
    
    /** Bonusova vlastnost 
        Navyseni nejake vlastnosti jednotky. Tyto bonusy jednotka nakupuje,
        nebo ziskava prechodem na vyssi uroven.
    */
    struct UNIT_PROPERTY_BONUS
    {
        /** ID bonusu */
        int id;

        /** ID nazvu bonusu */
        int nameid;
        
        /** ID obrazku */
        int imageid;
    
        /** vlastnost, ktera je povysovana */ 
        UNIT_PROPERTY property;

        /** financni naklady na porizeni bonusu */
        int cost;
        
        /** hodnota, o kterou je vlastnost navysena
            (pripadne ID kouzla, ktere se jednotka naucila)
        */
        int bonus;          
    };

    /// Maximalni uroven jednotky
    #define MAX_UNIT_LEVEL 7

    /// Pocet bodu zkusenosti nutnych na postup do dalsiho levelu
    #define EXPERIENCE_LEVEL 1000

    /** Jednotka
        Herni figurka, postava, jednotka, zkratka zakladni zivouci objekt ve hre.
    */
    struct UNIT 
    {   
        /** Vojtovo ID */
        int id;

        /** ID nazvu jednotky */
        int nameid;

        /** ID modelu */
        int modelid;

        /** ID zvuku */
        int soundsid[MAX_ACTION_SOUNDS];
        
        /** ID ikony */
        int symbolid;



    /* ZAKLADNI CHARAKTERISTIKA */
                
        /** Typ jednotky */
        UNIT_TYPE type;



    /* POHYB */
    
        /** pocet (maximalni) bodu pohybu */
        int max_points_of_movement;

        /** Koeficienty pro pohyb v nadmorske vysce */
		float movement_elevation[ELEVATION_TYPES_COUNT]; //# vyskovych typu

        /** Koeficienty pro pohyb v terenu */
		float movement_terrain[TERRAIN_TYPES_COUNT + 1]; //# pocet terenu 

        /** Koeficienty pro pohyb za daneho pocasi */
        float movement_weather[WEATHER_STATES_COUNT]; //# pocet stavu pocasi
        
        /** Koeficient pro pohyb do kopce */
        float movement_uphill;

        /** Koeficient pro pohyb z kopce */
        float movement_downhill;
    
        /** Pocet muzu v rade */
        int men_in_row;

        /** Muze jednotka obsazovat budovy? */
        bool can_enter_building;


    /* BOJ */
    
        /** Utocne cislo */
        int attack;
        
        /** Obranne cislo */
        int defense;

        /** Sila protiutoku 
            Ma hodnotu 0, pokud jednotka pri napadeni vubec nemuze nepriteli zpusobit
            zraneni (napriklad katapult)
            */
        int counterattack;

        /** Kombinace pohybu a boje */
        MOVEMENT_AND_FIGHT_COMBINATION movement_and_fight_combination;

        /** Styl boje */
        FIGHT_TYPE fight_type;

        /** Bonus k utoku za teren na kterem stojim */
		float attack_bonus_from_terrain[TERRAIN_TYPES_COUNT + 1]; //# terenu 

        /** Bonus k obrane za teren na kterem stojim */
		float defense_bonus_from_terrain[TERRAIN_TYPES_COUNT + 1]; //# terenu

        /** Bonus k utoku za teren, na kterem stoji protivnik */
		float attack_bonus_against_terrain[TERRAIN_TYPES_COUNT + 1]; //# terenu;
        
        /** Bonusy pro boj s konkretnim typem jednotky 
            Rozlisuje se podle UNIT_TYPE protivnika
        */
        float attack_bonus_against_unit[UNIT_TYPES_COUNT]; // # pocet kategorii jednotek

        /** Bonus k utoku pri napadeni staveb 
            Plati na vsechny budovy, bez rozdilu
        */
        float attack_bonus_against_building; 

        /** Dosah utoku
            0 .. jednotka nemuze utocit
            1 .. utok na blizko (ztec)
            2 - n .. dostrel
        */
        int attack_range;
        
        /** Pravdepodobnost zasahu
            V rozsahu 0..1 uruje pro dobyvaci stroje pravdepodobnost zasahu
            cile.
        */
        float hit_probability;
        
        
        
    /* NAKUP A ZOLD */
    
        /** Implicitni vyskyt jednotky na cele mape 
            Jednotky s takovouto vlastnosti je mozno kupovat kdekoliv na mape.
            Ostatni jednotky jen tam, kde maji prislusne hexy nastaven jejich
            vyskyt.
        */
        int default_occurrence;

        /** Nakupni cena jednotky */
        int cost;

        /** Zold za jedno kolo */
        int pay;

        /** Bonusy.
            Seznam vlastnosti, ktera je jednotka schopna zlepsovat pomoci bonusu bonusu, ktere je jednotka schopna obdrzet
		*/
        int bonuses[UNIT_PROPERTY_COUNT];        
    
        

    /* ZIVOTY A LECENI */
        
        /** Maximalni pocet zivotu/muzu */
        int max_lives;
       
        /** Cena za vyleceneho muze/lehke zraneni */
        int cure_cost;

    
        
    /* VIDENI */

        /** Dosah videni */
        int eyesight_distance;
    
    

    /* STAVBA BUDOV */

        /** Priznak, jestli jednotka umi nebo neumi stavet/opravovat budovy 
            Umi-li, pak dokaze stavet vsechny. Na uroven jednotlivych budov se nerozlisuje.
        */
		int can_build_building;
    };

    /// Seznam hernich akci jednotky
    enum UNIT_ACTION
    {
        /// dobrovolna smrt jednotky
        UA_SUICIDE 

        /// pohyb po mape
       ,UA_MOVE

        /// utok na jednotku
       ,UA_ATTACK

        /// leceni (doplneni zivotu)
       ,UA_HEAL

        /// doplneni muzu (neelitni)
       ,UA_RECRUIT

        /// doplneni muzu (elitni)
       ,UA_RECRUIT_ELITE

        /// zahajeni stavby 
       ,UA_START_BUILDING

        /// preruseni stavby
       ,UA_STOP_BUILDING
        
        /// oprava budovy
       ,UA_REPAIR_BUILDING
    };

    /// pocet akci jednotky
    #define UNIT_ACTION_COUNT 9

    /// Priznaky urcujici, co se muze o jednotce zobrazit
    enum UNIT_KNOWLEDGE {
        /// o jednotce nesmim zobrazit nic
        UK_NONE

        /// jednotka patri nepriteli, zobrazuji jen nektere polozky
       ,UK_ENEMY

        /// jednotka je moje, zobrazuji vsechno
       ,UK_MINE
    };

    /// Stav zivouci jednotky
    enum UNIT_STATE {
        /// volna jednotka
        US_IDLE

        /// jednotka venujici se stavbe/oprave budovy
       ,US_BUILDING

        /// jednotka v tomto kole zajajila stavbu nebo opravu
       ,US_HAS_STARTED_BUILDING

        /// jednotka se v tomto kole jiz pohla
       ,US_HAS_MOVED

        /// jednotka v tomto kole jiz bojovala
       ,US_HAS_FOUGHT

        /// jednotka bojovala a pak se pohla
       ,US_HAS_FOUGHT_AND_MOVED

        /// jednotka, ktera jiz provedla nejakou akci a nemuze dale nic konat
       ,US_EXHAUSTED
    };

    /// Stav panacka v jednotce, resp. stav zraneni jednotlivce
    #define UW_OK 0 ///< bez zraneni
    #define UW_KO 3 ///< na smrtelne posteli - pokud se panacek nevyleci, pristi kolo zemre.

    /** Zivouci jednotka
        Kompletni datovy zaznam zivouci jednotky 
    */
	struct LIVING_UNIT 
	{
        /// typ jednotky 
        UNIT_TYPE_ID type;
        
        /// hrac, ke kteremu jednotka patri 
		PLAYER_ID player;

		/// jmeno jednotky
		K8_CHAR name[MAX_STRLEN];
		
		/// bonusy, ktere jednotka koupila
		K8_ARRAY<UNIT_PROPERTY_BONUS_ID> bonuses;

        /// stav jednotky
        UNIT_STATE state;

		/// Pocet zivotu (resp. zivych muzu) 
		int lives;

		/// Pocitadlo doby zraneni 
		int wounds[MAX_LIVES]; 

		/// Pocet napadeni v tomto kole 
		int assaults_count;

		/// zbyvajici body pohybu 
		int points_of_movement;

		/// uroven jednotky
		int level;
		
		/// zkusenosti jednotky
		int experience;

        /// orientace jednotky
		int orientation;
	};

    /** Typ jednotky
        Objektovy obal struktury UNIT, napojeni na mapu a dalsi redundantni udaje
        */
    class TUnit : public TStructHolder<UNIT>, public TTCL_Struct
    {
    public:
        TUnit();

        TUnit(UNIT * data, int id = 0);

        void readFromTCL(TTCL_Interpreter & interpreter, const char * varName);
        
        void writeToTCL(TTCL_Interpreter & interpreter, const char * varName);
    };
    
    /// Typ bonusu
    class TBonus : public TStructHolder<UNIT_PROPERTY_BONUS>, public TTCL_Struct
    {
    public:
        TBonus();

        TBonus(UNIT_PROPERTY_BONUS * data, int id = 0);

        void readFromTCL(TTCL_Interpreter & interpreter, const char * varName);
        
        void writeToTCL(TTCL_Interpreter & interpreter, const char * varName);
    };

    /** Zivouci jednotka 
        Objektovy obal struktury LIVING_UNIT
    */        
    class TLivingUnit : public TStructHolder<LIVING_UNIT>, public TTCL_XML_Struct, public TMutex
    {
    private:
        HEX_ID _hex;

        BUILDING_ID _building;
        
        World::TWorld * _world;

        friend class TWorld;      

        friend class TWorldEngine;
    public:
        TLivingUnit();

        ~TLivingUnit();

        TLivingUnit(LIVING_UNIT * data, int id = 0);

        TLivingUnit(int id);

        void readFromXML(TPackage * package);

        void writeToXML(TPackage * package);

        void readFromTCL(TTCL_Interpreter & interpreter, const char * varName);
        
        void writeToTCL(TTCL_Interpreter & interpreter, const char * varName);

        HEX_ID hex();

        World::THex * tHex();

        BUILDING_ID building();

        World::TLivingBuilding * tBuilding();

        /// muze tato jednotka v tomto kle keste neco uzitecneho udelat
        /// Pouziva se pri listovani jednotkami pomoci tlacitek "Dalsi" a "Predchozi"
        bool isOperational();

        /// soucet bonusu k utoku
        int attackBonus();

        /// soucet bonusu k obrane
        int defenseBonus();

        /// soucet bonusu k poctu zivotu
        int maxLivesBonus();
        
        /// soucet bonusu k bodum pohybu
        int movementPointsBonus();

        /// soucet bonusu k dosahu utoku
        int attackRangeBonus();
    };
}

#endif
/*****************************************************************************/
