/** @file /world/world.h

	Spolecne definice, konstanty a nastavani obou modulu World (Server a Client)

	@see /world/world_client.h
	@see /world/world_server.h

	@author Petr Wolf
*/

#ifndef WORLD_H
#define WORLD_H

class TPacket_RCT_UNIT_ATTACK;

#include "common/mm.h"
#include "common/Log.h"
#include "world/typedefs.h"
#include "world/rules.h"
#include "world/weather.h"
#include "world/da_holder.h"
#include "world/sa_holder.h"
#include "ai/Diplomacy/diptypes.h"

/** Modul pravidel WORLD.
    Namespace pro WORLD_SERVER, WORLD_CLIENT, reprezentaci datovych struktur a metod
*/
namespace World
{
    /// seznam jednotek na mape
    typedef TDAHolder_Mutex<LIVING_UNIT, UNIT_ID, TLivingUnit> TLivingUnitsContainer;
    /// iterator pres seznam jednotek na mape
    typedef TLivingUnitsContainer::iterator TLivingUnitsIterator;

    /// seznam budov na mape
    typedef TDAHolder_Mutex<LIVING_BUILDING, BUILDING_ID, TLivingBuilding> TLivingBuildingsContainer;
    /// iterator pres seznam budov
    typedef TLivingBuildingsContainer::iterator TLivingBuildingsIterator;
   
    /// seznam mest na mape
    typedef TDAHolder_Mutex<TOWN, int, TTown> TTownsContainer;
    /// iterator pres seznam map
    typedef TTownsContainer::iterator TTownsIterator;

    /// seznam kralovstvi v mape
    typedef TSAHolder_Mutex<KINGDOM, int, TKingdom> TKingdomsContainer;
    /// iterator pres seznam kralovstvi
    typedef TKingdomsContainer::iterator TKingdomsIterator;

    /// umisteni modulu WORLD
	enum WORLD_SITUATION
	{
		WS_SINGLE
	   ,WS_CLIENT
	   ,WS_SERVER
	};

	/** Zakladni trida reprezentace herniho sveta
		Obsahuje veskere informace o hernim svete a jeho hracich, kompletni pozice vsech
		jednotek, celou mapu, ale i masky viditelnosti jednotlivych hracu a dalsi
		udaje dulezite pro hru, vcetne aktivniho stavu hry (kdo je na rade, vztahy mezi hraci atd.)

		Objekt teto tridy existuje v kazde hre v jedne instanci u kazdeho klienta a v jedne instanci 
		na hernim serveru. Mezi nimi pak probiha synchronizace.
		
		Ve specialnim pripade singleplayer hry na lokale muze existovat jen jednou 
		a plnit obe role soucasne. V takovem pripade budou sice synchornizacni funkce volat, ale nebudou
		nic delat.
	*/
    class TWorld : public TMutex
	{
    private:
        bool rules_ownership;
    public:
        /// pravidla, tabulky, koeficienty, skripty
        World::TRules * rules;  

        /// herni mapa
        World::TMap map;

        /// kralovstvi
        TKingdomsContainer kingdoms;

        /// mesta
        TTownsContainer towns;

        /// mapy viditelnosti
        TVisibilityMapsContainer visibility_maps;

        /// instance zivych jednotek
        TLivingUnitsContainer units;

        /// "archiv" zabitych jednotek
        TLivingUnitsContainer corpses;
        
        /// instance zivych budov
        TLivingBuildingsContainer buildings;

        /// "archiv" znicenych budov
        TLivingBuildingsContainer ruins;

        /// seznam hracu
        TPlayerContainer players;

        /// pocasi
        TWeather weather;

        /// cislo herniho kola
        int turn_id;
        
        /// Konstruktor
        TWorld(); 

        /// Destruktor
        ~TWorld();

        /// Inicializace, previrajici odkaz na objekt pravidel
        void init(World::TRules * external_rules = NULL);

        /// Vyprazdneni dat
        void reset();

        /** Vyplneni redundantnich vztahu a zkratek
            Funkce vyplni v objektech vsechny redundanti vzajemne ukazatele, 
            slouzici k urcyhleni.

            @param  refresh  nastavit na \c TRUE, neni-li svet tesne po inicializaci
                             a je nejprve treba smazat stara data.
            */
        void fillRelations(bool refresh = false);

        /** Nove kolo.
            Zaneseni zmen v dusledku noveho kola.
            */
        void nextTurn(MONEY moneyGain, TIntArray & movementPoints, TIntArray & buildingLives, TIntContainer & unitsInterrupted, TIntArray * menDied = NULL);

        /** porazeni hrace */
        void playerDefeated(PLAYER_ID player_id);

        /** Vyvtoreni jednotky.
            Metoda zanese do sveta vznik nove jednotky, aktualizuje vazby mezi entitami a odecte
            patricnou sumu z hracovych penez.
        */
        TLivingUnit * unitCreated(BUILDING_ID building_id, UNIT_TYPE_ID unit_type, UNIT_ID unit_id, MONEY cost);

        /// presun jednotky na hex
        void unitStep(UNIT_ID unit_id, HEX_ID to);

        /** Pohyb jednotky.
            Metoda zanese do struktur sveta pohyb jednotky po zadane ceste.
        */
        void unitMove(UNIT_ID unit_id, HEX_ID to, TPath * path = NULL);

        /// Zaneseni zmen do struktur po smrti jednotky.
        void unitDied(UNIT_ID unit_id);

        /** Vytvoreni budovy.
            Metoda zanese do sveta vznik nove bodovy, aktualizuje vazby mezi entitami a odecte
            patricnou sumu z hracovych penez.
        */
        TLivingBuilding * buildingCreated(UNIT_ID unit_id, BUILDING_TYPE_ID building_type, BUILDING_ID building_id, HEX_ID location, int lives, MONEY cost, int orientation);

        /** Oprava budovy.
            Metoda zanese do sveta opravu budovy.
        */
        void buildingRepaired(UNIT_ID unit_id, BUILDING_ID building_id, int lives, MONEY cost);

        /** Preruseni stavby jednotky */
        void stopBuilding(UNIT_ID unit_id);

        /** Prodej bonusu jednotce v budove */
        void sellBonus(BUILDING_ID building_id, UNIT_PROPERTY_BONUS_ID bonus_id, MONEY cost, UNIT_PROPERTY_BONUS_ID replace);

        /** Zraneni jednotky */
        void unitWound(UNIT_ID unit_id, int wounds);

        /** Pricteni zkusenosti jednotce */
        void unitExperience(UNIT_ID unit_id, int experience, int level);

        /** Leceni jednotky */
        void unitHeal(UNIT_ID unit_id, int lives);
        
        /** Doplneni jednotky (neelitni) */
        void unitRecruit(UNIT_ID unit_id, int lives,  int cost, int level, int experience);

        /** Doplneni jednotky (elitni) */
        void unitRecruitElite(UNIT_ID unit_id, int lives, int cost);

        /// Zaneseni zmen do struktur po zniceni budovu
        void buildingDestroyed(BUILDING_ID building_id);

        /** Rozhodnuti, je-li hrac momentalne schopen akce */
        bool playerOperational(PLAYER_ID player_id);

        /** Rozhodnuti, je-li jednotka momentalne akceschopna.
            Funkce provede nekoliik jednoduchych testu urcujicich, jestli
            jednotka patri hraci na tahu a jestli se o akci s ni
            pokousi skutecne jeji majitel.
        */            
        bool unitOperational(UNIT_ID unit_id, PLAYER_ID player_id);
        
        /** Rozhodnuti, je-li budova momentalne akceschopna.
            Funkce provede nekoliik jednoduchych testu urcujicich, jestli
            budova patri hraci na tahu a jestli se o akci s ni
            pokousi skutecne jeji majitel.
        */ 
        bool buildingOperational(BUILDING_ID building_id, PLAYER_ID player_id);

        /** Zmeny v diplomatickych relacich mezi hraci */
        bool diplomacyChange(int source_diplomat_world_id, int target_diplomat_world_id, ai_ns::diplomacy_ns::TRelationship new_relationship, ai_ns::diplomacy_ns::TRelationship new_offered_relationship, MONEY source_diplomat_penalty, MONEY target_diplomat_penalty);
        
        /** Nastaveni pocasi */
        void setWeather(WEATHER_STATE state);

        void visibilityShow(PLAYER_ID player_id, TIntContainer & shown);

        void visibilityHide(PLAYER_ID player_id, TIntContainer & shown);

        /// Hromadne zamceni vsech zamykatelnych objektu uvnitr (mapy, jednotek, ...)
        int lock();

        /// Hromadne odemceni vsech zamykatelnych objektu uvnitr (mapy, jednotek, ...)
        int unlock();

	};    
};

#include "world/world_messages.h"

#endif
/*****************************************************************************/
