/** @file /world/world_messages.h
    
    Tela zprav modulu World
    
    @author Petr Wolf
*/

#ifndef WORLD_MESSAGES_H 
#define WORLD_MESSAGES_H

#include <stdarg.h>
#include <string>
#include "common/mm.h"
#include "common/rm/rminit.h"
#include "common/xml/xml_struct.h"
#include "ai/PathFind/pathenv.h"
#include "ai/Diplomacy/diptypes.h"
#include "world/useful.h"
#include "world/world.h"
#include "world/history.h"

// +--------------------------------------------------------------------------+
// |                                                                          |
// |             Z P R A V Y   P O S I L A N E   L O K A L N E                |
// |                                                                          |
// +--------------------------------------------------------------------------+

/** Chybove kody */
enum K8_ERROR {
    /// Obecna chyba, bez podrobnejsi specifikace
    ERR_General = 1   

    /// Chyba pri zpracovavani pozadavku na serveru
   ,ERR_ServerError

    /// Neplatne souradnice
   ,ERR_InvalidCoords
   
    /// Neplatne ID hexu
   ,ERR_InvalidHexId

    /// Neplatne ID jednotky
   ,ERR_InvalidUnitId   

    /// Neplatne ID budovy
   ,ERR_InvalidBuildingId

    /// Struktury sveta nebyly korektne inicializovany
   ,ERR_WorldNotInitialized

    /// Chybny vstup pro metodu (neplatna kombinace dat)
   ,ERR_InvalidInput

    /// Zvoleny objekt (hrac, jednotka, budova) neni momentalne akceschopna (neni na tahu)
   ,ERR_NotOperational

    /// Nedostatek penez pro zvolenou akci
   ,ERR_NotEnoughMoney

    /// Na zasobniku hernich akci uz nic neni.
    /// Zprava pro AI, ze se jiz dobrala na vrchol zasobniku hernich akci
    /// a muze pokracovat zaslanim dalsi akce, nebo ukoncit tah.
   ,ERR_NoMoreActions

    /// Hra uz byla ukoncena, jdi domu.
    /// Zprava pro AI, ktere by nadale posilalo pozadavky na akce
   ,ERR_GameOver
};

/// Univerzalni trida pro vraceni slozitejsich datovych struktur v MSG systemu.
class TPacket_SyncResult
{
public:
    /// Implicitni konstruktor
    TPacket_SyncResult();

    /// Konstruktor pro vraceni chyboveho kodu
    TPacket_SyncResult(K8_ERROR code);

    /** Status, se kterym volani zpravy skoncilo. Ma hodnotu 0, je-li vsechno 
        v poradku, jinak kod chyby.
    */
    int status;
};

/** Trida, ktera se posila GUI v momente, kdy ze serveru
    dorazi zprava o provedeni nejake herni akce.
    */
class TPacket_AsyncResult
{
public:
    TPacket_AsyncResult();

    /// ID hrace, kteremu je zprava posilana
    PLAYER_ID player_id;   
};

// oznameni o chybe
class TPacket_ErrorResult : public TPacket_AsyncResult
{
public:
    /// konstruktor
    TPacket_ErrorResult();

    /// kod chyby
    int error;
};

/// Zaklad zpravy prijate prez sit od modulu WORLD_SERVER pro WORLD_CLIENT 
struct PACKET_WORLD_CLIENT_SERVER
{
    /// typ zpravy (MSG_neco)
    int type;

    /// hrac, kteremu je zprava urcena
    int player_id;
};

/// Zaklad zpravy pro komunikace WORLD_CLIENT <-> GUI
struct PACKET_WORLD_GUI
{
    /// typ zpravy (MSG_neco)
    int type;

    /// hrac, kteremu je zprava urcena, nebo ktery zpravu posila
    int player_id;
};

/// Zaklad zpravy pro komunikaci WORLD_CLIENT, WORLD_SERVER <-> AI
#define PACKET_WORLD_AI PACKET_WORLD_GUI

/// Zaklad zpravy pro komunikace WORLD_CLIENT/WORLD_SERVER <-> TCL
struct PACKET_WORLD_TCL
{
    /// typ zpravy
    int type;

    /// pocet argumentu (vcetne nazvu volane funkce)
    int argc;

    /// ukazatele na hodnoty argumentu (delky argc + 1, posledni ma hodnotu NULL)
    char ** argv;

    /// jmeno promenne, do ktere ma byt vysledek vykonavani funkce ulozen
    char * varName;
};

/// Paket pro vyber rezimu hry
struct PACKET_GAME_MODE
{
    /// rezim hry
    GAME_MODE mode;

    /// IP adresa serveru, k nemuz se chce klient v pripade sitove hry pripojit
    char * address;

    /// nick
    char nick[MAX_STRLEN];
};

/// Nastaveni hrace
struct PACKET_PLAYER_SETTINGS
{
    /// id herni role, kterou si hrac zvolil
    int player_id;

    /// jmeno hrace
    char name[MAX_STRLEN];

    /// typ hrace
    World::PLAYER_TYPE type;

    /// command the player can give
    World::PLAYER_ALLOWED_COMMANDS acommands;

    /// id sitoveho klienta
    int client_id;
};

/// Paket pro start hry
struct PACKET_GAME_START
{
    /// id mapy
    int map_id;

    /// volba podminek ukonceni hry
    END_GAME_CONDITION end_game_condition;

    /** seznam hracu
        Pole by melo byt naplneno od indexu 0 dale (spojite).
        Nepouzite pozice by mely obsahovat jako \c id hodnotu -1.
        */
    PACKET_PLAYER_SETTINGS players[MAX_PLAYERS];
};

/// Kopie zpravy PACKET_GAME_START pro jeji zpracovani v samostatnem vlakne
void PACKET_GAME_START_Copy(const PACKET_GAME_START * source, PACKET_GAME_START * target);

/// Paket pro ulozeni hry
struct PACKET_SAVE_GAME 
{
    char filename[K8_MAX_FILE_NAME_LEN];
};

/** Telo zpravy MSG_LETS_GO.
    @see MSG_LETS_GO
*/
struct PACKET_LETS_GO
{
    /// struktury z pravidel
    World::TRules * rules;

    /// mapa
    World::MAP * map;

    /// masky viditelnosti
    World::TVisibilityMapsContainer * visibility_maps;

    /// pocasi
    World::TWeather * weather;

    /// mesta
    World::TTownsContainer * towns;

    /// kralovstvi
    World::TKingdomsContainer * kingdoms;

    /// zive jednotky
    World::TLivingUnitsContainer * living_units;

    /// mrtve jednotky
    World::TLivingUnitsContainer * corpses;

    /// zive budovy
    World::TLivingBuildingsContainer * living_buildings;

    /// znicene budovy
    World::TLivingBuildingsContainer * ruins;

    /// seznam hracu
    World::TPlayerContainer * players;

    /// cislo tahu
    int turn_id;

    /// hrac na tahu
    PLAYER_ID player_on_turn;

    /// nejblizsi lidsky hrac (0, neexistuje-li)
    PLAYER_ID human_player;
};

/** Zadost o odpojeni GUI.
    @see MSG_END_GAME_GUI
*/
struct PACKET_END_GAME_GUI : public PACKET_WORLD_GUI
{
    /// Nastavit na \c true, ma-li byt soucasne ukoncena cela hra
    /// Naopak nastavit na \c false, ma-li se pouze od hry odpojit hrac, 
    /// ktery zpravu posila.
    bool final;
};

/** Ukonceni hry na popud WORLDu.
    @see MSG_END_GAME
*/
class TPacket_END_GAME : public TXML_Struct
{
public:
    /// duvod, ktery vedl k ukonceni hry
    World::ENDGAME_REASON reason;

    /// ID hrace, ktery zvitezil, nebo 0, nevyhral-li nikdo
    PLAYER_ID winner;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zjisteni vzdalenosti mezi hexy pro konkretni jednotku.
    @see MSG_GET_DISTANCE
*/ 
struct PACKET_GET_DISTANCE
{
    /// ID jednotky, ktere se pohyb tyka
    UNIT_ID unit_id;

    /// body pohybu, ktere doted jednotka "usla" behem vypoctu pohybu
    int points_of_movement;

    /// ID startovniho hexu
    HEX_ID from;

    /// ID ciloveho hexu
    HEX_ID to;

    /// transparence hledani cesty
    ai_ns::pathfind_ns::TGetCostTransparency getcost_transparency;
};

/** Odpoved na PACKET_GET_DISTANCE 
    @see MSG_GET_DISTANCE
*/
class TPacket_RET_DISTANCE : public TPacket_SyncResult
{
public:
    /** Pocet bodu pohybu, ktere by pohyb jednotku stal, nebo zbyvajici pocet BP 
        v pripade, ze hex neni dostupny.
        */        
    int cost;
};

/** Vraceni mapy.
    @see MSG_GET_MAP
*/
class TPacket_RET_MAP : public TPacket_SyncResult
{
public:
    /// ukazatel na mapu
    World::MAP * map;
};

/** Ziskani rozsahu pohybu jednotky.
    @see MSG_GET_RANGE
    @see TPacket_RET_RANGE
*/
struct PACKET_GET_RANGE : public PACKET_WORLD_GUI
{
    /// ID jednotky, o kterou se zajimame
    UNIT_ID unit_id;    
};

/** Vraceni rozsahu pohybu jednotky.
    @see MSG_GET_RANGE
    @see PACKET_GET_RANGE
*/
class TPacket_RET_RANGE : public TPacket_SyncResult
{
public:
    std::vector<HEX_ID> range;
};

/** Ziskani rozsahu utoku jednotky.
    @see MSG_GET_ATTACK_RANGE
    @see TPacket_RET_ATTACK_RANGE
*/
struct PACKET_GET_ATTACK_RANGE : public PACKET_WORLD_GUI
{
    /// ID jednotky, o kterou se zajimame
    UNIT_ID unit_id;
};

/** Vraceni rozsahu utoku jednotky.
    @see MSG_GET_ATTACK_RANGE
    @see PACKET_GET_ATTACK_RANGE
*/
class TPacket_RET_ATTACK_RANGE : public TPacket_SyncResult
{
public:
    std::vector<HEX_ID> attack_range;
};

/** Ziskani moznych mist, odkud muze byt jednotka napadena
    @see MSG_GET_INVERSE_ATTACK_RANGE
    @see PACKET_RET_INVERSE_ATTACK_RANGE
*/
struct PACKET_GET_INVERSE_ATTACK_RANGE : public PACKET_WORLD_AI
{
    /// ID jednotky, ktera ma utocit
    UNIT_ID attacker;

    /// ID hexu, ktery ma byt napaden
    HEX_ID target;
};

/** odpoved na zpravu MSG_GET_INVERSE_ATTACK_RANGE
    @see MSG_GET_INVERSE_ATTACK_RANGE
    @see PACKET_GET_INVERSE_ATTACK_RANGE
*/
class TPacket_RET_INVERSE_ATTACK_RANGE : public TPacket_SyncResult
{
public:
    /// seznam hexu, z nichz muze byt cilovy hex napaden
    std::vector<HEX_ID> attack_range;
};

/** Zadost o vypocet PATHFIND algoritmu */
struct PACKET_PATHFIND_REQUEST
{
    /// jednotka, ktera ma pohyb vykonat
    UNIT_ID unit_id;

    /// pocet bodu pohybu jednotky
    int points_of_movement;

    /// startovni hex
    COORDS from;

    /// cilovy hex
    COORDS to;

    /// transparence hledani cesty
    ai_ns::pathfind_ns::TPathFindTransparency pathfind_transparency;
};

/** Zadost o vypocet FLOODFILL algoritmu */
struct PACKET_FLOODFILL_REQUEST
{
    /// jednotka, ktera ma pohyb vykonat
    UNIT_ID unit_id;

    /// startovni hex
    COORDS start_hex;

    /// markmapa, kde by mely byt nastaveny hodnoty true, kam hrac jemuz patri jednotka nevidi
    ai_ns::pathfind_ns::PF_MARKMAP mark_map;

    /// maska, ktera bude naplnena hodnotami flood fill algoritmu
    ai_ns::pathfind_ns::FLOOD_MASK flood_mask;

    /// transparence
    ai_ns::pathfind_ns::TPathFindTransparency flooding_transparency;
};

/** Akce - kliknuti na jednotku */
struct PACKET_ACT_UNIT_CLICK : public PACKET_WORLD_GUI
{
    /// id jednotky, na kterou se kliklo
    UNIT_ID unit_id;
};

/** Vyhodnoceni kliknuti na jednotku - zpristupnuje jeji polozky */
class TPacket_RCT_UNIT_CLICK : public TPacket_SyncResult
{
public:
    TPacket_RCT_UNIT_CLICK();

    /// seznam moznych akci (0 ... nemuze, 1 a vic ... muze)
    int actions[UNIT_ACTION_COUNT];

    /// seznam typu budov, ktere jednotka muze stavet
    std::vector<BUILDING_TYPE_ID> buildings;

    /// priznak, co se o jednotce smi zobrazit
    World::UNIT_KNOWLEDGE knowledge;
};
/** Akce - kliknuti na budovu */
struct PACKET_ACT_BUILDING_CLICK : public PACKET_WORLD_GUI
{
    /// id budovy, na kterou se kliklo
    BUILDING_ID building_id;
};

/** Vyhodnoceni kliknuti na budovu - zpristupnuje jeji polozky */
class TPacket_RCT_BUILDING_CLICK : public TPacket_SyncResult
{
public:
    TPacket_RCT_BUILDING_CLICK();

    /// seznam moznych akci (0 ... nemuze, 1 a vic ... muze)
    int actions[BUILDING_ACTION_COUNT];

    /// seznam jednotek vyrobitelnych v budove
    std::vector<int> recruitment;

    /// seznam bonusu, ktere je mozno v budove prodat
    std::vector<int> bonuses;

    /// priznak, co se o budove smi zobrazit
    World::BUILDING_KNOWLEDGE knowledge;
};

/** Zadost o sebevrazdu jednotky */
struct PACKET_ACT_UNIT_SUICIDE : public PACKET_WORLD_GUI
{
    /// id jednotky, ktera se chce znicit
    UNIT_ID unit_id;
};

/** Vyhodnoceni sebevrazdy jednotky */
class TPacket_RCT_UNIT_SUICIDE : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// ide znicene jednotky
    UNIT_ID unit_id;
    
    /// ztrata vyhledu
    TIntContainer hidden;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zadost o sebezniceni budovy */
struct PACKET_ACT_BUILDING_SUICIDE : public PACKET_WORLD_GUI
{
    /// id budovy, ktera se chce znicit
    BUILDING_ID building_id;
};

/** Vyhodnoceni sebezniceni budovy */
class TPacket_RCT_BUILDING_SUICIDE : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// id znicene budovy
    BUILDING_ID building_id;

    /// id jednotky, ktera pripadne zahynula
    UNIT_ID unit_killed;
    
    /// ztrata vyhledu
    TIntContainer hidden;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zadost o koupi nove jednotke */
struct PACKET_ACT_BUILDING_RECRUIT_UNIT : public PACKET_WORLD_GUI
{
    /// id budovy, ktera ma jednotku vyrobit
    BUILDING_ID building_id;

    /// typ jednotky, ktera ma byt zakoupena
    UNIT_TYPE_ID unit_type; 
};

/** Vyhodnoceni koupe jednotky */
class TPacket_RCT_BUILDING_RECRUIT_UNIT : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// id budovy, ktera vyrobila novou jednotku
    BUILDING_ID building_id;

    /// typ jednotky, ktera vyrobou vznikla
    UNIT_TYPE_ID unit_type; 

    /// id jednotkty, ktera vyrobou vznikla
    UNIT_ID unit_id;

    /// cena, kterou za to hrac zaplatil
    World::MONEY cost;

    /// hexy, ktere se vyrobenim jednotky objevi ve vyhledu
    TIntContainer shown;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zadost o seznam moznych umisteni pro stavbu budov */
struct PACKET_GET_BUILDING_RANGE : PACKET_WORLD_GUI
{
    /// id stavitelske jednotky
    UNIT_ID unit_id;

    /// typ budovy, ktera se ma stavet
    BUILDING_TYPE_ID building_type;
};

/** Seznam moznych umisteni a orientaci pro stavbu */
class TPacket_RET_BUILDING_RANGE_Position
{
public:
    TPacket_RET_BUILDING_RANGE_Position();

    TPacket_RET_BUILDING_RANGE_Position(HEX_ID _position, std::vector<int> _orientation);

    HEX_ID position;

    std::vector<int> orientation;
};

/** Zprava obsahujici seznam moznych umisteni pro stavbu */
class TPacket_RET_BUILDING_RANGE : public TPacket_SyncResult
{
public:
    /// seznam moznych umisteni  a orientaci budovy
    std::vector<TPacket_RET_BUILDING_RANGE_Position> locations;
};

/** Zadost o seznam moznych cilu k oprave */
struct PACKET_GET_REPAIR_RANGE : PACKET_WORLD_GUI
{
    /// id stavitelske jednotky
    UNIT_ID unit_id;
};

/** Vyhodnoceni seznamu moznych cilu k oprave */
class TPacket_RET_REPAIR_RANGE : public TPacket_SyncResult
{
public:
    /// seznam budov, ktere je mozne opravit
    std::vector<BUILDING_ID> buildings;
};

/** Zadost o zahajeni stavby */
struct PACKET_ACT_START_BUILDING : PACKET_WORLD_GUI
{
    /// id jednotky stavitele
    UNIT_ID unit_id;

    /// typ stavene budovy
    BUILDING_TYPE_ID building_type;

    /// hex, na kterem je staveno
    HEX_ID location;

    /// orientace budovy
    int orientation;
};

/** Vyhodnoceni zahajeni stavby */
class TPacket_RCT_START_BUILDING : public TPacket_AsyncResult
{
public:
    /// id jednotky stavitele
    UNIT_ID unit_id;

    /// typ stavene budovy
    BUILDING_TYPE_ID building_type;

    /// hex, na kterem je staveno
    HEX_ID location;

    /// orientace budovy
    int orientation;

    /// id stavene budovy
    BUILDING_ID building_id;
    
    /// pocet zivotu, se kterymi budova vznika
    int lives;
    
    /// cena, kterou hrace zahajeni stavby stoji
    World::MONEY cost;

    /// hexy, ktere se zahajenim stavby objevi ve vyhledu
    TIntContainer shown;

    /// hexy, ktere se zahajenim stavby ztrati z vyhledu
    TIntContainer hidden;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zadost o zahajeni opravy */
struct PACKET_ACT_REPAIR_BUILDING : PACKET_WORLD_GUI
{
    /// id jednotky stavitele
    UNIT_ID unit_id;

    /// id budovy
    BUILDING_ID building_id;
};

/** Vyhodnoceni zahajeni oprav */
class TPacket_RCT_REPAIR_BUILDING : public TPacket_AsyncResult
{
public:
    /// id jednotky stavitele
    UNIT_ID unit_id;

    /// id budovy
    BUILDING_ID building_id;

    /// pocet zivotu, ktere oprava budove pridala
    int lives;

    /// cena, kterou hrace zahajeni opravy stoji
    World::MONEY cost;

    /// hexy, ktere se zahajenim opravy objevi ve vyhledu
    TIntContainer shown;

    /// hexy, ktere se zahajenim opravy ztrati z vyhledu
    TIntContainer hidden;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zadost o preruseni stavby */
struct PACKET_ACT_STOP_BUILDING : public PACKET_WORLD_GUI
{
    /// ID jednotky, ktera doted stavela
    UNIT_ID unit_id;
};

/** Vyhodnoceni preruseni stavby */
class TPacket_RCT_STOP_BUILDING : public TPacket_AsyncResult
{
public:
    /// ID jednotky, ktera prestala stavet
    UNIT_ID unit_id;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zadost o prodej bonusu */
struct PACKET_ACT_SELL_BONUS : public PACKET_WORLD_GUI
{
    /// ID budovy prodavajici bonus
    BUILDING_ID building_id;

    /// ID kupovaneho bonusu
    UNIT_PROPERTY_BONUS_ID bonus_id;
};

/** Vyhodnoceni prodeje bonusu */
class TPacket_RCT_SELL_BONUS : public TPacket_AsyncResult
{
public:
    /// ID budovy prodavajici bonus
    BUILDING_ID building_id;

    /// ID kupovaneho bonusu
    UNIT_PROPERTY_BONUS_ID bonus_id;

    /// cena, kterou je treba za bonus odecist
    World::MONEY cost;

    /// ID bonusu, ktery je novym bonusem nahrazen
    UNIT_PROPERTY_BONUS_ID replace;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};
   
/** Zadost o pohyb jednotky */
struct PACKET_ACT_UNIT_MOVE : public PACKET_WORLD_GUI
{       
    /// id jednotky, ktera se ma pohnout
    UNIT_ID unit_id;

    /// cilovy bod, po ktere ma jit
    HEX_ID target;

    /// transparence, se kterou je pocitana cesta
    /// vyplnuje pouze AI; pro GUI neni tento parametr dulezity
    ai_ns::pathfind_ns::TPathFindTransparency transparency;
};

/// pohyb jednotky
/// @see MSG_ACT_UNIT_MOVE
class TPacket_RCT_UNIT_MOVE : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// id jednotky, ktera se pohnula
    UNIT_ID unit_id;

    /// cesta, po ktere se jednotka vydala
    TPath path;
    
    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/// zacatek noveho kola
class TPacket_RCT_NEXT_TURN : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// ID hrace, ktery se dostal na radu
    PLAYER_ID player_id_on_turn;

    /// nastalo nove kolo?
    bool newRound;

    /// financni prijem hrace
    int moneyGain;

    // varovani, ze v pristim kole mohou dojit penize na zold
    bool moneyLack;

    /// zmena pocasi (pouze, nastalo-li nove kolo)
    World::WEATHER_STATE weatherState;

    /// prirustek bodu pohybu u jednotek
    TIntArray movementPoints;

    /// prirustek zivotu staveb
    TIntArray buildingLives;

    // seznam jednotek, ktere v dusledku nedostatku penez na zold opustily vojsko
    TIntContainer unitsDeserted;

    // seznam jednotek, jejichz stavebni cinnost byla prerusena v dusledku nedostatku penez
    TIntContainer unitsInterrupted;

    // seznam jednotek ohrozenych smrti v dusledku zraneni
    TIntContainer unitsWounded;

    // seznam jednotek, ktere v dusledku zraneni zemrely
    TIntContainer unitsDied;

    // seznam muzu, kteri v dusledku zraneni zemreli
    TIntArray menDied;

    // hexy, o ktere se viditelnost rozsirila
    TIntContainer shown;

    // hexy, ktere se z vyhledu ztratily
    TIntContainer hidden;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zadost o utok jednotky */
struct PACKET_ACT_UNIT_ATTACK : public PACKET_WORLD_GUI
{
    /// id utocnika
    UNIT_ID attacker;

    /// id napadeneho hexu
    HEX_ID target;
};

/** Vyhodnoceni souboje jednotek */
class TPacket_RCT_UNIT_ATTACK : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// konstruktor
    TPacket_RCT_UNIT_ATTACK();

    /// id utocnika
    UNIT_ID attacker;

    /// id napadeneho hexu
    HEX_ID target;

    /// ztraty na zivotech utocnika
    int attacker_losses;

    /// zraneni utocnika
    int attacker_wounds;

    /// ztraty na zivotech obrance
    int defender_losses;

    /// zraneni obrance
    int defender_wounds;

    /// ztraty na napadene budove
    int building_losses;

    /// zkusenosti utocnika
    int attacker_experience;

    /// uroven utocnika
    int attacker_level;

    /// zkusenosti obrance
    int defender_experience;

    /// uroven obrance
    int defender_level;

    /// zahynul utocnik? 
    /// 0 pokud ne, jeho id pokud ano
    int attacker_killed;

    /// zahynul obrance?
    /// 0 pokud ne, jeho id pokud ano
    int defender_killed;

    /// byla znicena budova?
    /// 0 pokud ne, jeji id pokud ano
    int building_destroyed;

    /// ztrata vyhledu utocnika
    TIntContainer attacker_hidden;

    /// ztrata vyhledu obrance
    TIntContainer defender_hidden;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zadost o leceni jednotky */
struct PACKET_ACT_UNIT_HEAL : public PACKET_WORLD_GUI
{
    /// ID jednotky, ktera se chce lecit
    int unit_id;   
};

/** Vyhodnoceni leceni jednotky */
class TPacket_RCT_UNIT_HEAL : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// ID lecene jednotky
    int unit_id;

    /// pocet zraneni, ktere se podarilo vylecit
    int lives;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zadost o neelitni doplneni jednotky */
struct PACKET_ACT_UNIT_RECRUIT : public PACKET_WORLD_GUI
{
    /// ID jednotky, ktera se chce doplnit
    int unit_id;
};

/** Vyhodnoceni neelitniho doplneni jednotky */
class TPacket_RCT_UNIT_RECRUIT  : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// ID doplnene jednotky
    int unit_id;

    /// cena za doplneni jednotky
    int cost;

    /// pocet doplnenych zivotu
    int lives;

    /// nova uroven jednotky
    int level;

    /// novy pocet zkusenosti jednotky
    int experience;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zadost o elitni doplneni jednotky */
struct PACKET_ACT_UNIT_RECRUIT_ELITE : public PACKET_WORLD_GUI
{
    /// ID jednotky, ktera se chce doplnit
    int unit_id;
};

/** Vyhodnoceni neelitniho doplneni jednotky */
class TPacket_RCT_UNIT_RECRUIT_ELITE : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// ID doplnene jednotky
    int unit_id;

    /// cena za doplneni jednotky
    int cost;

    /// pocet doplnenych zivotu
    int lives;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zprava o obsazeni budovy */
class TPacket_BUILDING_OCCUPIED : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// ID obsazene budovy
    int building_id;

    /// ID noveho majitele
    int new_player_id;

    /// ID puvodniho majitele (jen kvuli polozce \c hidden)
    int former_player_id;

    /// hexy, ktere se obsazenim budovy ztrati z vyhledu jejiho dosavadniho majitele
    TIntContainer hidden;

    /// hexy, ktere obsazenim budovy ziskava do vyhledu jeji novy majitel
    TIntContainer shown;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zprava o obsazeni mesta */
class TPacket_TOWN_OCCUPIED : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// ID obsazeneho mesta
    int town_id;

    /// ID noveho majitele
    int new_player_id;

    /// ID puvodniho majitele
    int former_player_id;

    /// hexy, ktere se obsazenim mesta ztrati z vyhledu jeho dosavadniho majitele
    TIntContainer hidden;

    /// hexy, ktere obsazenim mesta ziskava do vyhledu jeho novy majitel
    TIntContainer shown;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zprava o obsazeni kralovstvi */
class TPacket_KINGDOM_OCCUPIED : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// ID obsazeneho kralovstvi
    int kingdom_id;

    /// ID noveho krale
    int new_player_id;

    /// ID puvodniho majitele
    int former_player_id;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Zprava o porazeni hrace */
class TPacket_PLAYER_DEFEATED : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// ID porazeneho hrace
    int defeated_player_id;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);

};

/** Zprava o zmene viditelnosti */
class TPacket_VISIBILITY_CHANGE : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// ID hrace, jehoz viditelnost se meni
    int visibility_mask_player_id;

    /// seznam odkrytych hexu
    TIntContainer shown;

    /// seznam skrytych hexu
    TIntContainer hidden;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

/** Navrh diplomatickeho vztahu */
struct PACKET_ACT_OFFER_RELATION : public PACKET_WORLD_GUI
{
    /// ID hrace navrhujiciho vztah
    int source_diplomat_world_id;

    /// ID druhe strany
    int target_diplomat_world_id;
    
    /// navrhovovany vztah
    ai_ns::diplomacy_ns::TRelationship offered_relationship;
};

/** Zprava o zmene diplomatickeho vztahu */
class TPacket_DIPLOMACY_RELATION_CHANGE : public TPacket_AsyncResult, public TXML_Struct
{
public:
    /// Implicitni konstruktor
    TPacket_DIPLOMACY_RELATION_CHANGE();

    /// Konstruktor kopirujici data ze strukturt TDipChange
    TPacket_DIPLOMACY_RELATION_CHANGE(ai_ns::diplomacy_ns::TDipChange change);

    /// ID zdrojoveho diplomata
    int source_diplomat_world_id;

    /// ID ciloveho diplomata
    int target_diplomat_world_id;

    /// stavajici platny vztah
    ai_ns::diplomacy_ns::TRelationship old_relationship;

    /// stavajici navrzeny vztah
    ai_ns::diplomacy_ns::TRelationship old_offered_relationship;
    
    /// novy platny vztah
    ai_ns::diplomacy_ns::TRelationship new_relationship;
    
    /// novy navrzeny vztah
    ai_ns::diplomacy_ns::TRelationship new_offered_relationship;

    /// financni penale, ktere je strhnuto hraci \c source_diplomat_world_id
    World::MONEY source_diplomat_penalty;

    /// financni penale, ktere je strhnuto hraci \c target_diplomat_world_id
    World::MONEY target_diplomat_penalty;

    /// zapis do XML
    void writeToXML(TPackage * package);

    /// cteni z XML
    void readFromXML(TPackage * package);
};

// +--------------------------------------------------------------------------+
// |                                                                          |
// |              Z P R A V Y   M E Z I   A I   A   W O R L D                 |
// |                                                                          |
// +--------------------------------------------------------------------------+

/// Zprava pro AI, ze svet je nacten.
/// @see MSG_AI_LETS_GO
class TPacket_AI_LetsGo 
{
public:
    /// struktury z pravidel
    World::TRules * rules;

    /// mapa
    World::TMap * map;

    /// mesta
    World::TTownsContainer * towns;

    /// kralovstvi
    World::TKingdomsContainer * kingdoms;

    /// masky viditelnosti
    World::TVisibilityMapsContainer * visibility_maps;

    /// zive jednotky
    World::TLivingUnitsContainer * living_units;

    /// mrtve jednotky
    World::TLivingUnitsContainer * corpses;

    /// zive budovy
    World::TLivingBuildingsContainer * living_buildings;

    /// znicene budovy
    World::TLivingBuildingsContainer * ruins;

    /// seznam hracu
    World::TPlayerContainer * players;    
};

/// Zprava pro AI, ze se ma ujmout tahu.
/// @see MSG_AI_ON_TURN
class TPacket_AI_ON_TURN
{
public:
    /// id hrace, ktereho ma umela inteligence predstavovat
    int player_id;

    /// aktualni maximalni ID herni akce obsazene v kontejneru historie
    int max_action_id;

    /// obtiznost pocitacoveho hrace, jenz je na tahu
    World::PLAYER_TYPE ai_level;
};

/** Ziskani akce ze zasobniku historie */
struct PACKET_AI_GET_HISTORY_PIECE : public PACKET_WORLD_AI
{
    /// dosavadni maximalni ID herni akce, ktere AI zpracovala
    int max_action_id;
};

/** Zprava vracejici akci ze zasobniku historie */
class TPacket_AI_RET_HISTORY_PIECE : public TPacket_SyncResult
{
public:
    /// konstruktor
    TPacket_AI_RET_HISTORY_PIECE();

    /// id akce
    ACTION_ID action_id;

    /// typ akce
    World::ACTION_TYPE action_type;

    /// data k akci vazana
    void * data;
};

// +--------------------------------------------------------------------------+
// |                                                                          |
// |        Z P R A V Y   M E Z I   A I   A   M A P  A N A L Y Z A T O R      |
// |                                                                          |
// +--------------------------------------------------------------------------+


/// Zprava pro Map Analyzator, ktera vyhleda pro daneho hrace nejblizsi produkcni budovu, ktera umi vyrobit danou jednotku.
/// @see MSG_MAPANALYZER_GET_NEAREST_FREE_PRODUCTION_BUILDING_FOR_UNIT_TRAIN
class TPacket_MapAnalyzer_NearestFreeProductionBuildingForUnitTrain
{
public:
    /// ID hrace, pro ktereho umisteni hledam
    int player_id;
    
    /// typ jednotky jenz chci produkovat
    int unit_type;

    /// souradnice hexu, od ktereho merim vzdalenost
    int start_hex;
};

/// Zprava pro Map Analyzator, ktera vyhleda pro daneho hrace nejlepsi misto pro stavbu budovy.
/// @see MSG_MAPANALYZER_GET_BEST_PRODUCTION_BUILDING_PLACE
class TPacket_MapAnalyzer_BestProductionBuildingPlace
{
public:
    /// Konstruktor
    TPacket_MapAnalyzer_BestProductionBuildingPlace() { planned_buildings_coordinates.clear(); }

    /// Destruktor
    ~TPacket_MapAnalyzer_BestProductionBuildingPlace() { planned_buildings_coordinates.clear(); }

    /// ID hrace, pro ktereho umisteni hledam
    int player_id;
    
    /// typ budovy jenz chci stavet
    int building_type;
    
    /// pozice stavitele, jenz ma stavbu prirazenu - ma hodnotu (-1), pokud jeste zadny stavitel prirazen neni
    int assigned_builder_position;

    /// list se souradnicemi planovanych budov, co maji byt postaveny
    std::list<int> planned_buildings_coordinates;
};

/// Zprava pro Map Analyzator, ktera najde pomyslny stred hracova uzemi.
/// @see MSG_MAPANALYZER_FIND_PLAYER_EMPIRE_CENTER
class TPacket_MapAnalyzer_FindPlayerEmpireCenter
{
public:
    /// ID hrace, pro ktereho stred hledam
    int player_id;
};

/// Zprava pro Map Analyzator, ktera najde nahodny neprozkoumany hex pro daneho hrace.
/// @see MSG_MAPANALYZER_GET_RANDOM_INVISIBLE_HEX
class TPacket_MapAnalyzer_GetRandomInvisibleHex
{
public:
    /// ID hrace, pro ktereho hledam nahodny neprozkoumany hex
    int player_id;
};

/// Zprava pro Map Analyzator, ktera zmeri prostou vzdalenost mezi dvema hexy.
/// @see MSG_MAPANALYZER_GET_HEX_DISTANCE
class TPacket_MapAnalyzer_GetHexDistance
{
public:
    /// IDcka hexu, mezi kterymi merim vzdalenost
    int hexid1;

    /// IDcka hexu, mezi kterymi merim vzdalenost
    int hexid2;
};

/// Zprava pro Map Analyzator, ktera zmeri prostou vzdalenost mezi dvema hexy.
/// @see MSG_MAPANALYZER_FIND_CITY_BY_POSITION
class TPacket_MapAnalyzer_FindCityByPosition
{
public:
    /// IDcko hexu s pozici mesta
    int position;
};

/// Packet pro Map Analyzator, ktery nese informace o pozadovanem akcnim radiusu od zadaneho hexu.
/// @see MSG_MAPANALYZER_GET_ACTION_RADIUS
/// @return TPacket_MapAnalyzer_ActionRadiusResponse
class TPacket_MapAnalyzer_GetActionRadius
{
public:
    /// IDcko hexu od ktereho chceme akcni radius
    int position;

    /// polomer akcniho radiusu
    int radius_size;
};

/// Navratova trida Map Analyzatoru, ktera obsahuje seznam hexu akcniho radiusu od zadaneho hexu.
/// @see MSG_MAPANALYZER_GET_ACTION_RADIUS
class TPacket_MapAnalyzer_ActionRadiusResponse
{
public:
  /// konstruktor
  TPacket_MapAnalyzer_ActionRadiusResponse() { radius_hexes.clear(); }
    
  /// destruktor
  ~TPacket_MapAnalyzer_ActionRadiusResponse() { radius_hexes.clear(); }
    
  /// seznam hexu akcniho radiusu
  std::list<int> radius_hexes;
};

/// Zprava pro Map Analyzator, ktera najde nejvhodnejsi umisteni pro stavbu mostu pro daneho hrace.
/// @see MSG_MAPANALYZER_BRIDGE_BUILDING_ANALYSIS
class TPacket_MapAnalyzer_BridgeBuildingAnalysis
{
public:
    /// ID hrace, pro ktereho hledam nejvhodnejsi umisteni mostu
    int player_id;
};

/// Zprava pro Map Analyzator, ktera urci, zdali je jednotku daneho typu mozne vytrenovat v budove na zadanem hexu.
/// @see MSG_MAPANALYZER_UNIT_TRAINABLE_IN_BUILDING
class TPacket_MapAnalyzer_UnitTrainableInBuilding
{
public:
    /// typ jednotky, na kterou se ptam
    int unit_type;

    /// ID hexu, na kterem stoji budova jenz testuji
    int building_placement;
};

/// Navratova trida Map Analyzatoru, ktera obsahuje informace o planovane stavbe mostu.
/// @see MSG_MAPANALYZER_BRIDGE_BUILDING_ANALYSIS
class TPacket_MapAnalyzer_BridgeBuildingResponse
{
public:
  /// umisteni mostu
  int bridge_placement;

  /// orientace mostu
  int bridge_orientation;

  /// cilovy hex stavitele, ze ktereho ma zacit stavet
  int builder_position;
};

/// Navratova trida Map Analyzatoru, ktera obsahuje informace o planovane stavbe mostu.
/// @see MSG_MAPANALYZER_FIND_UNIT_POSITION
class TPacket_MapAnalyzer_FindUnitPosition
{
public:
  /// World ID jednotky
  int unit_world_id;
};


// +--------------------------------------------------------------------------+
// |                                                                          |
// |            Z P R A V Y   M E Z I   A I   A   D I P L O M A C Y           |
// |                                                                          |
// +--------------------------------------------------------------------------+

/// Zprava pro Diplomacy, ze se je pridavan diplomat.
/// @see MSG_DIPLOMACY_ADD_DIPLOMAT
class TPacket_Diplomacy_AddDiplomat
{
public:
    /// jednoznacne ID diplomata
    int diplomat_world_id;

    // typ diplomata (DT_HUMAN nebo DT_AI)
    ai_ns::diplomacy_ns::TDiplomatType diplomat_type;
};

/// Zprava pro Diplomacy, ze se je odebiran diplomat.
/// @see MSG_DIPLOMACY_REMOVE_DIPLOMAT
class TPacket_Diplomacy_RemoveDiplomat
{
public:
    /// jednoznacne ID diplomata
    int diplomat_world_id;
};

/// Zprava pro Diplomacy o provedenem utoku.
/// @see MSG_DIPLOMACY_ATTACK_INFO
class TPacket_Diplomacy_AttackInfo
{
public:
    /// jednoznacne ID diplomatu
    int attacking_diplomat_world_id;
    /// jednoznacne ID diplomatu
    int attacked_diplomat_world_id;

    /// typy ucastnicich se jednotek
    int attacking_unit_type;
    /// typy ucastnicich se jednotek
    int attacked_unit_type;

    /// ceny ucastnicich se jednotek
    int attacking_unit_cost;
    /// ceny ucastnicich se jednotek
    int attacked_unit_cost;

    /// typ napadene budovy
    int attacked_building_type;

    /// cena napadene budovy
    int attacked_building_cost;

    /// urovne ucastnicich se jednotek
    int attacking_unit_level;
    /// urovne ucastnicich se jednotek
    int attacked_unit_level;

    /// plne pocty muzu ucastnicich se jednotek
    int attacking_unit_men_number_full;
    /// plne pocty muzu ucastnicich se jednotek
    int attacked_unit_men_number_full;

    /// plny pocet zivotu napadene budovy
    int attacked_building_lives_full;

    /// pocty muzu ucastnicich se jednotek pred utokem
    int attacking_unit_men_number_before;
    /// pocty muzu ucastnicich se jednotek pred utokem
    int attacked_unit_men_number_before;

    /// pocet zivotu napadene budovy pred utokem
    int attacked_building_lives_before;

    /// pocty muzu ucastnicich se jednotek po utoku
    int attacking_unit_men_number_after;
    /// pocty muzu ucastnicich se jednotek po utoku
    int attacked_unit_men_number_after;

    /// pocet zivotu napadene budovy po utoku
    int attacked_building_lives_after;

    /// priznak, zdali je utoceno na mesto
    bool attack_on_city;
};

/// Zprava pro Diplomacy o dobyti budovy.
/// @see MSG_DIPLOMACY_CONQUER_BUILDING
class TPacket_Diplomacy_ConquerBuilding
{
public:
    /// jednoznacne ID diplomatu
    int attacking_diplomat_world_id;
    /// jednoznacne ID diplomatu
    int attacked_diplomat_world_id;
    
    /// typ obsazene budovy
    int building_type;

    /// cena vyroby obsazene budovy
    int building_cost;
};

/// Zprava pro Diplomacy o dobyti mesta.
/// @see MSG_DIPLOMACY_CONQUER_CITY
class TPacket_Diplomacy_ConquerCity
{
public:
    /// jednoznacne ID diplomatu
    int attacking_diplomat_world_id;
    /// jednoznacne ID diplomatu
    int attacked_diplomat_world_id;
    
    /// prijem obsazeneho mesta
    int payment;
};

/// Zprava pro Diplomacy, kterou zjistuji oficialni vztah mezi dvema hraci.
/// Vraci ai_ns::diplomacy_ns::TRelationship*
/// @see MSG_DIPLOMACY_GET_RELATION
class TPacket_Diplomacy_GetRelation
{
public:
    /// jednoznacne ID diplomatu
    int source_diplomat_world_id;
    /// jednoznacne ID diplomatu
    int target_diplomat_world_id;
};

/// Zprava pro Diplomacy o zmene oficialniho diplomatickeho vztahu.
/// @see MSG_DIPLOMACY_SET_RELATION
class TPacket_Diplomacy_SetRelation
{
public:
    /// jednoznacne ID diplomatu
    int source_diplomat_world_id;
    /// jednoznacne ID diplomatu
    int target_diplomat_world_id;
    /// nastavovany vztah
    ai_ns::diplomacy_ns::TRelationship set_relationship;
};

/// Zprava pro Diplomacy, kterou zjistuji oficialni vztah mezi dvema hraci.
/// Vraci ai_ns::diplomacy_ns::TRelationship*
/// @see MSG_DIPLOMACY_GET_OFFER_RELATION
class TPacket_Diplomacy_GetOfferRelation
{
public:
    /// jednoznacne ID diplomatu
    int source_diplomat_world_id;
    /// jednoznacne ID diplomatu
    int target_diplomat_world_id;
};

/// Zprava pro Diplomacy o zmene navrhovaneho diplomatickeho vztahu.
/// @see MSG_DIPLOMACY_SET_OFFER_RELATION
class TPacket_Diplomacy_SetOfferRelation 
{
public:
    /// jednoznacne ID diplomatu
    int source_diplomat_world_id;
    /// jednoznacne ID diplomatu
    int target_diplomat_world_id;
    
    /// navrhovovany vztah
    ai_ns::diplomacy_ns::TRelationship offered_relationship;
};

/// Zprava pro Diplomacy, kterou zjistuji ostatni polozky diplomatickeho vztahu.
/// Vraci ai_ns::diplomacy_ns::TRelationshipProperties*
/// @see MSG_DIPLOMACY_GET_OTHER_RELATION_PROPERTIES
class TPacket_Diplomacy_GetOtherRelationProperties
{
public:
    /// jednoznacne ID diplomatu
    int source_diplomat_world_id;
    /// jednoznacne ID diplomatu
    int target_diplomat_world_id;
};

/// Zprava pro Diplomacy o zmene ostatnich polozek diplomatickeho vztahu.
/// @see MSG_DIPLOMACY_SET_OTHER_RELATION_PROPERTIES
class TPacket_Diplomacy_SetOtherRelationProperties 
{
public:
    /// jednoznacne ID diplomatu
    int source_diplomat_world_id;
    /// jednoznacne ID diplomatu
    int target_diplomat_world_id;
    
    /// navrhovovany vztah
    ai_ns::diplomacy_ns::TRelationshipProperties relationship_properties;
};

/// Zprava pro Diplomacy, ze dany diplomat zacina sve kolo.
/// @see MSG_DIPLOMACY_DIPLOMAT_TURN_BEGIN
class TPacket_Diplomacy_DiplomatTurnBegin
{
public:
    /// jednoznacne ID diplomata
    int diplomat_world_id;
};

/// Zprava pro Diplomacy, ze dany diplomat konci sve kolo.
/// @see MSG_DIPLOMACY_DIPLOMAT_TURN_END
class TPacket_Diplomacy_DiplomatTurnEnd
{
public:
    /// jednoznacne ID diplomata
    int diplomat_world_id;
};

/// Dotaz pro Diplomacy od AI na nejvetsiho nepritele
/// @see MSG_DIPLOMACY_FIND_WORST_ENEMY
class TPacket_Diplomacy_FindWorstEnemy
{
public:
    /// jednoznacne ID diplomata
    int diplomat_world_id;

    /// diplotame, kteri jsou vynechani
    std::list<int>* excluded_diplomats;
};

/// Zpusobi pro AI hrace vyhlaseni valky nejhorsimu nepriteli
/// @see MSG_DIPLOMACY_AI_WAR_DECLARATION
class TPacket_Diplomacy_AIWarDeclaration
{
public:
    /// World ID AI diplomata
    int ai_diplomat_world_id;

    /// World ID nepritele AI diplomata
    int enemy_diplomat_of_ai_world_id;
};

/// Zpusobi zmenu typu diplomata (Ai/clovek) na pozadovany typ.
/// @see MSG_DIPLOMACY_CHANGE_DIPLOMAT_TYPE
class TPacket_Diplomacy_ChangeDiplomatType
{
public:
    /// World ID diplomata
    int diplomat_world_id;

    /// novy typ diplomata
    ai_ns::diplomacy_ns::TDiplomatType new_diplomat_type;
};

/// Packet obsahujici aktualizacni informace o vztazich v diplomacii.
/// @see MSG_DIPLOMACY_CHANGE
class TPacket_DIPLOMACY_CHANGE
{
public:
  TPacket_DIPLOMACY_CHANGE() { changes.clear(); };
  ~TPacket_DIPLOMACY_CHANGE() { changes.clear(); };
  /// vektor se zmenami diplomacie
  std::vector<ai_ns::diplomacy_ns::TDipChange> changes;
};

// +--------------------------------------------------------------------------+
// |                                                                          |
// |          Z P R A V Y   S P E C I F I C K E   P R O   E D I T O R         |
// |                                                                          |
// +--------------------------------------------------------------------------+

/// inicializace a aktualizace dat WORLDu pri behu v editoru
struct PACKET_EDITOR_RELOAD
{
    /// hlavicka mapy
    rm::TMapInfo* map_info;

    /// telo mapy
    World::HEX ** hexes;

    /// mesta na mape
    DA<rm::TCityInfo *> * cities;

    /// kralovstvi v mape
    World::KINGDOM ** kingdoms;

    /// seznam jednotek
    DA<World::LIVING_UNIT *> * living_units;

    /// seznam budov
    DA<World::LIVING_BUILDING *> * living_buildings;

    /// Potvrzovat nacteni? Nastavit na \c true pri incializaci,
    /// ponechat na \c false pri prubezne aktualizaci.
    bool confirm;
};

/// odpoved na inicializaci WORLDu pri behu v editoru
class TPacket_EDITOR_RELOADED
{
public:
    /// struktury z pravidel
    World::TRules * rules;

    /// mapa
    World::TMap * map;

    /// mesta
    World::TTownsContainer * towns;

    /// kralovstvi
    World::TKingdomsContainer * kingdoms;

    /// masky viditelnosti
    World::TVisibilityMapsContainer * visibility_maps;

    /// zive jednotky
    World::TLivingUnitsContainer * living_units;

    /// mrtve jednotky
    World::TLivingUnitsContainer * corpses;

    /// zive budovy
    World::TLivingBuildingsContainer * living_buildings;

    /// znicene budovy
    World::TLivingBuildingsContainer * ruins;

    /// seznam hracu
    World::TPlayerContainer * players;    
};

// +--------------------------------------------------------------------------+
// |                                                                          |
// |           Z P R A V Y   P O S I L A N E   P O   S I T I                  |
// |                                                                          |
// +--------------------------------------------------------------------------+

/** Pripojeni hrace pres sit */
namespace PACKET_NET_CONNECTION
{
    /// id hrace
    const char player_id[] = "player_id";

    /// nastaveni hrace
    const char settings[]  = "settings";
};

/** synchornizacni balicek */
namespace PACKET_SYNCHRO
{
    /// id hrace
    const char player_id[] = "player_id";

    /// id akce
    const char action_id[] = "action_id";

    /// typ akce
    const char action_type[] = "action_type";
};

/** Potvrzeni o prijeti synchronizacniho balicku */
namespace PACKET_SYNCHRO_CONFIRMATION
{
    /// id hrace
    const char player_id[] = "player_id";

    /// id akce
    const char action_id[] = "action_id";

    /// vyhodnoceni zpracovani
    const char status[]    = "status";
};

/** Sitova zprava o chybe pri zpracovani */
namespace PACKET_NET_ERROR_REPLY
{
    /// specifikace chyby
    const char error[] = "error";
};

/// Paket pro pripojeni hrace do hry
namespace PACKET_JOIN_GAME 
{
    /// id hrace
    const char player_id[] = "player_id";

    /// id jeho sitoveho klienta
    const char client_id[] = "client_id";
};

/** Zaklad zpravy posilane po siti mezi moduly WORLD_SERVER a WORLD_CLIENT 
    @see PACKET_TO_CLIENT
*/
namespace PACKET_WORLD
{
    /// id hrace
    const char player_id[] = "player_id";
    
    /// typ zpravy
    const char type[]      = "type";
};

/// sitova zadost o sebevrazdu jednotky
namespace PACKET_NET_ACT_UNIT_SUICIDE
{
    /// id jednotky
    const char unit_id[] = "unit_id";
};

/// sitova zadost o sebezniceni budovy
namespace PACKET_NET_ACT_BUILDING_SUICIDE
{
    /// id budovy
    const char building_id[] = "building_id";
};

/// sitova zadost o koupi jednotky
namespace PACKET_NET_ACT_BUILDING_RECRUIT_UNIT
{
    /// id budovy
    const char building_id[] = "building_id";

    /// typ jednotky
    const char unit_type[] = "unit_type";
};

/// sitova zadost o zahajeni stavby
namespace PACKET_NET_ACT_START_BUILDING
{
    /// id jednotky
    const char unit_id[] = "unit_id";

    /// typ budovy
    const char building_type[] = "building_type";

    /// umisteni budovy
    const char location[] = "location";

    /// orientace budovy
    const char orientation[] = "orientation";
};

/// sitova zadost o zahajeni opravy
namespace PACKET_NET_ACT_REPAIR_BUILDING
{
    /// id jednotky
    const char unit_id[] = "unit_id";

    /// id budovy
    const char building_id[] = "building_id";
};

/// sitova zadost o prodej bonusu
namespace PACKET_NET_ACT_SELL_BONUS
{
    /// id budovy
    const char building_id[] = "building_id";

    /// id bonusu
    const char bonus_id[] = "bonus_id";
};

/// sitova zadost o preruseni stavby budovy
namespace PACKET_NET_ACT_STOP_BUILDING
{
    /// id jednotky
    const char unit_id[] = "unit_id";
};

/// sitova zadost o zahajeni pohybu jednotky
namespace PACKET_NET_ACT_UNIT_MOVE
{
    /// id jednotky
    const char unit_id[] = "unit_id";

    /// id ciloveho hexu
    const char to[]      = "to";
};

/// sitova zadost o souboj jednotek
namespace PACKET_NET_ACT_UNIT_ATTACK
{
    /// id utocici jednotky
    const char attacker[] = "unit_id";

    /// id napadeneho hexu
    const char target[]   = "target";
};

/// sitova zprava o zahajeni hry
namespace PACKET_NET_GAME_START
{
    /// id hrace na tahu
    const char player_on_turn[] = "player_on_turn";

    /// cislo tahu
    const char turn_id[] = "turn_id";
};

/// sitova zprava o ukonceni hry na popud hrace
namespace PACKET_NET_END_GAME_GUI
{
    /// rozliseni definitivniho ukonceni a odpojeni hrace
    const char final[] = "final";
};

/// sitova zprava o ukonceni hry
namespace PACKET_NET_END_GAME
{
    /// duvod ukonceni
    const char reason[] = "reason";

    /// id vitezneho hrace
    const char winner[] = "winner";
};

/// sitova zadost o leceni jednotky
namespace PACKET_NET_ACT_UNIT_HEAL
{
    /// id jednotky
    const char unit_id[] = "unit_id";
};

/// sitova zadost o neelitni doplneni jednotky
namespace PACKET_NET_ACT_UNIT_RECRUIT
{
    /// id jednotky
    const char unit_id[] = "unit_id";
};

/// sitova zadost o elitni doplneni jednotky
namespace PACKET_NET_ACT_UNIT_RECRUIT_ELITE
{
    /// id jednotky
    const char unit_id[] = "unit_id";
};

/// sitova zadost o zmenu diplomatickeho vztahu
namespace PACKET_NET_ACT_OFFER_RELATION
{
    /// id diplomata zadajiciho zmenu
    const char source_diplomat_world_id[] = "source_diplomat_world_id";

    /// id protistrany
    const char target_diplomat_world_id[] = "target_diplomat_world_id";

    /// navrhovany vztah
    const char offered_relationship[] = "offered_relationship";
};

#endif 
/*****************************************************************************/
